package app

import (
	"errors"
	"fmt"
	"net/http"
	"os"
	"sync"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/client"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/controller"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/grpcserver"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/logging"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/middleware"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/service"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry/wavefront"

	"github.com/fsnotify/fsnotify"
	"github.com/labstack/echo/v4"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

type Starter interface {
	Start(address string) error
}

type App struct {
	grpcStarter     *grpcserver.DeviceServiceServer
	Starter         Starter
	Stop            chan os.Signal
	ClearToShutdown <-chan struct{}
}

// Start initializes the web+grpc servers and fatally terminates the app if an error is encountered.
func (app *App) Start() {
	port := viper.GetUint16(config.WebPortKey)

	go app.grpcStarter.Start(app.ClearToShutdown)

	log.Info().Msgf("starting http server on [::]:%d", port)
	err := app.Starter.Start(fmt.Sprintf(":%d", port))

	if err != nil && !errors.Is(err, http.ErrServerClosed) {
		log.Fatal().Err(err).Msg("could not start http server")
	}

	<-app.ClearToShutdown
}

// New returns a new app with graceful shutdown helpers.
func New() *App {
	echoServer, grpcServer, shutdownHandler := InitAppWithLogger(&log.Logger)
	stop, clearToShutdown := GetShutdownChannel(echoServer, grpcServer, shutdownHandler)

	return &App{grpcStarter: grpcServer, Starter: echoServer, Stop: stop, ClearToShutdown: clearToShutdown}
}

func InitAppWithLogger(fatalLogger FatalLogger) (*echo.Echo, *grpcserver.DeviceServiceServer, ShutdownFunc) {
	err := config.Init()
	// ConfigFileNotFoundError has no Is method, so we need to use As for equivalence check
	if err != nil && !errors.As(err, &viper.ConfigFileNotFoundError{}) {
		fatalLogger.Fatal().Err(err).Msg("config failed to initialize")
	}

	logging.Configure()
	viper.OnConfigChange(func(_ fsnotify.Event) { logging.Configure() })

	configureTelemetry()

	_, _ = data.NewTable(data.NewDynamoDBClient(), data.DeviceTableSpecs())

	cbdClient := client.NewCbdClient(client.NewCbdClientConfig())
	cbdService := service.NewCbdService(cbdClient, viper.GetBool(config.CbdClientEnabledKey))

	deviceEventsProducerConfig := data.NewProducerClientConfig(config.KinesisStreamNameDeviceEventsKey)
	deviceEventsProducer := data.NewProducerClient(deviceEventsProducerConfig)
	deviceEventsService := service.NewDeviceEventsService(deviceEventsProducer,
		viper.GetBool(config.DeviceEventsEnabledKey), viper.GetBool(config.AppDryRunKey))
	deviceAuthClient := client.NewDeviceAuthClient(client.NewDeviceAuthClientCfg())

	evictService := service.NewEvictService(service.NewEvictServiceConfig(), deviceAuthClient)

	esManager := data.NewESClientManager(data.NewESConfigs()...)
	database := data.InitDB(data.NewDBConfigs())

	readinessService := service.NewReadinessService(deviceAuthClient, database, deviceEventsProducer, esManager)
	svc := service.New()

	grpcServer := grpcserver.NewDeviceServiceServer(database, evictService, deviceEventsService, cbdService)

	echoServer := newEcho()
	middleware.Apply(echoServer)
	controller.Route(echoServer, svc, readinessService)

	svcManagers := []service.Manager{svc, database, evictService}

	return echoServer, grpcServer, generateShutdownHandler(svcManagers)
}

func configureTelemetry() {
	sender := wavefront.InitSender()
	telemetry.SetupMetrics(sender)
	telemetry.SetupTracing(sender)
}

func newEcho() *echo.Echo {
	echoServer := echo.New()

	echoServer.HideBanner = !viper.GetBool(config.DevModeKey)
	echoServer.HidePort = !viper.GetBool(config.DevModeKey)
	echoServer.Debug = viper.GetBool(config.DevModeKey)
	echoServer.Binder = model.NewProtobufBinder()
	echoServer.Server.ReadHeaderTimeout = viper.GetDuration(config.HTTPReadHeaderTimeoutKey)
	echoServer.Server.ReadTimeout = viper.GetDuration(config.HTTPReadTimeoutKey)

	return echoServer
}

//nolint:godox // temporary TODO for template
// TODO: gracefully shutdown the contexts and connections.

func generateShutdownHandler(services []service.Manager) ShutdownFunc {
	return func(reporter ShutdownReporter) *sync.WaitGroup {
		appName := viper.GetString(config.AppNameKey)

		reporter.Stopping(appName)
		defer reporter.Stopped(appName)

		waitGroup := &sync.WaitGroup{}
		waitGroup.Add(len(services))

		for _, svc := range services {
			// avoid services blocking one another while attempting to close
			go func(svc service.Manager, waitGroup *sync.WaitGroup) {
				// stop service, close + flush any connections/channels
				_ = svc.FlushAndClose()

				reporter.Stopping(svc.Name())
				defer reporter.Stopped(svc.Name())

				waitGroup.Done()
			}(svc, waitGroup)
		}

		return waitGroup
	}
}
