package app_test

import (
	"fmt"
	"net/http"
	"net/url"
	"os"
	"strings"
	"testing"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/app"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	"github.com/labstack/echo/v4"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

//nolint:gochecknoglobals // singleton test controller
var ctrl *gomock.Controller

func TestApp(t *testing.T) {
	ctrl = gomock.NewController(t)
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "App")
}

var _ = Describe("App", func() {
	var testServer *echo.Echo
	var mockWriter *mocks.Writer
	var mockResponseWriter *mocks.ResponseWriter
	var responseHeaders http.Header
	var request *http.Request
	var appName string

	BeforeEach(func() {
		os.Clearenv()
		viper.Reset()
		Expect(len(viper.AllKeys())).To(BeZero())
		Expect(len(os.Environ())).To(BeZero())
		os.Setenv(strings.ToUpper(strings.ReplaceAll(config.TracingEnabledKey, ".", "_")), "false")
		os.Setenv(strings.ToUpper(strings.ReplaceAll(config.LogLevelKey, ".", "_")), "info")
		// allow time for debugging
		// os.Setenv(strings.ToUpper(strings.Replace(config.HTTPRequestTimeoutKey, ".", "_", -1)), "5m")
	})

	Context("happy InitApp method", func() {
		BeforeEach(func() {
			testServer, _, _ = app.InitAppWithLogger(&log.Logger)
			appName = viper.GetString(config.AppNameKey)
		})

		It("should set config defaults", func() {
			Expect(len(viper.AllKeys())).ToNot(BeZero())
		})

		It("should create a pscweb app", func() {
			Expect(testServer).ToNot(BeNil())
		})

		It("should apply middleware", func() {
			//nolint:exhaustruct // only used for testing, just need method, url, header
			request = &http.Request{
				Method: http.MethodPost,
				//nolint:exhaustruct // only used for testing, just need path
				URL:    &url.URL{Path: fmt.Sprintf("/%s/v1", appName)},
				Header: make(http.Header),
			}

			mockResponseWriter = mocks.NewResponseWriter(ctrl)
			responseHeaders = make(http.Header)
			mockWriter = mocks.NewWriter(ctrl)
			logger := log.With().Logger().Output(mockWriter)
			request = request.WithContext(logger.WithContext(request.Context()))

			var testMiddlewareLog string
			var serviceLog string
			var requestLoggerLog string

			gomock.InOrder(
				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&testMiddlewareLog),
				).Times(1),
				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&serviceLog),
				).Times(1),
				mockResponseWriter.EXPECT().Header().Return(responseHeaders).Times(1),
				mockResponseWriter.EXPECT().WriteHeader(http.StatusAccepted).Times(1),
				mockResponseWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Times(1),
				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&requestLoggerLog),
				).Times(1),
			)

			testServer.Use(func(next echo.HandlerFunc) echo.HandlerFunc {
				return func(ctx echo.Context) error {
					logger := zerolog.Ctx(ctx.Request().Context())
					logger.Warn().Msg("app middleware called")

					return next(ctx)
				}
			})

			testServer.ServeHTTP(mockResponseWriter, request)

			Expect(testMiddlewareLog).To(ContainSubstring("\"message\":\"app middleware called\""))
			Expect(testMiddlewareLog).To(ContainSubstring("\"request_id\":"))

			Expect(serviceLog).To(ContainSubstring("\"message\":\"service be doin stuff\""))
			Expect(serviceLog).To(ContainSubstring("\"request_id\":"))

			Expect(requestLoggerLog).To(ContainSubstring("\"message\":\"request\""))
			Expect(requestLoggerLog).To(ContainSubstring("\"request_id\":"))
		})
	})

	Context("sad InitApp method", func() {
		It("should error with bad config", func() {
			logger := helper.NewNonFatalLogger()

			helper.GenerateBadConfig()
			defer helper.CleanupConfig()

			Expect(func() { app.InitAppWithLogger(logger) }).To(Panic())
			Expect(logger.FatalCallsCount).To(Equal(1))
		})
	})

	Context("Start method", func() {
		It("should succeed", func() {
			viper.Set(config.AppShutdownWaitKey, "0")

			app := app.New()
			Expect(app).ToNot(BeNil())

			go func() {
				app.Start()
			}()
			time.Sleep(10 * time.Millisecond)
			Eventually(app.Stop).ShouldNot(BeClosed())
			Eventually(app.ClearToShutdown).ShouldNot(BeClosed())

			app.Stop <- os.Interrupt
			time.Sleep(10 * time.Millisecond)
			Eventually(app.ClearToShutdown).Should(BeClosed())
		})
	})
})
