package app

import (
	"context"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"

	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

type Shutdowner interface {
	Shutdown(context.Context) error
}

type ShutdownReporter interface {
	Stopping(string)
	Stopped(string)
}

type ShutdownFunc func(reporter ShutdownReporter) *sync.WaitGroup

type shutdownReporter struct {
	log *zerolog.Event
	mu  *sync.Mutex
}

func (reporter *shutdownReporter) Stopping(name string) {
	reporter.mu.Lock()
	defer reporter.mu.Unlock()
	reporter.log.Str(name, "stopping")
}

func (reporter *shutdownReporter) Stopped(name string) {
	reporter.mu.Lock()
	defer reporter.mu.Unlock()
	reporter.log.Str(name, "stopped")
}

func GetShutdownChannel(
	server Shutdowner,
	grpcServer Shutdowner,
	shutdownFunc ShutdownFunc,
) (chan os.Signal, <-chan struct{}) {
	stop := make(chan os.Signal, 1)
	clearToShutdown := make(chan struct{})

	go func() {
		signal.Notify(stop, os.Interrupt, syscall.SIGINT, syscall.SIGTERM)
		<-stop

		// Wait to shutdown app to ensure ingress controller has updated list of running pods
		shutdownWait := viper.GetDuration(config.AppShutdownWaitKey)
		time.Sleep(shutdownWait)

		log.Info().Msg("Attempting graceful shutdown...")
		logEvent := log.Info()
		reporter := &shutdownReporter{
			logEvent,
			&sync.Mutex{},
		}

		reporter.Stopping("http")

		if err := server.Shutdown(context.Background()); err != nil {
			log.Err(err).Msg("HTTP shutdown failed")
		} else {
			reporter.Stopped("http")
		}

		reporter.Stopping("grpc")

		if err := grpcServer.Shutdown(context.Background()); err != nil {
			log.Err(err).Msg("gRPC shutdown failed")
		} else {
			reporter.Stopped("grpc")
		}

		// shutdownFunc closes all outstanding clients/connections to infra
		if shutdownFunc != nil {
			shutdownFunc(reporter).Wait()
		}

		reporter.mu.Lock()
		logEvent.Msg("Graceful shutdown complete")
		reporter.mu.Unlock()

		close(clearToShutdown)
	}()

	return stop, clearToShutdown
}
