package app_test

import (
	"errors"
	"os"
	"sync"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/app"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
)

var ErrShutdownFailure = errors.New("failed to shutdown")

func generateHandleShutdown() (chan struct{}, app.ShutdownFunc) {
	shutdownHandled := make(chan struct{})

	handleShutdownFunc := func(reporter app.ShutdownReporter) *sync.WaitGroup {
		Expect(reporter).ToNot(BeNil())
		reporter.Stopping("test")
		reporter.Stopped("test")
		close(shutdownHandled)

		return &sync.WaitGroup{}
	}

	return shutdownHandled, handleShutdownFunc
}

var _ = Describe("Shutdown", func() {
	BeforeEach(func() {
		viper.Reset()
	})

	Context("getShutdownChannel function", func() {
		var stop chan os.Signal
		var clearToShutdown <-chan struct{}
		var shutdownHandled chan struct{}
		var handleShutdownFunc app.ShutdownFunc
		var mockShutdowner *mocks.Shutdowner
		var mockGrpcShutdowner *mocks.Shutdowner

		BeforeEach(func() {
			mockShutdowner = mocks.NewShutdowner(ctrl)
			mockGrpcShutdowner = mocks.NewShutdowner(ctrl)
			shutdownHandled, handleShutdownFunc = generateHandleShutdown()
		})

		Context("with successful shutdowner shutdown", func() {
			BeforeEach(func() {
				mockShutdowner.EXPECT().Shutdown(gomock.Any()).Times(1)
				mockGrpcShutdowner.EXPECT().Shutdown(gomock.Any()).Times(1)
			})

			It("should close the clear to shutdown channel", func() {
				stop, clearToShutdown = app.GetShutdownChannel(mockShutdowner, mockGrpcShutdowner, handleShutdownFunc)
				Expect(stop).ToNot(BeClosed())
				stop <- os.Interrupt

				Eventually(shutdownHandled).Should(BeClosed())
				Eventually(clearToShutdown).Should(BeClosed())
			})

			It("should wait before shutting down", func() {
				viper.SetDefault(config.AppShutdownWaitKey, "1s")

				stop, clearToShutdown = app.GetShutdownChannel(mockShutdowner, mockGrpcShutdowner, handleShutdownFunc)
				Expect(stop).ToNot(BeClosed())
				stop <- os.Interrupt

				// give time for shutdown goroutine to complete
				time.Sleep(900 * time.Millisecond)

				Expect(shutdownHandled).ToNot(BeClosed())
				Expect(clearToShutdown).ToNot(BeClosed())
				Eventually(shutdownHandled).Should(BeClosed())
				Eventually(clearToShutdown).Should(BeClosed())
			})
		})

		Context("with unsuccessful shutdowner shutdown", func() {
			BeforeEach(func() {
				mockShutdowner.EXPECT().Shutdown(gomock.Any()).Return(ErrShutdownFailure).Times(1)
				mockGrpcShutdowner.EXPECT().Shutdown(gomock.Any()).Return(ErrShutdownFailure).Times(1)
			})

			It("should close the clear to shutdown channel", func() {
				stop, clearToShutdown = app.GetShutdownChannel(mockShutdowner, mockGrpcShutdowner, handleShutdownFunc)
				Expect(stop).ToNot(BeClosed())
				stop <- os.Interrupt

				Eventually(shutdownHandled).Should(BeClosed())
				Eventually(clearToShutdown).Should(BeClosed())
			})
		})
	})
})
