package client

import (
	"bytes"
	"context"
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"mime/multipart"
	"net/http"
	"strconv"
	"strings"
	"time"

	"github.com/opentracing/opentracing-go"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
)

var ErrCbdClientUnexpectedStatusCode = errors.New("cbd client received non-200 response")

type CbdCaller interface {
	Call(ctx context.Context, host, uri, headers, message string) error
}

type CbdClientConfig struct {
	requestTimeout time.Duration
	maxIdleConns   int
	maxConns       int
	idleTimeout    time.Duration
}

type CbdClient struct {
	telemetryPrefix string
	Client          *http.Client
}

func NewCbdClient(config CbdClientConfig) *CbdClient {
	client := &http.Client{
		Timeout: config.requestTimeout,
		Transport: &http.Transport{
			MaxIdleConns:        config.maxIdleConns,
			MaxIdleConnsPerHost: config.maxIdleConns,
			MaxConnsPerHost:     config.maxConns,
			IdleConnTimeout:     config.idleTimeout,
		},
	}

	return &CbdClient{
		telemetryPrefix: "cbdclient",
		Client:          client,
	}
}

func NewCbdClientConfig() CbdClientConfig {
	return CbdClientConfig{
		requestTimeout: viper.GetDuration(config.CbdClientRequestTimeoutKey),
		maxIdleConns:   viper.GetInt(config.CbdClientMaxIdleConnsKey),
		maxConns:       viper.GetInt(config.CbdClientMaxConnsKey),
		idleTimeout:    viper.GetDuration(config.CbdClientIdleConnTimeoutKey),
	}
}

func (cbdc *CbdClient) Call(ctx context.Context, host, uri, headers, message string) error {
	var err error

	span, ctx := opentracing.StartSpanFromContext(ctx, "CbdClientHTTPCall")
	defer span.Finish()

	api := strings.ReplaceAll(uri, "/", "")
	dims := map[string]string{}

	defer telemetry.CounterInc(fmt.Sprintf("%s.request.%s", cbdc.telemetryPrefix, api), dims, 1)

	var cbdRequest *http.Request

	if cbdRequest, err = createCbdRequest(ctx, host, uri, headers, message); err != nil {
		return fmt.Errorf("failed to create cbd request for %s %w", uri, err)
	}

	span.SetTag("http.url", fmt.Sprintf("%s://%s%s", cbdRequest.URL.Scheme, cbdRequest.URL.Host, uri))
	span.SetTag("http.method", cbdRequest.Method)

	//nolint:bodyclose // body is closed in closeRespBody below
	resp, err := cbdc.Client.Do(cbdRequest)
	if err != nil {
		span.SetTag("error", true)

		return fmt.Errorf("failed to call cbd request: %v | error: %w", cbdRequest, err)
	}

	defer closeRespBody(resp.Body)

	dims["code"] = strconv.Itoa(resp.StatusCode)
	span.SetTag("http.code", resp.StatusCode)

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf(
			"%w: expected %d received %d",
			ErrCbdClientUnexpectedStatusCode,
			http.StatusOK,
			resp.StatusCode)
	}

	return nil
}

func createCbdRequest(ctx context.Context, host, uri, headers, message string) (*http.Request, error) {
	httpHeaders, err := parseHeaders(headers)
	if err != nil {
		return nil, fmt.Errorf("failed to parse original request headeres from metadata %w", err)
	}

	var buf bytes.Buffer

	multipartWriter := multipart.NewWriter(&buf)

	messageBytes, err := base64.StdEncoding.DecodeString(message)
	if err != nil {
		return nil, fmt.Errorf("failed to base64 decode original msgpack request from metadata %w", err)
	}

	messageReader := bytes.NewReader(messageBytes)

	var fieldWriter io.Writer

	if fieldWriter, err = multipartWriter.CreateFormField("message"); err != nil {
		return nil, fmt.Errorf("failed to create form field message in multipart writer %w", err)
	}

	if _, err := io.Copy(fieldWriter, messageReader); err != nil {
		return nil, fmt.Errorf("failed to write message from reader to multipart writer %w", err)
	}

	if err = multipartWriter.Close(); err != nil {
		return nil, fmt.Errorf("failed to close multipart writer %w", err)
	}

	scheme := "https"

	if viper.GetString(config.AppEnvKey) == "local" {
		scheme = "http"
	}

	var req *http.Request

	req, err = http.NewRequestWithContext(
		ctx, http.MethodPost, fmt.Sprintf("%s://%s%s", scheme, host, uri), &buf)

	if err != nil {
		return nil, fmt.Errorf("failed to create Cbd Request %w", err)
	}

	delete(httpHeaders, "Content-Type")
	req.Header = httpHeaders
	req.Header.Set("Content-Type", multipartWriter.FormDataContentType())

	return req, nil
}

func parseHeaders(headers string) (http.Header, error) {
	headerMap := make(map[string][]string)
	canonicalHeaderMap := http.Header{}

	if err := json.Unmarshal([]byte(headers), &headerMap); err != nil {
		return nil, fmt.Errorf("failed to unmarshall json to map %w", err)
	}

	for key, values := range headerMap {
		for _, value := range values {
			canonicalHeaderMap.Add(key, value)
		}
	}

	return canonicalHeaderMap, nil
}
