package client_test

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"io"
	"mime/multipart"
	"net/http"
	"net/http/httptest"
	"strings"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/client"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
)

var _ = Describe("cbd client", func() {
	var cbdClient *client.CbdClient
	var server *httptest.Server
	var headers string
	var message string
	var host string

	Context("calls", func() {
		setupServer := func(expectedStatus int) {
			server = httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, req *http.Request) {
				// Confirm the msgpack message field is in the request
				reader, err := getMessageFieldReader(req)
				Expect(err).To(BeNil())
				msg, err := io.ReadAll(reader)
				Expect(err).To(BeNil())
				Expect(string(msg)).To(Equal("testmessage"))

				// Send expected response
				w.WriteHeader(expectedStatus)
				_, err = w.Write([]byte(`response`))
				if err != nil {
					return
				}
			}))

			host = strings.Split(server.URL, "://")[1]
			headerMap := map[string][]string{
				"whatever": {"test"},
			}
			headerBytes, err := json.Marshal(headerMap)
			Expect(err).To(BeNil())
			headers = string(headerBytes)
			message = base64.StdEncoding.EncodeToString([]byte("testmessage"))

			cbdClient = client.NewCbdClient(client.CbdClientConfig{})
			// substitute client for test client
			cbdClient.Client = server.Client()
		}

		BeforeEach(func() {
			viper.Set(config.AppEnvKey, "local")
			telemetry.SetupMetrics(nil)
		})

		It("status", func() {
			setupServer(http.StatusOK)
			defer server.Close()
			err := cbdClient.Call(context.Background(), host, "/services/emd_status/", headers, message)
			Expect(err).To(BeNil())
		})

		It("metadata", func() {
			setupServer(http.StatusOK)
			defer server.Close()
			err := cbdClient.Call(context.Background(), host, "/services/emd_metadata/", headers, message)
			Expect(err).To(BeNil())
		})
	})
})

func getMessageFieldReader(req *http.Request) (io.Reader, error) {
	var multipartReader *multipart.Reader

	var err error

	if multipartReader, err = req.MultipartReader(); err != nil {
		//nolint:errorlint // already have a wrapped err
		return nil, fmt.Errorf("%w", err)
	}

	currentPrt, err := multipartReader.NextPart()

	for err == nil && currentPrt.FormName() != "message" {
		currentPrt, err = multipartReader.NextPart()
	}

	if err != nil {
		//nolint:errorlint // already have a wrapped err
		return nil, fmt.Errorf("%w", err)
	}

	return currentPrt, nil
}
