package client

import (
	"context"
	"errors"
	"fmt"
	"io"
	"net/http"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"

	"github.com/opentracing/opentracing-go"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

const (
	deviceAuthEvictDeviceURL = "%s:%d/ems/v1/orgs/%s/device/%d"
	deviceAuthHealthCheckURL = "%s:%d/ems/health_check"
)

var (
	ErrDeviceAuthUnexpectedStatusCode = errors.New("device-auth responded with unexpected status code")
	ErrInvalidParamsProvided          = errors.New("error invalid params provided for delete request")
)

type Evictor interface {
	EvictDevice(ctx context.Context, orgKey string, deviceID int) error
	Close() error
}

type DeviceAuthClientCfg struct {
	Host           string
	idleTimeout    time.Duration
	maxIdleConns   int
	maxConns       int
	Port           int
	requestTimeout time.Duration
	skipReadyCheck bool
}

type DeviceAuthClient struct {
	cfg    DeviceAuthClientCfg
	Client *http.Client
}

var _ Evictor = (*DeviceAuthClient)(nil)

func NewDeviceAuthClientCfg() DeviceAuthClientCfg {
	// viper inits
	return DeviceAuthClientCfg{
		Host:           viper.GetString(config.DeviceAuthHostKey),
		idleTimeout:    viper.GetDuration(config.DeviceAuthClientIdleConnTimeoutKey),
		maxIdleConns:   viper.GetInt(config.DeviceAuthClientMaxIdleConnsKey),
		maxConns:       viper.GetInt(config.DeviceAuthClientMaxConnsKey),
		Port:           viper.GetInt(config.DeviceAuthPortKey),
		requestTimeout: viper.GetDuration(config.DeviceAuthClientRequestTimeoutKey),
		skipReadyCheck: viper.GetString(config.AppEnvKey) == "local" || !viper.GetBool(config.DeviceAuthClientEnabled),
	}
}

func NewDeviceAuthClient(cfg DeviceAuthClientCfg) *DeviceAuthClient {
	//nolint:exhaustruct
	client := &http.Client{
		Timeout: cfg.requestTimeout,
		Transport: &http.Transport{
			MaxIdleConns:        cfg.maxIdleConns,
			MaxIdleConnsPerHost: cfg.maxIdleConns,
			MaxConnsPerHost:     cfg.maxConns,
			IdleConnTimeout:     cfg.idleTimeout,
		},
	}

	return &DeviceAuthClient{cfg: cfg, Client: client}
}

func (dac *DeviceAuthClient) Close() error {
	dac.Client.CloseIdleConnections()

	return nil
}

func (dac *DeviceAuthClient) IsReady() error {
	if dac.cfg.skipReadyCheck {
		log.Trace().Msg("Skipping device auth ready check")

		return nil
	}

	log.Info().Msgf("Attempting to call device auth health check at %v:%v - %v",
		dac.cfg.Host, dac.cfg.Port, deviceAuthHealthCheckURL)

	url := fmt.Sprintf(deviceAuthHealthCheckURL, dac.cfg.Host, dac.cfg.Port)
	if err := dac.makeRequestWithContext(context.Background(), http.MethodGet, url, http.StatusOK); err != nil {
		return err
	}

	dac.cfg.skipReadyCheck = true

	log.Info().Msgf("successfully connected to %s", url)

	return nil
}

func (dac *DeviceAuthClient) EvictDevice(ctx context.Context, orgKey string, deviceID int) error {
	if orgKey == "" || deviceID <= 0 {
		return fmt.Errorf("%w: orgKey %s deviceID %d", ErrInvalidParamsProvided, orgKey, deviceID)
	}

	url := fmt.Sprintf(deviceAuthEvictDeviceURL, dac.cfg.Host, dac.cfg.Port, orgKey, deviceID)

	return dac.makeRequestWithContext(ctx, http.MethodDelete, url, http.StatusNoContent)
}

func (dac *DeviceAuthClient) makeRequestWithContext(
	ctx context.Context,
	method string,
	url string,
	expectedStatusCode int,
) error {
	var (
		err error
		req *http.Request
	)

	span, ctx := opentracing.StartSpanFromContext(ctx, "DeviceAuthHTTPCall")
	defer span.Finish()

	span.SetTag("http.method", method)
	span.SetTag("http.url", url)

	if req, err = http.NewRequestWithContext(ctx, method, url, nil); err != nil {
		return fmt.Errorf("failed to create request for method %s and url %s: %w", method, url, err)
	}

	var resp *http.Response

	//nolint:bodyclose // body is closed if it is not nil
	resp, err = dac.Client.Do(req)
	if resp != nil {
		defer closeRespBody(resp.Body)
	}

	if err != nil {
		span.SetTag("error", true)

		return fmt.Errorf("failed to connect to %s: %w", url, err)
	}

	if resp != nil {
		span.SetTag("http.code", resp.StatusCode)

		if resp.StatusCode != expectedStatusCode {
			return fmt.Errorf(
				"%w: expected %d received %d",
				ErrDeviceAuthUnexpectedStatusCode,
				expectedStatusCode,
				resp.StatusCode)
		}
	}

	return nil
}

func closeRespBody(body io.ReadCloser) {
	// to ensure proper connection re-use, the response body must be read to EOF and properly closed
	// https://pkg.go.dev/net/http#Client.Do
	if _, err := io.Copy(io.Discard, body); err != nil {
		log.Err(err).Msg("failed to read and discard response body")
	}

	if err := body.Close(); err != nil {
		log.Err(err).Msg("failed to close response body")
	}
}
