package client_test

import (
	"context"
	"errors"
	"fmt"
	"net/http"
	"net/http/httptest"
	"strings"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/client"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
)

var _ = Describe("device auth client", func() {
	var cfg client.DeviceAuthClientCfg
	var dac *client.DeviceAuthClient
	var server *httptest.Server

	Context("initializing", func() {
		BeforeEach(func() {
			viper.Set(config.DeviceAuthHostKey, "hello")
			viper.Set(config.DeviceAuthClientEnabled, true)
			viper.Set(config.DeviceAuthPortKey, 5004)
			viper.Set(config.AppEnvKey, "local")
			cfg = client.NewDeviceAuthClientCfg()
			Expect(cfg).To(Not(BeNil()))
			Expect(cfg.Host).To(Equal(viper.Get(config.DeviceAuthHostKey)))
			Expect(cfg.Port).To(Equal(viper.Get(config.DeviceAuthPortKey)))
			dac = client.NewDeviceAuthClient(cfg)
			Expect(dac).To(Not(BeNil()))
		})

		It("should return nil when in local mode", func() {
			err := dac.IsReady()
			Expect(err).To(BeNil())
		})
	})

	Context("client should do stuff with real server", func() {
		// Start a local HTTP server
		setupServer := func(expectedStatus int) {
			server = httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, req *http.Request) {
				// Send expected response
				w.WriteHeader(expectedStatus)
				_, err := w.Write([]byte(`response`))
				if err != nil {
					return
				}
			}))

			urlParts := strings.Split(server.URL, ":")
			viper.Set(config.DeviceAuthClientEnabled, true)
			viper.Set(config.DeviceAuthHostKey, fmt.Sprintf("%s:%s", urlParts[0], urlParts[1]))
			viper.Set(config.DeviceAuthPortKey, urlParts[2])
			viper.Set(config.AppEnvKey, "test")
			cfg = client.NewDeviceAuthClientCfg()
			dac = client.NewDeviceAuthClient(cfg)
			Expect(dac).To(Not(BeNil()))
			// substitute client for test client
			dac.Client = server.Client()
		}

		It("should respond with 200 when checking healthcheck", func() {
			setupServer(http.StatusOK)
			defer server.Close()
			err := dac.IsReady()
			Expect(err).To(BeNil())
		})

		It("should respond without an error when healthcheck responds with non 200 status", func() {
			setupServer(http.StatusInternalServerError)
			defer server.Close()
			err := dac.IsReady()
			Expect(err).To(Not(BeNil()))
		})

		It("should respond with an error when healthcheck responds with non 200 status", func() {
			setupServer(http.StatusInternalServerError)
			defer server.Close()
			err := dac.IsReady()
			Expect(err).To(Not(BeNil()))
		})

		It("should respond without error when evict device request is successful", func() {
			setupServer(http.StatusNoContent)
			defer server.Close()
			err := dac.EvictDevice(context.Background(), "test123", 10)
			Expect(err).To(BeNil())
		})

		It("should respond with error when evict device is called with invalid params", func() {
			err := dac.EvictDevice(context.Background(), "", 10)
			Expect(err).To(Not(BeNil()))
			Expect(errors.Is(err, client.ErrInvalidParamsProvided))
		})

		It("should respond with error when evict device request responds with non 200 status", func() {
			setupServer(http.StatusServiceUnavailable)
			defer server.Close()
			err := dac.EvictDevice(context.Background(), "test123", 10)
			Expect(err).To(Not(BeNil()))
		})
	})
})
