package config_test

import (
	"errors"
	"os"
	"strings"
	"testing"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
)

func TestConfig(t *testing.T) {
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "Config")
}

var _ = Describe("Config", func() {
	BeforeEach(func() {
		viper.Reset()
	})

	Context("Init method", func() {
		It("should set defaults", func() {
			Expect(len(viper.AllKeys())).To(BeZero())
			err := config.Init()
			Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())
			Expect(len(viper.AllKeys())).ToNot(BeZero())

			Expect(viper.GetInt(config.WFFlushIntervalSecsKey)).ToNot(BeZero())
			Expect(viper.GetString(config.LogFormatKey)).To(Equal(""))
			Expect(viper.GetString(config.LogLevelKey)).To(Equal("warn"))
		})

		Context("with app name set", func() {
			const appName = "test"

			It("should set app name from standard env key", func() {
				key := strings.ReplaceAll(strings.ToUpper(config.AppNameKey), ".", "_")
				os.Setenv(key, appName)
				err := config.Init()
				Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())
				Expect(viper.GetString(config.AppNameKey)).To(Equal(appName))
			})
		})

		Context("with dev mode set", func() {
			It("should set app name from standard env key", func() {
				key := strings.ReplaceAll(strings.ToUpper(config.DevModeKey), ".", "_")
				os.Setenv(key, "true")
				err := config.Init()
				Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())
				Expect(viper.GetBool(config.DevModeKey)).To(BeTrue())
				Expect(viper.GetString(config.LogFormatKey)).To(Equal("pretty"))
				Expect(viper.GetString(config.LogLevelKey)).To(Equal("info"))
				Expect(viper.GetBool(config.LogCallerKey)).To(BeTrue())
			})
		})

		Context("with good config file", func() {
			It("should succeed", func() {
				helper.GenerateGoodConfig()
				defer helper.CleanupConfig()
				err := config.Init()
				Expect(err).To(BeNil())
			})
		})

		Context("with bad config file", func() {
			It("should error", func() {
				helper.GenerateBadConfig()
				defer helper.CleanupConfig()
				err := config.Init()
				Expect(err).ToNot(BeNil())
				Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeFalse())
			})
		})
	})

	Context("accessors", func() {
		BeforeEach(func() {
			viper.Reset()
			err := config.Init()
			Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())
		})

		Context("IsRedactedHeader method", func() {
			It("should return true for headers that should be redacted", func() {
				Expect(config.IsRedactedHeader(config.RequestContextHeader)).To(BeTrue())
				Expect(config.IsRedactedHeader(config.UserContextHeader)).To(BeTrue())
				Expect(config.IsRedactedHeader(config.ConnectorAuthHeader)).To(BeTrue())
			})
		})

		Context("GetWFAppTags method", func() {
			BeforeEach(func() {
				viper.SetDefault(config.WFTagsClusterKey, "local")
				viper.SetDefault(config.WFTagsShardKey, "test")
			})

			It("should return app tags", func() {
				tags := config.GetWFAppTags()
				Expect(tags).NotTo(BeNil())
			})
		})

		Context("GetWFSamplerRate method", func() {
			Context("unset", func() {
				It("should return sampler rate of 0", func() {
					rate := config.GetWFSamplerRate()
					Expect(rate).To(BeZero())
				})
			})

			Context("under max", func() {
				BeforeEach(func() {
					viper.SetDefault(config.WFSamplerRateKey, 99)
				})

				It("should return sampler rate", func() {
					rate := config.GetWFSamplerRate()
					Expect(rate).To(Equal(uint64(99)))
				})
			})

			Context("over max", func() {
				BeforeEach(func() {
					viper.SetDefault(config.WFSamplerRateKey, 100)
				})

				It("should return sampler rate of 0", func() {
					rate := config.GetWFSamplerRate()
					Expect(rate).To(BeZero())
				})
			})
		})
	})
})
