package controller_test

import (
	"errors"
	"fmt"
	"net/http"
	"net/url"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/controller"
	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	"github.com/labstack/echo/v4"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog"
	"github.com/spf13/viper"
)

var (
	ErrFailure              = errors.New("failure")
	ErrMockedServiceFailure = errors.New("mocked service failure")
)

var _ = Describe("Router", func() {
	const prefix = "/test"
	var testEcho *echo.Echo
	var mockPerformer *mocks.Performer
	var mockReadinessChecker *mocks.ReadinessChecker

	BeforeSuite(func() {
		testEcho = echo.New()
		Expect(len(testEcho.Routes())).To(Equal(0))

		mockPerformer = mocks.NewPerformer(ctrl)
		mockReadinessChecker = mocks.NewReadinessChecker(ctrl)

		viper.Set(config.WebRoutesPrefixKey, prefix)
		controller.Route(testEcho, mockPerformer, mockReadinessChecker)
	})

	Context("route method", func() {
		It("should route", func() {
			Expect(len(testEcho.Routes())).To(Equal(2))

			helper.ExpectRoutes(testEcho.Routes(), []string{
				"GET", fmt.Sprintf("%s%s", prefix, config.HealthCheckPath),
				"POST", fmt.Sprintf("%s/v1", prefix),
			})
		})

		Context("v1 handler", func() {
			var mockResponseWriter *mocks.ResponseWriter
			var responseHeaders http.Header
			var request *http.Request
			var requestLogger *zerolog.Logger

			BeforeEach(func() {
				//nolint:exhaustruct // only used for testing, just need method and url
				request = &http.Request{
					Method: http.MethodPost,
					//nolint:exhaustruct // only used for testing, just need path
					URL: &url.URL{Path: "/test/v1"},
				}
				requestLogger = zerolog.Ctx(request.Context())
				responseHeaders = http.Header{}
				mockResponseWriter = mocks.NewResponseWriter(ctrl)
			})

			It("should succeed when service succeeds", func() {
				mockResponseWriter.EXPECT().WriteHeader(http.StatusAccepted).Times(1)
				mockPerformer.EXPECT().Perform(requestLogger).Times(1)
				testEcho.ServeHTTP(mockResponseWriter, request)
			})

			It("should fail when service fails", func() {
				mockResponseWriter.EXPECT().Header().Return(responseHeaders).Times(1)
				mockResponseWriter.EXPECT().WriteHeader(http.StatusInternalServerError).Times(1)
				mockResponseWriter.EXPECT().Write(gomock.Any()).Times(1)
				mockPerformer.EXPECT().Perform(requestLogger).Return(ErrMockedServiceFailure).Times(1)
				testEcho.ServeHTTP(mockResponseWriter, request)
			})
		})
	})
})
