package data

import (
	"context"
	"database/sql"
	"encoding/json"
	"fmt"
	"math/rand"
	"net/url"
	"regexp"

	"github.com/opentracing/opentracing-go"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"

	_ "github.com/lib/pq" // pg driver
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

var ErrNoDBConn = fmt.Errorf("unable to get db connection")

const (
	rowsUpdatedMetric = "db.rows.updated"
	maxTrafficRatio   = 100
)

type DBConfig struct {
	ClusterEndpoint        string
	ClusterReaderEndpoint  string
	DataBaseName           string
	Username               string
	Pass                   string
	Port                   string
	ConnectionString       string
	ConnectionStringReader string
}

type databaseConnections struct {
	Reader *sql.DB
	Writer *sql.DB
	Env    string
}

type DatabaseManager struct {
	Connections map[string]*databaseConnections
	config      *DatabaseManagerConfig
}

type DatabaseManagerConfig struct {
	dryRun                     bool
	simulateDBLoad             bool
	simulateDBLoadTrafficRatio int
}

type DB interface {
	FlushAndClose() error
	IsReady() error
	Update(context.Context, *model.Device) (int, error)
	MSMProcessDevice(context.Context, *model.Device) error
}

var _ DB = (*DatabaseManager)(nil)

func NewDBConfigs() map[string]DBConfig {
	connections := make(map[string]DBConfig)
	multiDBJSON := viper.GetString(config.DBHostMultiDBKey)

	if err := json.Unmarshal([]byte(multiDBJSON), &connections); multiDBJSON != "" && err != nil {
		log.Fatal().Err(err).Msgf("problem deserializing multi_db_json %v", multiDBJSON)
	}

	if len(connections) != 0 {
		log.Info().Msgf("Running in MultiDB Mode ... %v", multiDBJSON)

		return newMultiDBConfig(connections)
	}

	log.Info().Msgf("Running SingleDB Mode ... env is %v", viper.GetString(config.DBEnvKey))

	return newSingleDBConfig(connections)
}

func newDatabaseManagerConfig() *DatabaseManagerConfig {
	return &DatabaseManagerConfig{
		dryRun:                     viper.GetBool(config.AppDryRunKey),
		simulateDBLoad:             viper.GetBool(config.DBSimulateLoadEnabledKey),
		simulateDBLoadTrafficRatio: viper.GetInt(config.DBSimulateLoadTrafficRatioKey),
	}
}

func InitDB(dbToEnvConfig map[string]DBConfig) *DatabaseManager {
	dbConnections := make(map[string]*databaseConnections)

	for env, envConfig := range dbToEnvConfig {
		log.Info().Msgf("For env %v connecting to writer %v", env, envConfig.ClusterEndpoint)

		writerConnection, err := sql.Open(viper.GetString(config.DBDriverKey), envConfig.ConnectionString)
		if err != nil {
			log.Fatal().Msgf("error creating connection pool to db writer %v", maskPGCreds(err.Error()))
		}

		var readerConnection *sql.DB

		if envConfig.ConnectionStringReader != "" {
			log.Info().Msgf("For env %v connecting to reader %v", env, envConfig.ClusterReaderEndpoint)

			readerConnection, err = sql.Open(viper.GetString(config.DBDriverKey), envConfig.ConnectionStringReader)
			if err != nil {
				log.Fatal().Msgf("error creating connection pool to db reader %v", maskPGCreds(err.Error()))
			}
		}

		dbConnections[env] = &databaseConnections{
			Writer: writerConnection,
			Reader: readerConnection,
			Env:    env,
		}
	}

	dbm := &DatabaseManager{
		config:      newDatabaseManagerConfig(),
		Connections: dbConnections,
	}

	return dbm
}

func (db *DatabaseManager) IsReady() error {
	for env, connection := range db.Connections {
		if connection.Reader != nil {
			log.Trace().Str("env", env).Interface("stats", connection.Reader.Stats()).Msg("Attempting to ping DB reader")

			if err := connection.Reader.Ping(); err != nil {
				return fmt.Errorf("could not ping reader DB: %w", err)
			}
		}

		log.Trace().Str("env", env).Interface("stats", connection.Writer.Stats()).Msg("Attempting to ping DB writer")

		if err := connection.Writer.Ping(); err != nil {
			return fmt.Errorf("could not ping writer DB: %w", err)
		}

		log.Trace().Str("env", env).Msg("Successfully pinged DB(s)")
	}

	return nil
}

func (db *DatabaseManager) FlushAndClose() error {
	for _, connection := range db.Connections {
		err := connection.Writer.Close()
		if err != nil {
			return fmt.Errorf("%w", err)
		}

		if connection.Reader != nil {
			err = connection.Reader.Close()

			if err != nil {
				return fmt.Errorf("%w", err)
			}
		}
	}

	return nil
}

func (db *DatabaseManager) connection(env string) (*databaseConnections, bool) {
	if conn, ok := db.Connections[env]; ok {
		return conn, true
	}

	// failed to get requested env conn, default to first conn
	for _, conn := range db.Connections {
		log.Warn().
			Str("env", env).
			Str("conn", conn.Env).
			Msg("failed to get requested db connection corresponding to env, query will use the conn from conn env")

		return conn, false
	}

	return &databaseConnections{}, false
}

func (db *DatabaseManager) update(ctx context.Context, device *model.Device) (int, error) {
	span, ctx := opentracing.StartSpanFromContext(ctx, "DBDeviceUpdate")
	defer span.Finish()

	conn, connOK := db.connection(device.Env)
	if !connOK && conn.Env == "" {
		span.SetTag("error", true)

		return 0, ErrNoDBConn
	}

	result, err := conn.Writer.ExecContext(
		ctx,
		sqlUpdateDevice,
		device.ID,
		device.Type,
		device.Hostname,
		device.Email,
		device.DeploymentType.String(),
		device.OrgKey,
		device.PropsJSON,
	)
	if err != nil {
		log.Debug().Err(err).Msg("updated 0 rows")
		span.SetTag("error", true)

		return 0, fmt.Errorf("failed to exec sqlUpdateDevice: %w", err)
	}

	if updated, err := result.RowsAffected(); err != nil {
		log.Err(err).Msg("failed to get rows affected")
		log.Debug().Err(err).Msg("updated 0 rows")
		span.SetTag("error", true)
	} else {
		log.Debug().Msgf("updated %v rows", updated)
		if updated != 0 {
			telemetry.CounterInc(rowsUpdatedMetric, map[string]string{"env": device.Env}, updated)

			return device.ID, nil
		}
	}

	return 0, nil
}

func (db *DatabaseManager) Update(ctx context.Context, device *model.Device) (int, error) {
	if db.config.dryRun || device.DryRunMsg {
		log.Debug().
			Str("env", device.Env).
			Int("device-id", device.ID).
			Int("device-type", device.Type).
			Str("name", device.Hostname).
			Str("email", device.Email).
			Str("deployment-type", device.DeploymentType.String()).
			Str("props", device.PropsJSON).
			Msg("dry run device update")

		//nolint:gosec // rand lib is not being used for auth purposes (pw, keys, secrets, etc.)
		if db.config.simulateDBLoad && rand.Intn(maxTrafficRatio) < db.config.simulateDBLoadTrafficRatio {
			if err := db.findDeviceByID(ctx, device.ID, device.Env); err != nil {
				log.Err(err).
					Str("env", device.Env).
					Int("device-id", device.ID).
					Int("device-type", device.Type).
					Str("name", device.Hostname).
					Str("email", device.Email).
					Str("deployment-type", device.DeploymentType.String()).
					Str("props", device.PropsJSON).
					Msg("failed to find device")
			}
		}

		return 0, nil
	}

	return db.update(ctx, device)
}

// findDeviceByID is meant to only be used for testing load against the database while receiving DryRun messages.
func (db *DatabaseManager) findDeviceByID(ctx context.Context, deviceID int, env string) error {
	span, ctx := opentracing.StartSpanFromContext(ctx, "DBFindDeviceByID")
	defer span.Finish()

	conn, connOK := db.connection(env)
	if !connOK && conn.Env == "" {
		span.SetTag("error", true)

		return ErrNoDBConn
	}

	row := conn.Reader.QueryRowContext(ctx, fmt.Sprintf(sqlFindDevice, deviceID))

	// allocate slice of size 19 for number of columns in device row
	if err := row.Err(); err != nil {
		span.SetTag("error", true)

		return fmt.Errorf("failed to exec sqlFindDevice: %w", err)
	}

	return nil
}

func (db *DatabaseManager) MSMProcessDevice(
	ctx context.Context,
	device *model.Device,
) error {
	span, ctx := opentracing.StartSpanFromContext(ctx, "DBMSMProcessDevice")
	defer span.Finish()

	conn, connOK := db.connection(device.Env)
	if !connOK && conn.Env == "" {
		span.SetTag("error", true)

		return ErrNoDBConn
	}

	_, err := conn.Writer.ExecContext(
		ctx,
		sqlInsertDeviceToMSM,
		device.ID,
	)
	if err != nil {
		span.SetTag("error", true)

		return fmt.Errorf("failed to exec sqlInsertDeviceToMSM: %w", err)
	}

	return nil
}

func (db *DatabaseManager) Name() string {
	return "DB"
}

func newSingleDBConfig(connections map[string]DBConfig) map[string]DBConfig {
	var host string

	var hostReader string

	if host = viper.GetString(config.DBEndpointOverrideKey); host == "" {
		host = viper.GetString(config.DBHostKey)
		hostReader = viper.GetString(config.DBHostReaderKey)
	} else {
		hostReader = viper.GetString(config.DBEndpointOverrideKey)
	}

	port := viper.GetString(config.DBPortKey)
	username := viper.GetString(config.DBUserKey)
	pass := viper.GetString(config.DBPassKey)
	databasename := viper.GetString(config.DBDatabaseKey)
	sslMode := viper.GetString(config.DBSSLModeKey)

	connections[viper.GetString(config.DBEnvKey)] = DBConfig{
		Port:                   port,
		Username:               username,
		Pass:                   pass,
		ClusterEndpoint:        host,
		ClusterReaderEndpoint:  hostReader,
		DataBaseName:           databasename,
		ConnectionString:       connectionString(username, pass, host, port, databasename, sslMode),
		ConnectionStringReader: connectionString(username, pass, hostReader, port, databasename, sslMode),
	}

	return connections
}

func newMultiDBConfig(connections map[string]DBConfig) map[string]DBConfig {
	port := viper.GetString(config.DBPortKey)
	sslMode := viper.GetString(config.DBSSLModeKey)

	for env, envConfig := range connections {
		envConfig.Port = port
		envConfig.Username = getEnvPrefixedEnvVariable(env, config.DBUserKey)
		envConfig.Pass = getEnvPrefixedEnvVariable(env, config.DBPassKey)
		envConfig.ConnectionString = connectionString(
			getEnvPrefixedEnvVariable(env, config.DBUserKey),
			getEnvPrefixedEnvVariable(env, config.DBPassKey),
			envConfig.ClusterEndpoint, port, envConfig.DataBaseName, sslMode)
		envConfig.ConnectionStringReader = connectionString(
			getEnvPrefixedEnvVariable(env, config.DBUserKey),
			getEnvPrefixedEnvVariable(env, config.DBPassKey),
			envConfig.ClusterReaderEndpoint, port, envConfig.DataBaseName, sslMode)

		connections[env] = envConfig
	}

	return connections
}

func connectionString(username, password, host, port, databaseName, sslMode string) string {
	return fmt.Sprintf("postgresql://%v:%v@%v:%v/%v?sslmode=%v", // pragma: allowlist secret
		username, url.QueryEscape(password), host, port, databaseName, sslMode)
}

func getEnvPrefixedEnvVariable(env string, config string) string {
	if env != "" {
		return viper.GetString(fmt.Sprintf("%v.%v", env, config))
	}

	return viper.GetString(config)
}

func maskPGCreds(errorString string) string {
	maskRegex := regexp.MustCompile("postgresql://.+@")

	return maskRegex.ReplaceAllString(errorString, "postgresql://XXXXX:XXXXX@") // pragma: allowlist secret
}
