package data_test

import (
	"context"
	"errors"
	"fmt"
	"os"
	"strings"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"

	_ "github.com/lib/pq"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
	"github.com/testcontainers/testcontainers-go"
)

const connectionStringTemplate = "postgresql://%v:%v@%v:%v/%v?sslmode=%v" // pragma: allowlist secret

type logConsumer struct{}

func (*logConsumer) Accept(msg testcontainers.Log) {
	log.Debug().Str("type", msg.LogType).Msg(string(msg.Content))
}

var _ = Describe("DB Client", func() {
	const EAP01 = "EAP01_"
	const PROD01 = "PROD01_"
	const PROD02 = "PROD02_"
	const PROD05 = "PROD05_"

	Context("initializes ", func() {
		BeforeEach(func() {
			viper.Reset()
		})
		Context("Single DB Configuration", func() {
			var dbToEnvConfigExpected map[string]data.DBConfig

			goodDevice := &model.Device{
				DeploymentType: model.DeploymentTypeEndpoint,
				DryRunMsg:      false,
				Email:          "testEmail",
				Env:            "local",
				Hostname:       "testHost",
				ID:             1,
				OrgID:          1,
				OrgKey:         "R76KWHRN",
				PropsJSON:      "{\"testJSON\":\"value\"}",
				Type:           1,
			}

			BeforeEach(func() {
				const env = "local"
				const hostReader = ""
				host, _ := container.Host(pgCtx)
				const username = "user"
				const sslMode = "disable"
				const pass = "password"
				const databaseName = "database"
				port, _ := container.MappedPort(pgCtx, "5432")

				dbConfig := data.DBConfig{
					ClusterReaderEndpoint: hostReader,
					ClusterEndpoint:       host,
					Port:                  port.Port(),
					DataBaseName:          databaseName,
					Username:              username,
					Pass:                  pass,
					ConnectionString: fmt.Sprintf(connectionStringTemplate,
						username, pass, host, port.Port(), databaseName, sslMode),
					ConnectionStringReader: fmt.Sprintf(connectionStringTemplate,
						username, pass, hostReader, port.Port(), databaseName, sslMode),
				}

				dbToEnvConfigExpected = map[string]data.DBConfig{env: dbConfig}

				dbHostMultiDBConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBHostMultiDBKey), ".", "_")
				dbEnvConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBEnvKey), ".", "_")
				dbHostConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBHostKey), ".", "_")
				dbHostReaderConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBHostReaderKey), ".", "_")
				dbUserConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBUserKey), ".", "_")
				dbPasswordConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBPassKey), ".", "_")
				dbDatabaseConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBDatabaseKey), ".", "_")
				dbPortConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBPortKey), ".", "_")
				dbSSLModeConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBSSLModeKey), ".", "_")
				wfModeConfigKey := strings.ReplaceAll(strings.ToUpper(config.WFModeKey), ".", "_")

				os.Setenv(dbHostMultiDBConfigKey, "")
				os.Setenv(dbEnvConfigKey, env)
				os.Setenv(dbHostConfigKey, host)
				os.Setenv(dbHostReaderConfigKey, hostReader)
				os.Setenv(dbUserConfigKey, username)
				os.Setenv(dbPasswordConfigKey, pass)
				os.Setenv(dbDatabaseConfigKey, databaseName)
				os.Setenv(dbPortConfigKey, port.Port())
				os.Setenv(dbSSLModeConfigKey, "disable")
				os.Setenv(wfModeConfigKey, "disabled")

				err := config.Init()
				Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())

				telemetry.SetupMetrics(nil)
			})

			It("should correctly provide a configuration for single DB environment", func() {
				dbToEnvConfig := data.NewDBConfigs()
				Expect(dbToEnvConfigExpected).To(Equal(dbToEnvConfig))
			})

			It("should connect to the DB and check for readiness without error", func() {
				viper.SetDefault(config.AppDryRunKey, true)
				dbm := data.InitDB(data.NewDBConfigs())
				Expect(dbm.IsReady()).To(Succeed())
			})

			It("should connect to the DB and close without error", func() {
				viper.SetDefault(config.AppDryRunKey, true)
				dbm := data.InitDB(data.NewDBConfigs())
				Expect(dbm.FlushAndClose()).To(Succeed())
			})

			It("should not update the db with empty data", func() {
				dbm := data.InitDB(data.NewDBConfigs())
				updatedDevice, err := dbm.Update(context.Background(), &model.Device{})

				Expect(err).To(HaveOccurred())
				Expect(updatedDevice).To(Equal(0))
			})

			It("should not update the db with valid data, but dry run config", func() {
				viper.SetDefault(config.AppDryRunKey, true)
				dbm := data.InitDB(data.NewDBConfigs())
				updatedDevice, err := dbm.Update(context.Background(), goodDevice)
				Expect(err).ToNot(HaveOccurred())
				Expect(updatedDevice).To(Equal(0))
			})

			It("should not update the db with invalid device", func() {
				dbm := data.InitDB(data.NewDBConfigs())
				device := &model.Device{
					DeploymentType: model.DeploymentTypeEndpoint,
					DryRunMsg:      false,
					Email:          "testEmail",
					Env:            "local",
					Hostname:       "testHost",
					ID:             2,
					OrgID:          1,
					OrgKey:         "R76KWHRN",
					PropsJSON:      "{\"testJSON\":\"value\"}",
					Type:           1,
				}
				updatedDevice, err := dbm.Update(context.Background(), device)

				Expect(err).ToNot(HaveOccurred())
				Expect(updatedDevice).To(Equal(0))
			})

			It("should update the db with valid data", func() {
				dbm := data.InitDB(data.NewDBConfigs())
				updatedDevice, err := dbm.Update(context.Background(), goodDevice)
				Expect(err).ToNot(HaveOccurred())
				Expect(updatedDevice).To(Equal(1))
			})

			It("should update the db with missing orgkey, use first db conn", func() {
				dbm := data.InitDB(data.NewDBConfigs())
				device := &model.Device{
					DeploymentType: model.DeploymentTypeEndpoint,
					DryRunMsg:      false,
					Email:          "testEmail",
					Env:            "local",
					Hostname:       "testHost",
					ID:             1,
					OrgID:          1,
					OrgKey:         "",
					PropsJSON:      "{\"testJSON\":\"value\"}",
					Type:           1,
				}
				updatedDevice, err := dbm.Update(context.Background(), device)

				Expect(err).ToNot(HaveOccurred())
				Expect(updatedDevice).To(Equal(1))
			})

			It("should process msm device", func() {
				dbm := data.InitDB(data.NewDBConfigs())
				device := &model.Device{
					DeploymentType: model.DeploymentTypeEndpoint,
					DryRunMsg:      false,
					Email:          "testEmail",
					Env:            "local",
					Hostname:       "testHost",
					ID:             1,
					OrgID:          1,
					OrgKey:         "",
					PropsJSON:      "{\"testJSON\":\"value\"}",
					Type:           1,
				}
				err := dbm.MSMProcessDevice(context.Background(), device)

				Expect(err).ToNot(HaveOccurred())
			})
		})

		Context("Multi DB Configuration", func() {
			var dbToEnvConfigExpected map[string]data.DBConfig

			BeforeEach(func() {
				const eap01User = "eap01.user"
				const eap01Pass = "eap01.password"
				const prod01User = "prod01.user"
				const prod01Pass = "prod01.password"
				const prod02User = "prod02.user"
				const prod02Pass = "prod02.password"
				const prod05User = "prod05.user"
				const prod05Pass = "prod05.password"
				const sslMode = "disable"
				const port = "5432"

				dbToEnvConfigExpected = map[string]data.DBConfig{
					"eap01": {
						Username:              eap01User,
						Pass:                  eap01Pass,
						ClusterEndpoint:       "localhost.eap01",
						ClusterReaderEndpoint: "localhostreader.eap01",
						DataBaseName:          "pscdbaurora.eap01",
						Port:                  port,
						ConnectionString: fmt.Sprintf(connectionStringTemplate,
							eap01User, eap01Pass, "localhost.eap01", port, "pscdbaurora.eap01", sslMode),
						ConnectionStringReader: fmt.Sprintf(connectionStringTemplate,
							eap01User, eap01Pass, "localhostreader.eap01", port, "pscdbaurora.eap01", sslMode),
					},
					"prod01": {
						Username:              prod01User,
						Pass:                  prod01Pass,
						ClusterEndpoint:       "localhost.prod01",
						ClusterReaderEndpoint: "localhostreader.prod01",
						DataBaseName:          "pscdbaurora.prod01",
						Port:                  port,
						ConnectionString: fmt.Sprintf(connectionStringTemplate,
							prod01User, prod01Pass, "localhost.prod01", port, "pscdbaurora.prod01", sslMode),
						ConnectionStringReader: fmt.Sprintf(connectionStringTemplate,
							prod01User, prod01Pass, "localhostreader.prod01", port, "pscdbaurora.prod01", sslMode),
					},
					"prod02": {
						Username:              prod02User,
						Pass:                  prod02Pass,
						ClusterEndpoint:       "localhost.prod02",
						ClusterReaderEndpoint: "localhostreader.prod02",
						DataBaseName:          "pscdbaurora.prod02",
						Port:                  port,
						ConnectionString: fmt.Sprintf(connectionStringTemplate,
							prod02User, prod02Pass, "localhost.prod02", port, "pscdbaurora.prod02", sslMode),
						ConnectionStringReader: fmt.Sprintf(connectionStringTemplate,
							prod02User, prod02Pass, "localhostreader.prod02", port, "pscdbaurora.prod02", sslMode),
					},
					"prod05": {
						Username:              prod05User,
						Pass:                  prod05Pass,
						ClusterEndpoint:       "localhost.prod05",
						ClusterReaderEndpoint: "localhostreader.prod05",
						DataBaseName:          "pscdbaurora.prod05",
						Port:                  port,
						ConnectionString: fmt.Sprintf(connectionStringTemplate,
							prod05User, prod05Pass, "localhost.prod05", port, "pscdbaurora.prod05", sslMode),
						ConnectionStringReader: fmt.Sprintf(connectionStringTemplate,
							prod05User, prod05Pass, "localhostreader.prod05", port, "pscdbaurora.prod05", sslMode),
					},
				}

				eap01UserConfigKey := strings.ReplaceAll(strings.ToUpper(EAP01+config.DBUserKey), ".", "_")
				eap01PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(EAP01+config.DBPassKey), ".", "_")
				prod01UserConfigKey := strings.ReplaceAll(strings.ToUpper(PROD01+config.DBUserKey), ".", "_")
				prod01PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(PROD01+config.DBPassKey), ".", "_")
				prod02UserConfigKey := strings.ReplaceAll(strings.ToUpper(PROD02+config.DBUserKey), ".", "_")
				prod02PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(PROD02+config.DBPassKey), ".", "_")
				prod05UserConfigKey := strings.ReplaceAll(strings.ToUpper(PROD05+config.DBUserKey), ".", "_")
				prod05PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(PROD05+config.DBPassKey), ".", "_")
				dbPortConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBPortKey), ".", "_")
				dbSSLModeConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBSSLModeKey), ".", "_")

				os.Setenv(eap01UserConfigKey, eap01User)
				os.Setenv(eap01PasswordConfigKey, eap01Pass)
				os.Setenv(prod01UserConfigKey, prod01User)
				os.Setenv(prod01PasswordConfigKey, prod01Pass)
				os.Setenv(prod02UserConfigKey, prod02User)
				os.Setenv(prod02PasswordConfigKey, prod02Pass)
				os.Setenv(prod05UserConfigKey, prod05User)
				os.Setenv(prod05PasswordConfigKey, prod05Pass)
				os.Setenv(dbPortConfigKey, port)
				os.Setenv(dbSSLModeConfigKey, sslMode)
			})

			It("should correctly provide a configuration for multi DB environment", func() {
				dbHostMultiDB := `{
						"eap01": {
							"clusterEndpoint": "localhost.eap01",
							"clusterReaderEndpoint": "localhostreader.eap01",
							"databaseName": "pscdbaurora.eap01"
						},
						"prod01": {
							"clusterEndpoint": "localhost.prod01",
							"clusterReaderEndpoint": "localhostreader.prod01",
							"databaseName": "pscdbaurora.prod01"
						},
						"prod02": {
							"clusterEndpoint": "localhost.prod02",
							"clusterReaderEndpoint": "localhostreader.prod02",
							"databaseName": "pscdbaurora.prod02"
						},
						"prod05": {
							"clusterEndpoint": "localhost.prod05",
							"clusterReaderEndpoint": "localhostreader.prod05",
							"databaseName": "pscdbaurora.prod05"
						}
					}`

				err := config.Init()
				Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())

				dbHostMultiDBConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBHostMultiDBKey), ".", "_")
				os.Setenv(dbHostMultiDBConfigKey, dbHostMultiDB)
				dbToEnvConfig := data.NewDBConfigs()
				Expect(dbToEnvConfigExpected).To(Equal(dbToEnvConfig))
			})

			It("should connect to all DBs in configuration and check for readiness without error", func() {
				const eap01User = "user"
				const eap01Pass = "password"
				const prod01User = "user"
				const prod01Pass = "password"
				const prod02User = "user"
				const prod02Pass = "password"
				const prod05User = "user"
				const prod05Pass = "password"
				const sslMode = "disable"
				host, _ := container.Host(pgCtx)
				port, _ := container.MappedPort(pgCtx, "5432")

				dbHostMultiDBTemplate := `{
						"eap01": {
							"clusterEndpoint": "%v",
							"clusterReaderEndpoint": "%v",
							"databaseName": "database"
						},
						"prod01": {
							"clusterEndpoint": "%v",
							"clusterReaderEndpoint": "%v",
							"databaseName": "database"
						},
						"prod02": {
							"clusterEndpoint": "%v",
							"clusterReaderEndpoint": "%v",
							"databaseName": "database"
						},
						"prod05": {
							"clusterEndpoint": "%v",
							"clusterReaderEndpoint": "%v",
							"databaseName": "database"
						}
					}`

				dbHostMultiDB := fmt.Sprintf(dbHostMultiDBTemplate, host, host, host, host, host, host, host, host)

				dbHostMultiDBConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBHostMultiDBKey), ".", "_")
				eap01UserConfigKey := strings.ReplaceAll(strings.ToUpper(EAP01+config.DBUserKey), ".", "_")
				eap01PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(EAP01+config.DBPassKey), ".", "_")
				prod01UserConfigKey := strings.ReplaceAll(strings.ToUpper(PROD01+config.DBUserKey), ".", "_")
				prod01PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(PROD01+config.DBPassKey), ".", "_")
				prod02UserConfigKey := strings.ReplaceAll(strings.ToUpper(PROD02+config.DBUserKey), ".", "_")
				prod02PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(PROD02+config.DBPassKey), ".", "_")
				prod05UserConfigKey := strings.ReplaceAll(strings.ToUpper(PROD05+config.DBUserKey), ".", "_")
				prod05PasswordConfigKey := strings.ReplaceAll(strings.ToUpper(PROD05+config.DBPassKey), ".", "_")
				dbPortConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBPortKey), ".", "_")
				dbSSLModeConfigKey := strings.ReplaceAll(strings.ToUpper(config.DBSSLModeKey), ".", "_")

				os.Setenv(eap01UserConfigKey, eap01User)
				os.Setenv(eap01PasswordConfigKey, eap01Pass)
				os.Setenv(prod01UserConfigKey, prod01User)
				os.Setenv(prod01PasswordConfigKey, prod01Pass)
				os.Setenv(prod02UserConfigKey, prod02User)
				os.Setenv(prod02PasswordConfigKey, prod02Pass)
				os.Setenv(prod05UserConfigKey, prod05User)
				os.Setenv(prod05PasswordConfigKey, prod05Pass)
				os.Setenv(dbPortConfigKey, port.Port())
				os.Setenv(dbSSLModeConfigKey, sslMode)
				os.Setenv(dbHostMultiDBConfigKey, dbHostMultiDB)

				err := config.Init()
				Expect(errors.As(err, &viper.ConfigFileNotFoundError{})).To(BeTrue())

				viper.SetDefault(config.AppDryRunKey, true)
				dbm := data.InitDB(data.NewDBConfigs())
				Expect(dbm.Connections).To(HaveLen(4))
				for _, connection := range dbm.Connections {
					Expect(connection.Writer).ToNot(BeNil())
					Expect(connection.Reader).ToNot(BeNil())
				}

				Expect(dbm.IsReady()).Should(Succeed())
			})
		})
	})
})
