package data

import (
	"context"
	"fmt"
	"strings"

	"carbonblack.com/protocols/psc/ems/server/emdstatus"
	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/aws/retry"
	"github.com/aws/aws-sdk-go-v2/feature/dynamodb/attributevalue"
	"github.com/aws/aws-sdk-go-v2/service/dynamodb"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
)

const (
	endpointURLFlag = "--endpoint-url="
)

type DBClient interface {
	Ready(tableName string) error
	Update(ctx context.Context, tableName string, data *emdstatus.DeviceStatusMsg) error
}

// DynamoAPIClient interface was created to make mocking calls to dynamodb api easier.
type DynamoAPIClient interface {
	DescribeTable(context.Context,
		*dynamodb.DescribeTableInput,
		...func(*dynamodb.Options),
	) (*dynamodb.DescribeTableOutput, error)
	PutItem(context.Context,
		*dynamodb.PutItemInput,
		...func(*dynamodb.Options),
	) (*dynamodb.PutItemOutput, error)
}

// DynamoDBClient is a struct wrapping a dynamo api client interface.
type DynamoDBClient struct {
	Client DynamoAPIClient
}

// NewDynamoDBClient returns a DynamoDBClient struct implementing the DBClient interface.
func NewDynamoDBClient() *DynamoDBClient {
	apiClient := dynamodb.NewFromConfig(getAWSConfig(), func(opts *dynamodb.Options) {
		if viper.GetString(config.DynamoDBEndpointKey) != "" {
			endpoint := strings.ReplaceAll(viper.GetString(config.DynamoDBEndpointKey), endpointURLFlag, "")
			opts.EndpointResolver = dynamodb.EndpointResolverFromURL(endpoint)
		}

		opts.Retryer = retry.NewStandard(func(reOpts *retry.StandardOptions) {
			reOpts.MaxAttempts = viper.GetInt(config.DynamoDBMaxAttemptsKey)
			reOpts.MaxBackoff = viper.GetDuration(config.DynamoDBMaxBackoffKey)
		})
	})

	return &DynamoDBClient{
		Client: apiClient,
	}
}

// Ready checks that the dynamodb client can successfully connect to the dynamodb table.
func (d *DynamoDBClient) Ready(tableName string) error {
	ctx, cancel := context.WithTimeout(context.TODO(), viper.GetDuration(config.DynamoDBRequestTimeoutKey))
	defer cancel()

	// Setting max retries to 0 so the request is retried until it succeeds or non-retryable error is returned.
	_, err := d.Client.DescribeTable(ctx, &dynamodb.DescribeTableInput{TableName: aws.String(tableName)},
		func(opts *dynamodb.Options) {
			opts.Retryer = retry.AddWithMaxAttempts(retry.NewStandard(), 0)
		},
	)
	if err != nil {
		return fmt.Errorf("error describing dynamodb %s table: %w", tableName, err)
	}

	log.Info().Msgf("dynamodb client successfully connected to %s table", tableName)

	return nil
}

func (d *DynamoDBClient) Update(ctx context.Context, tableName string, data *emdstatus.DeviceStatusMsg) error {
	item, err := attributevalue.MarshalMap(data.Metadata)
	if err != nil {
		return fmt.Errorf("error marshaling metadata into dynamo attribute values: %w", err)
	}

	//nolint:exhaustruct // all fields within PutItemInput are not necessary.
	input := &dynamodb.PutItemInput{
		TableName: aws.String(tableName),
		Item:      item,
	}

	_, err = d.Client.PutItem(ctx, input)
	if err != nil {
		return fmt.Errorf("error putting data in dynamodb table %s: %w", tableName, err)
	}

	return nil
}
