package data_test

import (
	"context"
	"errors"

	"carbonblack.com/protocols/psc/ems/server/emdstatus"
	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/feature/dynamodb/attributevalue"
	"github.com/aws/aws-sdk-go-v2/service/dynamodb"
	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"
)

var errDynamo = errors.New("something happened in the dynamo")

//nolint:exhaustruct
var _ = Describe("DynamoDB", func() {
	var dynamoDB *data.DynamoDBClient
	var mockDynamoAPIClient *mocks.DynamoAPIClient
	tableName := "testTable"

	Context("initialized", func() {
		BeforeEach(func() {
			mockDynamoAPIClient = mocks.NewDynamoAPIClient(ctrl)
			dynamoDB = data.NewDynamoDBClient()
			dynamoDB.Client = mockDynamoAPIClient
			Expect(dynamoDB).ToNot(BeNil())
			Expect(dynamoDB.Client).ToNot(BeNil())
		})

		Context("Ready method", func() {
			input := &dynamodb.DescribeTableInput{TableName: aws.String(tableName)}
			output := &dynamodb.DescribeTableOutput{}

			It("should return nil when describeTable operation is successful", func() {
				mockDynamoAPIClient.EXPECT().DescribeTable(gomock.Any(), input, gomock.Any()).Return(output, nil)
				err := dynamoDB.Ready(tableName)
				Expect(err).To(BeNil())
			})

			It("should return error when describeTable operation fails", func() {
				mockDynamoAPIClient.EXPECT().DescribeTable(gomock.Any(), input, gomock.Any()).Return(nil, errDynamo)
				err := dynamoDB.Ready(tableName)
				Expect(err).ToNot(BeNil())
			})
		})

		Context("Update method", func() {
			partitionKey1 := "RegistrationID"
			partitionKey2 := "DeviceGUID"
			sortKey := "LastUpdateTime"
			data := &emdstatus.DeviceStatusMsg{
				Metadata: map[string]string{partitionKey1: "val1", partitionKey2: "val2", sortKey: "val3"},
			}
			item, err := attributevalue.MarshalMap(data.Metadata)
			input := &dynamodb.PutItemInput{TableName: aws.String(tableName), Item: item}
			output := &dynamodb.PutItemOutput{}
			Expect(err).To(BeNil())

			It("should return nil when put operation is successful", func() {
				mockDynamoAPIClient.EXPECT().PutItem(gomock.Any(), input).Return(output, nil)
				err := dynamoDB.Update(context.TODO(), tableName, data)
				Expect(err).To(BeNil())
			})

			It("should return error when put operation fails", func() {
				mockDynamoAPIClient.EXPECT().PutItem(gomock.Any(), input).Return(nil, errDynamo)
				err := dynamoDB.Update(context.TODO(), tableName, data)
				Expect(err).ToNot(BeNil())
			})
		})
	})

	Context("local deployment", func() {
		It("should create a dynamodb client", func() {
			viper.SetDefault(config.DynamoDBEndpointKey, "someEndPoint")
			dynamoDB = data.NewDynamoDBClient()
			Expect(dynamoDB).ToNot(BeNil())
		})
	})

	Context("aws deployment", func() {
		It("should create a dynamodb client", func() {
			viper.SetDefault(config.DynamoDBEndpointKey, "")
			dynamoDB = data.NewDynamoDBClient()
			Expect(dynamoDB).ToNot(BeNil())
		})
	})
})
