package data

import (
	"encoding/json"
	"fmt"
	"io"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"

	"github.com/elastic/go-elasticsearch/v5"
	"github.com/elastic/go-elasticsearch/v5/esapi"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

var (
	ErrClusterStatus   = fmt.Errorf("unhealthy cluster status")
	ErrClusterResponse = fmt.Errorf("bad response from cluster")
)

type ClusterInfo struct {
	Version struct {
		Number string `json:"number"`
	} `json:"version"`
}

type ClusterHealth struct {
	Status string `json:"status"`
}

type ESConfig struct {
	URL      string
	Env      string
	Username string
	Password string
}

type esMultiConfig map[string]struct {
	ClusterDomainName string `json:"clusterDomainName"`
	ClusterEndpoint   string `json:"clusterEndpoint"`
}

type ESClientManager struct {
	Clients map[string]*elasticsearch.Client
	ready   bool
}

func decodeESResponse(response *esapi.Response, err error, result any) error {
	if response != nil {
		defer response.Body.Close()
	}

	if err != nil {
		return err
	}

	if response.IsError() {
		return fmt.Errorf("%w status code: %v", ErrClusterResponse, response.StatusCode)
	}

	if err := json.NewDecoder(response.Body).Decode(result); err != nil {
		_, _ = io.Copy(io.Discard, response.Body) // Ensure the body is completely read before closed

		return fmt.Errorf("error parsing ES response body: %w", err)
	}

	return nil
}

func (m *ESClientManager) clusterHealth(esClient *elasticsearch.Client) error {
	clusterHealthResponse := &ClusterHealth{}

	res, err := esClient.Cluster.Health()

	if err = decodeESResponse(res, err, clusterHealthResponse); err != nil {
		return err
	}

	log.Info().Msgf("ES cluster status is %v", clusterHealthResponse.Status)

	if clusterHealthResponse.Status != "green" {
		return fmt.Errorf("%w: ES Cluster status is %v", ErrClusterStatus, clusterHealthResponse.Status)
	}

	return nil
}

func (m *ESClientManager) clusterInfo(esClient *elasticsearch.Client) error {
	clusterInfoResponse := &ClusterInfo{}

	res, err := esClient.Info()

	if err = decodeESResponse(res, err, clusterInfoResponse); err != nil {
		return err
	}

	log.Info().Msgf("ES cluster version is %v", clusterInfoResponse.Version.Number)

	return nil
}

func (m *ESClientManager) IsReady() error {
	if m.ready {
		return nil
	}

	for env, esClient := range m.Clients {
		log.Info().Msgf("Checking if cluster %v is ready", env)

		if err := m.clusterInfo(esClient); err != nil {
			log.Err(err).Msgf("Failed cluster info call")

			return err
		}

		if err := m.clusterHealth(esClient); err != nil {
			log.Err(err).Msgf("Failed cluster health call")

			return err
		}
	}

	m.ready = true

	return nil
}

func NewESClientManager(configs ...*ESConfig) *ESClientManager {
	clients := make(map[string]*elasticsearch.Client)

	for _, esConfig := range configs {
		log.Info().Msgf("Creating ES Client Manager for env %v with URL: %v", esConfig.Env, esConfig.URL)

		client, err := elasticsearch.NewClient(elasticsearch.Config{
			Addresses: []string{esConfig.URL},
			Username:  esConfig.Username,
			Password:  esConfig.Password, // pragma: allowlist secret
		})
		if err != nil {
			log.Err(err).Msgf("failed to create ES client for %v", esConfig.Env)
			panic("failed to create ES client")
		}

		clients[esConfig.Env] = client
	}

	return &ESClientManager{clients, false}
}

func FormatHostURL(endpoint, env string) string {
	insecureEnvs := viper.GetStringSlice(config.ElasticsearchInsecureEnvsKey)
	for _, insecureEnv := range insecureEnvs {
		if insecureEnv == env {
			return fmt.Sprintf("http://%v:9200", endpoint)
		}
	}

	return fmt.Sprintf("https://%v:443", endpoint)
}

func NewESConfigs() []*ESConfig {
	configs := NewESConfigsFromJSON()

	if len(configs) > 0 {
		return configs
	}

	return []*ESConfig{
		NewSingleESConfig(),
	}
}

func NewESConfigsFromJSON() []*ESConfig {
	var esConfigs []*ESConfig

	// Handle JSON blob for multi-environments and ES clusters
	esConfigJSON := viper.GetString(config.ElasticsearchMultiConfigKey)

	if esConfigJSON != "" {
		multiConfig := &esMultiConfig{}
		if err := json.Unmarshal([]byte(esConfigJSON), multiConfig); err != nil {
			panic(fmt.Sprintf("bad JSON for %v: %v", config.ElasticsearchMultiConfigKey, err))
		}

		log.Debug().Msgf("ES Config JSON: %v", multiConfig)

		for env, esConfig := range *multiConfig {
			esConfigs = append(esConfigs, &ESConfig{
				URL:      FormatHostURL(esConfig.ClusterEndpoint, viper.GetString(config.AppEnvKey)),
				Env:      env,
				Username: viper.GetString(config.ElasticsearchUsernameKey), // I think there's a single username/pw for all?
				Password: viper.GetString(config.ElasticsearchPasswordKey), // I think there's a single username/pw for all?
			})
		}
	}

	return esConfigs
}

func NewSingleESConfig() *ESConfig {
	return &ESConfig{
		URL:      FormatHostURL(viper.GetString(config.ElasticsearchEndpointKey), viper.GetString(config.AppEnvKey)),
		Env:      viper.GetString(config.AppEnvKey),
		Username: viper.GetString(config.ElasticsearchUsernameKey),
		Password: viper.GetString(config.ElasticsearchPasswordKey),
	}
}
