package data_test

import (
	"fmt"
	"io"
	"strings"

	"github.com/elastic/go-elasticsearch/v5"
	"github.com/elastic/go-elasticsearch/v5/esapi"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
)

var ErrES = fmt.Errorf("nope")

var _ = Describe("Elasticsearch", func() {
	Context("initialized", func() {
		BeforeEach(func() {
			viper.Reset()
		})

		Context("NewESConfigsFromJSON", func() {
			It("should return a config only if the property is set", func() {
				viper.Set(config.ElasticsearchMultiConfigKey,
					"{\"bar\": {\"clusterEndpoint\": \"foo\", \"clusterDomainName\": \"beepbeep\"}}")
				viper.Set(config.ElasticsearchUsernameKey, "boo")
				viper.Set(config.ElasticsearchPasswordKey, "far")

				esConfig := data.NewESConfigsFromJSON()
				Expect(esConfig).ToNot(BeNil())
				Expect(esConfig).To(HaveLen(1))
				Expect(esConfig).To(ContainElement(HaveField("URL", "https://foo:443")))
				Expect(esConfig).To(ContainElement(HaveField("Env", "bar")))
				Expect(esConfig).To(ContainElement(HaveField("Username", "boo")))
				Expect(esConfig).To(ContainElement(HaveField("Password", "far")))
			})

			It("should not return a config if the property is not set", func() {
				esConfig := data.NewESConfigsFromJSON()
				Expect(esConfig).To(HaveLen(0))
			})
		})

		Context("NewSingleESConfig", func() {
			It("should always return a config", func() {
				viper.Set(config.ElasticsearchEndpointKey, "foo")
				viper.Set(config.AppEnvKey, "bar")
				viper.Set(config.ElasticsearchUsernameKey, "boo")
				viper.Set(config.ElasticsearchPasswordKey, "far")

				esConfig := data.NewSingleESConfig()
				Expect(esConfig).ToNot(BeNil())
				Expect(esConfig).To(HaveField("URL", "https://foo:443"))
				Expect(esConfig).To(HaveField("Env", "bar"))
				Expect(esConfig).To(HaveField("Username", "boo"))
				Expect(esConfig).To(HaveField("Password", "far"))
			})
		})

		Context("NewESConfigs", func() {
			It("should return at least one ESConfig when no multiconfig property is set", func() {
				viper.Set(config.ElasticsearchMultiConfigKey, "")
				viper.Set(config.AppEnvKey, "local")
				configs := data.NewESConfigs()
				Expect(configs).To(HaveLen(1))
				Expect(configs).To(ContainElement(HaveField("Env", "local")))
			})

			It("should return multiple ESConfigs when the multiconfig property is set", func() {
				viper.Set(config.ElasticsearchMultiConfigKey, "{"+
					"\"env1\": {\"clusterEndpoint\": \"foo\"}, "+
					"\"env2\": {\"clusterEndpoint\": \"bar\"}"+
					"}")
				configs := data.NewESConfigs()
				Expect(configs).To(HaveLen(2))

				Expect(configs).To(ContainElements(HaveField("Env", "env1"), HaveField("Env", "env2")))
				Expect(configs).To(ContainElements(HaveField("URL", "https://foo:443"), HaveField("URL", "https://bar:443")))
			})

			It("fails with malformed JSON", func() {
				viper.Set(config.ElasticsearchMultiConfigKey,
					"{\"bar\": [{\"clusterEndpoint\": \"foo\", \"clusterDomainName\": \"beepbeep\"}]}")
				viper.Set(config.ElasticsearchUsernameKey, "boo")
				viper.Set(config.ElasticsearchPasswordKey, "far")

				Expect(func() { data.NewESConfigsFromJSON() }).To(Panic())
			})
		})

		Context("IsReady", func() {
			const client1 = "unittest1"
			const client2 = "unittest2"

			var mockMan *data.ESClientManager
			infoCallCount := 0
			healthCallCount := 0

			BeforeEach(func() {
				mockMan = &data.ESClientManager{Clients: map[string]*elasticsearch.Client{
					client1: {
						API: &esapi.API{
							Info: func(o ...func(*esapi.InfoRequest)) (*esapi.Response, error) {
								infoCallCount++

								return &esapi.Response{
									StatusCode: 200,
									Header:     nil,
									Body: io.NopCloser(strings.NewReader(
										"{\"version\": {\"number\": \"6.8.0\"}}")),
								}, nil
							},
							Cluster: &esapi.Cluster{
								Health: func(o ...func(*esapi.ClusterHealthRequest)) (*esapi.Response, error) {
									healthCallCount++

									return &esapi.Response{
										StatusCode: 200,
										Header:     nil,
										Body:       io.NopCloser(strings.NewReader("{\"status\": \"green\"}")),
									}, nil
								},
							},
						},
					},
					client2: {
						API: &esapi.API{
							Info: func(o ...func(*esapi.InfoRequest)) (*esapi.Response, error) {
								infoCallCount++

								return &esapi.Response{
									StatusCode: 200,
									Header:     nil,
									Body: io.NopCloser(strings.NewReader(
										"{\"version\": {\"number\": \"6.8.0\"}}")),
								}, nil
							},
							Cluster: &esapi.Cluster{
								Health: func(o ...func(*esapi.ClusterHealthRequest)) (*esapi.Response, error) {
									healthCallCount++

									return &esapi.Response{
										StatusCode: 200,
										Header:     nil,
										Body:       io.NopCloser(strings.NewReader("{\"status\": \"green\"}")),
									}, nil
								},
							},
						},
					},
				}}
			})

			It("should return nil when ready, should cache response", func() {
				err := mockMan.IsReady()
				Expect(err).ShouldNot(HaveOccurred())

				err = mockMan.IsReady()
				Expect(err).ShouldNot(HaveOccurred())

				Expect(infoCallCount).To(Equal(2))
				Expect(healthCallCount).To(Equal(2))
			})

			It("should return an error when the cluster status is red", func() {
				res := func(o ...func(*esapi.ClusterHealthRequest)) (*esapi.Response, error) {
					return &esapi.Response{
						StatusCode: 200,
						Header:     nil,
						Body:       io.NopCloser(strings.NewReader("{\"status\": \"red\"}")),
					}, nil
				}
				mockMan.Clients[client2].Cluster.Health = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})

			It("should return an error when the ES info response status code is bad", func() {
				res := func(o ...func(*esapi.InfoRequest)) (*esapi.Response, error) {
					return &esapi.Response{
						StatusCode: 503,
						Header:     nil,
						Body:       io.NopCloser(strings.NewReader("")),
					}, nil
				}
				mockMan.Clients[client1].Info = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})

			It("should return an error when the ES health response status code is bad", func() {
				res := func(o ...func(*esapi.ClusterHealthRequest)) (*esapi.Response, error) {
					return &esapi.Response{
						StatusCode: 503,
						Header:     nil,
						Body:       io.NopCloser(strings.NewReader("")),
					}, nil
				}
				mockMan.Clients[client1].Cluster.Health = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})

			It("should return an error when the ES info response status json is bad", func() {
				res := func(o ...func(*esapi.InfoRequest)) (*esapi.Response, error) {
					return &esapi.Response{
						StatusCode: 200,
						Header:     nil,
						Body:       io.NopCloser(strings.NewReader("nope")),
					}, nil
				}
				mockMan.Clients[client1].Info = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})

			It("should return an error when the ES health response status json is bad", func() {
				res := func(o ...func(*esapi.ClusterHealthRequest)) (*esapi.Response, error) {
					return &esapi.Response{
						StatusCode: 200,
						Header:     nil,
						Body:       io.NopCloser(strings.NewReader("nope")),
					}, nil
				}
				mockMan.Clients[client1].Cluster.Health = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})

			It("should return an error when the ES info response is an error", func() {
				res := func(o ...func(*esapi.InfoRequest)) (*esapi.Response, error) {
					return nil, ErrES
				}
				mockMan.Clients[client1].Info = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})

			It("should return an error when the ES health response is an error", func() {
				res := func(o ...func(*esapi.ClusterHealthRequest)) (*esapi.Response, error) {
					return nil, ErrES
				}
				mockMan.Clients[client1].Cluster.Health = res

				err := mockMan.IsReady()
				Expect(err).Should(HaveOccurred())
			})
		})

		Context("FormatHostURL", func() {
			BeforeEach(func() {
				viper.Set(config.ElasticsearchInsecureEnvsKey, "local prod02 prod05")
			})

			It("should return an http URL when env is insecure", func() {
				url := data.FormatHostURL("foo.bar", "local")
				Expect(url).To(HavePrefix("http://"))
				Expect(url).To(HaveSuffix(":9200"))
			})

			It("should return an https URL when env is not insecure", func() {
				url := data.FormatHostURL("foo.bar", "dev")
				Expect(url).To(HavePrefix("https://"))
				Expect(url).To(HaveSuffix(":443"))
			})
		})

		Context("NewEsClientManager", func() {
			It("Creates a client from each config", func() {
				viper.Set(config.ElasticsearchMultiConfigKey, "{"+
					"\"env1\": {\"clusterEndpoint\": \"foo\"}, "+
					"\"env2\": {\"clusterEndpoint\": \"bar\"}"+
					"}")
				configs := data.NewESConfigs()
				clientManager := data.NewESClientManager(configs...)

				Expect(clientManager).ToNot(BeNil())
				Expect(clientManager.Clients).To(HaveLen(2))
				Expect(clientManager.Clients).To(HaveKeyWithValue("env1", Not(BeNil())))
				Expect(clientManager.Clients).To(HaveKeyWithValue("env2", Not(BeNil())))
				Expect(clientManager.Clients).ToNot(HaveKey("local"))
			})

			It("panics on new client error", func() {
				viper.Set(config.ElasticsearchMultiConfigKey, "{"+
					// 7F is delete character, gets past json parsing, but will fail url parsing
					"\"env1\": {\"clusterEndpoint\": \""+string([]byte{0x7f})+"\"} "+
					"}")
				configs := data.NewESConfigs()

				Expect(func() { data.NewESClientManager(configs...) }).To(Panic())
			})
		})
	})
})
