package data

import (
	"context"
	"fmt"
	"strings"
	"time"

	"github.com/aws/aws-sdk-go-v2/aws"
	"github.com/aws/aws-sdk-go-v2/aws/retry"
	"github.com/aws/aws-sdk-go-v2/service/kinesis"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
)

type Producer interface {
	PutRecord(context.Context, StreamMessager) error
}

type StreamMessager interface {
	Marshal() ([]byte, error)
	PartitionKey() (string, error)
}

type KinesisProducer interface {
	DescribeStream(
		context.Context,
		*kinesis.DescribeStreamInput,
		...func(*kinesis.Options),
	) (*kinesis.DescribeStreamOutput, error)

	PutRecord(
		context.Context,
		*kinesis.PutRecordInput,
		...func(*kinesis.Options),
	) (*kinesis.PutRecordOutput, error)
}

type ProducerClientConfig struct {
	Stream         string
	EndpointURL    string
	MaxAttempts    int
	MaxBackoff     time.Duration
	RequestTimeout time.Duration
	Client         KinesisProducer // Client is an optional override
}

type ProducerClient struct {
	client              KinesisProducer
	cfg                 *ProducerClientConfig
	connectionValidated bool
}

func NewProducerClientConfig(streamKey string) *ProducerClientConfig {
	return &ProducerClientConfig{
		Stream:         viper.GetString(streamKey),
		EndpointURL:    strings.ReplaceAll(viper.GetString(config.KinesisEndpointKey), "--endpoint-url=", ""),
		MaxAttempts:    viper.GetInt(config.KinesisMaxAttemptsKey),
		MaxBackoff:     viper.GetDuration(config.KinesisMaxBackoffKey),
		RequestTimeout: viper.GetDuration(config.KinesisRequestTimeoutKey),
		Client:         nil,
	}
}

func NewProducerClient(cfg *ProducerClientConfig) *ProducerClient {
	if cfg.Client != nil {
		return &ProducerClient{cfg.Client, cfg, false}
	}

	client := kinesis.NewFromConfig(getAWSConfig(), generateClientConfigurer(cfg))

	return &ProducerClient{client, cfg, false}
}

func generateClientConfigurer(cfg *ProducerClientConfig) func(opts *kinesis.Options) {
	return func(opts *kinesis.Options) {
		if cfg.EndpointURL != "" {
			opts.EndpointResolver = kinesis.EndpointResolverFromURL(cfg.EndpointURL)
		}

		opts.Retryer = retry.NewStandard(func(reOpts *retry.StandardOptions) {
			reOpts.MaxAttempts = cfg.MaxAttempts
			reOpts.MaxBackoff = cfg.MaxBackoff
		})
	}
}

// IsReady returns nil if client connection is successful.
func (p *ProducerClient) IsReady() error {
	if p.connectionValidated {
		return nil
	}

	ctx, cancel := context.WithTimeout(context.TODO(), p.cfg.RequestTimeout)
	defer cancel()

	out, err := p.client.DescribeStream(ctx, &kinesis.DescribeStreamInput{
		StreamName: &p.cfg.Stream,

		ExclusiveStartShardId: nil,
		Limit:                 nil,
		StreamARN:             nil,
	})
	if err != nil {
		log.Err(err).Str("stream", p.cfg.Stream).Msg("failed to connect to kinesis stream")

		return nil
	}

	log.Info().Str("stream", p.cfg.Stream).Msg("connected to kinesis")
	log.Debug().Str("stream", p.cfg.Stream).Interface("DescribeStreamOutput", out).Msg("kinesis output received")

	p.connectionValidated = true

	return nil
}

func (p *ProducerClient) PutRecord(ctx context.Context, msg StreamMessager) error {
	ctx, cancel := context.WithTimeout(ctx, p.cfg.RequestTimeout)
	defer cancel()

	partitionKey, err := msg.PartitionKey()
	if err != nil {
		return fmt.Errorf("%w: failed to get partition key", err)
	}

	data, err := msg.Marshal()
	if err != nil {
		return fmt.Errorf("%w: failed to get message data", err)
	}

	out, err := p.client.PutRecord(ctx, &kinesis.PutRecordInput{
		StreamName:   &p.cfg.Stream,
		PartitionKey: aws.String(partitionKey),
		Data:         data,

		ExplicitHashKey:           nil,
		SequenceNumberForOrdering: nil,
		StreamARN:                 nil,
	})

	log.Debug().Err(err).Interface("PutRecordOutput", out).Msg("got response from kinesis")

	if err != nil {
		return fmt.Errorf("%w: failed to put record on stream", err)
	}

	return nil
}
