package data_test

import (
	"context"
	"errors"
	"time"

	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"

	"github.com/aws/aws-sdk-go-v2/service/kinesis"
	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/service"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"
)

var errKinesis = errors.New("kinesis can't even")

var _ = Describe("KinesisProducer", func() {
	Context("NewProducerClient", func() {
		It("should return a new producer client", func() {
			client := data.NewProducerClient(&data.ProducerClientConfig{}) //nolint:exhaustruct // testing here
			Expect(client).NotTo(BeNil())
		})

		Context("with endpoint", func() {
			It("should return a new producer client", func() {
				//nolint:exhaustruct // testing here
				client := data.NewProducerClient(&data.ProducerClientConfig{EndpointURL: "http://a.url"})
				Expect(client).NotTo(BeNil())
			})
		})
	})

	Context("with mocked kinesis client", func() {
		var mockKinesisProducer *mocks.KinesisProducer
		var producerClient data.Producer
		streamName := "test-stream"

		BeforeEach(func() {
			mockKinesisProducer = mocks.NewKinesisProducer(ctrl)
			//nolint:exhaustruct // testing here
			producerClient = data.NewProducerClient(&data.ProducerClientConfig{Stream: streamName, Client: mockKinesisProducer})

			Expect(producerClient).NotTo(BeNil())
		})

		Context("IsReady", func() {
			var mockWriter *mocks.Writer
			var readinessClient service.ReadinessChecker
			var describeStreamCall *gomock.Call
			var delay time.Duration
			var ogLogger zerolog.Logger

			BeforeEach(func() {
				mockWriter = mocks.NewWriter(ctrl)
				ogLogger = log.Logger
				log.Logger = zerolog.New(mockWriter).Level(zerolog.InfoLevel)

				var ok bool
				readinessClient, ok = producerClient.(service.ReadinessChecker)
				Expect(ok).To(BeTrue())
				Expect(readinessClient).NotTo(BeNil())

				describeStreamCall = mockKinesisProducer.EXPECT().DescribeStream(
					gomock.Any(),
					gomock.Any(),
				).Do(func(
					ctx context.Context,
					input *kinesis.DescribeStreamInput,
					optFuncs ...func(*kinesis.Options),
				) (*kinesis.DescribeStreamOutput, error) {
					Expect(input.StreamName).To(Equal(&streamName))
					time.Sleep(delay)

					return &kinesis.DescribeStreamOutput{}, nil //nolint:exhaustruct // testing here
				})
			})

			AfterEach(func() {
				log.Logger = ogLogger
			})

			It("should log error if unable to connect to Kinesis", func() {
				var errMsgLog string

				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&errMsgLog),
				).Times(1)

				describeStreamCall.Return(nil, errKinesis)

				err := readinessClient.IsReady()

				Expect(err).NotTo(HaveOccurred())
				Expect(errMsgLog).To(ContainSubstring("failed to connect to kinesis"))
			})

			XIt("should timeout", func() {
				var errMsgLog string

				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&errMsgLog),
				).Times(1)

				delay = 2 * time.Second

				err := readinessClient.IsReady()

				Expect(err).NotTo(HaveOccurred())
				Expect(errMsgLog).To(ContainSubstring("\"message\":\"connected to kinesis\""))
			})

			It("should return nil, cache response", func() {
				var errMsgLog string

				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&errMsgLog),
				).Times(1)

				describeStreamCall.Times(1)

				err := readinessClient.IsReady()

				Expect(err).NotTo(HaveOccurred())
				Expect(errMsgLog).To(ContainSubstring("\"message\":\"connected to kinesis\""))

				err = readinessClient.IsReady()
				Expect(err).NotTo(HaveOccurred())
			})
		})

		Context("PutRecord", func() {
			var mockMessage *mocks.StreamMessager

			BeforeEach(func() {
				mockMessage = mocks.NewStreamMessager(ctrl)
			})

			It("should return error if unable to get partition key", func() {
				mockMessage.EXPECT().PartitionKey().Return("", errKinesis)

				err := producerClient.PutRecord(context.Background(), mockMessage)
				Expect(err).To(HaveOccurred())
			})

			It("should return error if unable to get data", func() {
				mockMessage.EXPECT().PartitionKey()
				mockMessage.EXPECT().Marshal().Return([]byte{}, errKinesis)

				err := producerClient.PutRecord(context.Background(), mockMessage)
				Expect(err).To(HaveOccurred())
			})

			It("should return error if PutRecord fails", func() {
				mockMessage.EXPECT().PartitionKey()
				mockMessage.EXPECT().Marshal()
				mockKinesisProducer.EXPECT().PutRecord(gomock.Any(), gomock.Any()).Return(nil, errKinesis)

				err := producerClient.PutRecord(context.Background(), mockMessage)
				Expect(err).To(HaveOccurred())
			})

			XIt("should timeout", func() {
				mockMessage.EXPECT().PartitionKey()
				mockMessage.EXPECT().Marshal()
				mockKinesisProducer.EXPECT().PutRecord(gomock.Any(), gomock.Any()).Do(func(
					ctx context.Context,
					input *kinesis.PutRecordInput,
					optFuncs ...func(*kinesis.Options),
				) (*kinesis.PutRecordOutput, error) {
					time.Sleep(2 * time.Second)

					return nil, errKinesis
				})

				err := producerClient.PutRecord(context.Background(), mockMessage)
				Expect(err).NotTo(HaveOccurred())
			})

			It("should succeed if well-formed", func() {
				mockMessage.EXPECT().PartitionKey()
				mockMessage.EXPECT().Marshal()
				mockKinesisProducer.EXPECT().PutRecord(gomock.Any(), gomock.Any())

				err := producerClient.PutRecord(context.Background(), mockMessage)
				Expect(err).NotTo(HaveOccurred())
			})
		})
	})
})
