package data

import (
	"context"
	"errors"
	"fmt"
	"strings"

	"carbonblack.com/protocols/psc/ems/server/emdstatus"
	"github.com/rs/zerolog/log"
)

var errDeviceStatusMsgMissingKeys = errors.New("device status message missing required key")

type Tabler interface {
	Update(ctx context.Context, data *emdstatus.DeviceStatusMsg) error
}

// Table is a struct wrapping a DBClient interface. The client is used to interact with data in the underlying table.
type Table struct {
	dbClient DBClient
	specs    *TableSpecs
	name     string
}

// NewTable returns a Table struct implementing the Tabler interface.
func NewTable(dbClient DBClient, specs *TableSpecs) (*Table, error) {
	table := Table{
		dbClient: dbClient,
		specs:    specs,
		name:     specs.TableName,
	}

	if err := table.IsReady(); err != nil {
		log.Error().Err(err).Msgf("failed to init %s table", table.name)

		return nil, fmt.Errorf("table init: %w", err)
	}

	log.Info().Msgf("%s table successfully init", table.name)

	return &table, nil
}

// IsReady checks that the table can successfully connect to the underlying table.
func (t *Table) IsReady() error {
	if err := t.dbClient.Ready(t.name); err != nil {
		return fmt.Errorf("table isready error: %w", err)
	}

	return nil
}

func (t *Table) Update(ctx context.Context, data *emdstatus.DeviceStatusMsg) error {
	if err := validateMessage(data, t.specs); err != nil {
		return err
	}

	if err := t.dbClient.Update(ctx, t.name, data); err != nil {
		return fmt.Errorf("table update error: %w", err)
	}

	return nil
}

// validateMessage validates that the given device status message contains the required keys.
// After message is validated the composite hash key is set on the message.
func validateMessage(data *emdstatus.DeviceStatusMsg, specs *TableSpecs) error {
	sortKeyValues := make([]string, 0, len(specs.sortKeys))

	for _, key := range specs.sortKeys {
		if data.Metadata[key] == "" {
			return errDeviceStatusMsgMissingKeys
		}

		sortKeyValues = append(sortKeyValues, data.Metadata[key])
	}

	partitionKeyValues := make([]string, 0, len(specs.partitionKeys))

	for _, key := range specs.partitionKeys {
		if data.Metadata[key] == "" {
			return errDeviceStatusMsgMissingKeys
		}

		partitionKeyValues = append(partitionKeyValues, data.Metadata[key])
	}

	data.Metadata[sortKey] = strings.Join(sortKeyValues, specs.compositeKeySeparator)
	data.Metadata[partitionKey] = strings.Join(partitionKeyValues, specs.compositeKeySeparator)

	return nil
}
