package grpcserver

import (
	"context"
	"fmt"
	"net"
	"strings"

	"carbonblack.com/protocols/psc/ems/server/deviceevent"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/client"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/grpcserver/grpcmiddleware"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/service"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	"gitlab.bit9.local/psc/platform/ems/device-service/proto/grpc/deviceservice"

	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
	"google.golang.org/grpc"
	metadata2 "google.golang.org/grpc/metadata"
)

var ErrBadRequest = fmt.Errorf("bad request")

type DeviceServiceServer struct {
	deviceservice.UnimplementedDeviceServiceServer
	server   *grpc.Server
	database data.DB
	evictor  client.Evictor
	des      service.DeviceEventsPublisher
	cbd      service.Cbd
}

func NewDeviceServiceServer(
	database data.DB,
	evictor client.Evictor,
	des service.DeviceEventsPublisher,
	cbd service.Cbd,
) *DeviceServiceServer {
	s := grpc.NewServer(grpc.UnaryInterceptor(grpcmiddleware.UnaryInterceptor))
	dss := &DeviceServiceServer{
		server: s, database: database, evictor: evictor, des: des, cbd: cbd,
	}

	deviceservice.RegisterDeviceServiceServer(s, dss)

	return dss
}

func (dss *DeviceServiceServer) Start(shutdown <-chan struct{}) {
	port := viper.GetUint16(config.GrpcPortKey)
	log.Info().Msgf("starting grpc server on [::]:%d", port)

	var lis net.Listener

	var err error

	if lis, err = net.Listen("tcp", fmt.Sprintf(":%d", port)); err != nil {
		log.Fatal().Msgf("failed to create listener: %v", err)
	}

	if err = dss.server.Serve(lis); err != nil {
		log.Fatal().Msgf("failed to start grpc sever: %v", err)
	}

	<-shutdown
}

func (dss *DeviceServiceServer) Shutdown(context.Context) error {
	log.Info().Msg("shutting down grpc server gracefully")
	dss.server.GracefulStop()

	return nil
}

//nolint:cyclop
func (dss *DeviceServiceServer) Update(
	ctx context.Context,
	request *deviceservice.DeviceUpdateRequest,
) (*deviceservice.DeviceUpdateResponse, error) {
	logger := zerolog.Ctx(ctx)
	logger.Info().Msgf("received update request %v", request)

	var spanCtx opentracing.SpanContext

	if incomingMetadata, ok := metadata2.FromIncomingContext(ctx); ok {
		logger.Debug().Msgf("grpc metadata from incoming context %v", incomingMetadata)
		spanCtx = telemetry.SpanContextFromGRPC(incomingMetadata)

		// Add traceID etc to logger context
		logger.UpdateContext(func(c zerolog.Context) zerolog.Context {
			updatedCtx := c
			for key, values := range incomingMetadata {
				value := strings.Join(values, ",")
				updatedCtx = c.Str(key, value)
			}

			return updatedCtx
		})
	}

	updateDeviceSpan, ctx := opentracing.StartSpanFromContext(ctx, "UpdateDevice", ext.RPCServerOption(spanCtx))
	defer updateDeviceSpan.Finish()

	metadata, err := validateRequest(request)
	if err != nil {
		return nil, fmt.Errorf("error parsing req, error=%w, req=%v", err, request)
	}

	device, err := model.NewDevice(metadata)
	if err != nil {
		return nil, fmt.Errorf("error parsing device message, error=%w, metadata=%v", err, metadata)
	}

	if err := dss.cbd.Update(ctx, metadata); err != nil {
		return nil, fmt.Errorf("error sending update to cbd client, error=%w, device-msg", err)
	}

	if updatedDeviceID, err := dss.database.Update(ctx, device); err != nil {
		updateDeviceSpan.SetTag("error", true)

		return nil, fmt.Errorf("failed to update db: %w", err)
	} else if updatedDeviceID > 0 {
		if err := dss.evictor.EvictDevice(ctx, metadata[model.DataFieldOrgKey], updatedDeviceID); err != nil {
			logger.Err(err).Int("device-id", updatedDeviceID).Msg("failed to evict device")
		}

		if err := dss.publishDeviceEventsRecord(ctx, device); err != nil {
			logger.Err(err).Msg("publishDeviceEventsRecord:")
		}

		if err := dss.database.MSMProcessDevice(ctx, device); err != nil {
			logger.Err(err).Int("device-id", updatedDeviceID).Msg("failed to add device to MSM processing table")
		}
	}

	return &deviceservice.DeviceUpdateResponse{}, nil
}

func validateRequest(request *deviceservice.DeviceUpdateRequest) (map[string]string, error) {
	msg := request.GetMsg()
	if msg == nil {
		return nil, fmt.Errorf("%w: no status message provided", ErrBadRequest)
	}

	metadata := msg.GetMetadata()
	if metadata == nil {
		return nil, fmt.Errorf("%w: no metadata provided", ErrBadRequest)
	}

	return metadata, nil
}

func (dss *DeviceServiceServer) publishDeviceEventsRecord(ctx context.Context, device *model.Device) error {
	logger := zerolog.Ctx(ctx)
	span, ctx := opentracing.StartSpanFromContext(ctx, "PublishDeviceEventsRecord")

	defer span.Finish()

	if isEligibleForDeviceNotification(device) {
		logger.Debug().Msgf("sending device event notification for device %d", device.ID)

		if err := dss.des.PublishDeviceEventsRecord(ctx, device, deviceevent.EmsChangeEvent_UPDATE); err != nil {
			span.SetTag("error", true)

			return fmt.Errorf("error while publishing device event record to kinesis: %w", err)
		}
	}

	return nil
}

// Only to back support Win 3.8 & Linux 2.13, we will check for deploymentType.
// For all future sensor, we will determine this based on infrastructureProvider & vdiProvider.
// To exclude sending events for HCoA, we will check if the VM is supported by Azure infra and VdiProvider is HORIZON.
func isEligibleForDeviceNotification(device *model.Device) bool {
	return device.DeploymentType == model.DeploymentTypeAWS ||
		device.InfrastructureProvider == model.InfrastructureProviderAWS ||
		device.InfrastructureProvider == model.InfrastructureProviderGCP ||
		(device.InfrastructureProvider == model.InfrastructureProviderAzure &&
			device.VDIProvider != model.VDIProviderHorizon)
}
