package grpcserver_test

import (
	"context"
	"testing"

	"carbonblack.com/protocols/psc/ems/server/deviceevent"
	"carbonblack.com/protocols/psc/ems/server/emdstatus"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/grpcserver"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/proto/grpc/deviceservice"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

//nolint:gochecknoglobals // singleton test controller
var ctrl *gomock.Controller

func TestDeviceService(t *testing.T) {
	ctrl = gomock.NewController(t)
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "DeviceService")
}

var _ = Describe("DeviceService", func() {
	var (
		ctx                  context.Context
		dss                  *grpcserver.DeviceServiceServer
		mockDB               *mocks.DB
		mockDevEventsService *mocks.DeviceEventsPublisher
		mockEvictor          *mocks.Evictor
		mockCbdService       *mocks.Cbd
	)

	data := map[string]string{
		model.DataFieldRegistrationID:         "1-1",
		model.DataFieldDeviceType:             "1",
		model.DataFieldOSMajorVersion:         "2320",
		model.DataFieldOSMinorVersion:         "5",
		model.DataFieldMaintenanceRelease:     "5",
		model.DataFieldOSBuildNumber:          "5",
		model.DataFieldPlatformType:           "2",
		model.DataFieldOrgKey:                 "R76KWHRN",
		model.DataFieldInfrastructureProvider: "AWS",
		model.DataFieldHostname:               "testHost",
		model.DataFieldProductVersion:         "1",
		model.DataFieldOriginalRequestHeaders: "",
		model.DataFieldOriginalRequestMsgPack: "",
		model.DataFieldMessageType:            "8370",
	}

	Context("Update", func() {
		BeforeEach(func() {
			ctx = context.Background()
			mockDB = mocks.NewDB(ctrl)
			mockDevEventsService = mocks.NewDeviceEventsPublisher(ctrl)
			mockEvictor = mocks.NewEvictor(ctrl)
			mockCbdService = mocks.NewCbd(ctrl)
			dss = grpcserver.NewDeviceServiceServer(mockDB, mockEvictor, mockDevEventsService, mockCbdService)
		})

		It("should notify device event when there's been an update to the database and device is eligible", func() {
			req := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: data}}
			mockDB.EXPECT().Update(gomock.Any(), gomock.Any()).Return(1, nil)
			mockDB.EXPECT().MSMProcessDevice(gomock.Any(), gomock.Any())
			mockEvictor.EXPECT().EvictDevice(gomock.Any(), data[model.DataFieldOrgKey], 1).Return(nil)
			mockCbdService.EXPECT().Update(gomock.Any(), gomock.Any()).Return(nil)
			mockDevEventsService.EXPECT().PublishDeviceEventsRecord(gomock.Any(), gomock.Any(),
				deviceevent.EmsChangeEvent_UPDATE).Return(nil)

			_, err := dss.Update(ctx, req)
			Expect(err).To(BeNil())
		})

		It("should not notify device event and evict service when there isn't an update to the database", func() {
			req := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: data}}
			mockDB.EXPECT().Update(gomock.Any(), gomock.Any()).Return(0, nil)
			mockEvictor.EXPECT().EvictDevice(gomock.Any(), gomock.Any(), gomock.Any()).Times(0)
			mockCbdService.EXPECT().Update(gomock.Any(), gomock.Any()).Return(nil)
			mockDevEventsService.EXPECT().PublishDeviceEventsRecord(gomock.Any(), gomock.Any(), gomock.Any()).Times(0)

			_, _ = dss.Update(ctx, req)
		})

		It("should return error when validating request without a message", func() {
			req := &deviceservice.DeviceUpdateRequest{Msg: nil}
			_, err := dss.Update(ctx, req)
			Expect(err).ToNot(BeNil())
		})

		It("should return error when validating a request without metadata", func() {
			req := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: nil}}
			_, err := dss.Update(ctx, req)
			Expect(err).ToNot(BeNil())
		})

		It("should return error when parsing empty device data", func() {
			req := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: map[string]string{}}}
			_, err := dss.Update(ctx, req)
			Expect(err).ToNot(BeNil())
		})

		It("should notify device event service when infra is Azure and VDI is not Horizon", func() {
			data := map[string]string{
				model.DataFieldRegistrationID:         "1-1",
				model.DataFieldDeviceType:             "1",
				model.DataFieldOSMajorVersion:         "2320",
				model.DataFieldOSMinorVersion:         "5",
				model.DataFieldMaintenanceRelease:     "5",
				model.DataFieldOSBuildNumber:          "5",
				model.DataFieldPlatformType:           "2",
				model.DataFieldOrgKey:                 "R76KWHRN",
				model.DataFieldInfrastructureProvider: "Azure",
				model.DataFieldVDIProvider:            "CITRIX",
				model.DataFieldHostname:               "testHost",
				model.DataFieldProductVersion:         "1",
				model.DataFieldOriginalRequestHeaders: "",
				model.DataFieldOriginalRequestMsgPack: "",
				model.DataFieldMessageType:            "8370",
			}

			msg := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: data}}
			mockDB.EXPECT().Update(gomock.Any(), gomock.Any()).Return(1, nil)
			mockDB.EXPECT().MSMProcessDevice(gomock.Any(), gomock.Any())
			mockEvictor.EXPECT().EvictDevice(gomock.Any(), data[model.DataFieldOrgKey], 1).Return(nil)
			mockCbdService.EXPECT().Update(gomock.Any(), gomock.Any()).Return(nil)
			mockDevEventsService.EXPECT().PublishDeviceEventsRecord(gomock.Any(), gomock.Any(),
				deviceevent.EmsChangeEvent_UPDATE).Return(nil)

			_, err := dss.Update(ctx, msg)
			Expect(err).To(BeNil())
		})

		It("notify device event service when deployment type has changed and infra provider is GCP", func() {
			data := map[string]string{
				model.DataFieldRegistrationID:         "1-1",
				model.DataFieldDeviceType:             "1",
				model.DataFieldOSMajorVersion:         "2320",
				model.DataFieldOSMinorVersion:         "5",
				model.DataFieldMaintenanceRelease:     "5",
				model.DataFieldOSBuildNumber:          "5",
				model.DataFieldPlatformType:           "2",
				model.DataFieldOrgKey:                 "R76KWHRN",
				model.DataFieldInfrastructureProvider: "GCP",
				model.DataFieldHostname:               "testHost",
				model.DataFieldProductVersion:         "1",
				model.DataFieldOriginalRequestHeaders: "",
				model.DataFieldOriginalRequestMsgPack: "",
				model.DataFieldMessageType:            "8193",
			}

			msg := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: data}}
			mockDB.EXPECT().Update(gomock.Any(), gomock.Any()).Return(1, nil)
			mockDB.EXPECT().MSMProcessDevice(gomock.Any(), gomock.Any())
			mockEvictor.EXPECT().EvictDevice(gomock.Any(), data[model.DataFieldOrgKey], 1).Return(nil)
			mockCbdService.EXPECT().Update(gomock.Any(), gomock.Any()).Return(nil)
			mockDevEventsService.EXPECT().PublishDeviceEventsRecord(gomock.Any(), gomock.Any(),
				deviceevent.EmsChangeEvent_UPDATE).Return(nil)

			_, err := dss.Update(ctx, msg)
			Expect(err).To(BeNil())
		})

		It("should not notify device event service when device is not eligible", func() {
			data := map[string]string{
				model.DataFieldRegistrationID:         "1-1",
				model.DataFieldDeviceType:             "1",
				model.DataFieldOSMajorVersion:         "2320",
				model.DataFieldOSMinorVersion:         "5",
				model.DataFieldMaintenanceRelease:     "5",
				model.DataFieldOSBuildNumber:          "5",
				model.DataFieldPlatformType:           "2",
				model.DataFieldOrgKey:                 "R76KWHRN",
				model.DataFieldInfrastructureProvider: "NONE",
				model.DataFieldHostname:               "testHost",
				model.DataFieldProductVersion:         "1",
				model.DataFieldOriginalRequestHeaders: "",
				model.DataFieldOriginalRequestMsgPack: "",
				model.DataFieldMessageType:            "8193",
			}

			msg := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: data}}
			mockDB.EXPECT().Update(gomock.Any(), gomock.Any()).Return(1, nil)
			mockDB.EXPECT().MSMProcessDevice(gomock.Any(), gomock.Any())
			mockEvictor.EXPECT().EvictDevice(gomock.Any(), data[model.DataFieldOrgKey], 1).Return(nil)
			mockCbdService.EXPECT().Update(gomock.Any(), gomock.Any()).Return(nil)
			mockDevEventsService.EXPECT().PublishDeviceEventsRecord(gomock.Any(), gomock.Any(),
				deviceevent.EmsChangeEvent_UPDATE).Times(0)

			_, err := dss.Update(ctx, msg)
			Expect(err).To(BeNil())
		})
	})
})
