package grpcmiddleware

import (
	"context"
	"strconv"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	"gitlab.bit9.local/psc/platform/ems/device-service/proto/grpc/deviceservice"

	"github.com/google/uuid"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"google.golang.org/grpc"
	"google.golang.org/grpc/metadata"
	"google.golang.org/grpc/status"
	"google.golang.org/protobuf/proto"
)

type contextKey string

const contextRequestID contextKey = "request-id"

func UnaryInterceptor(
	ctx context.Context,
	req interface{},
	info *grpc.UnaryServerInfo,
	handler grpc.UnaryHandler,
) (interface{}, error) {
	startTime := time.Now()

	// create new request ID until status msg has trace ID attached
	requestID := uuid.New().String()
	ctx, logger := attachRequestLogger(ctx, requestID)

	resp, err := handler(ctx, req)

	latency := time.Since(startTime)
	respStatus := int(status.Code(err))

	dims := map[string]string{
		"method": info.FullMethod,
		"status": strconv.Itoa(respStatus),
		"error":  strconv.FormatBool(err != nil),
	}

	telemetry.CounterInc("grpc", dims, 1)
	telemetry.HistogramUpdateTimeSince("grpc.timer", dims, startTime)

	var logEvent *zerolog.Event
	if respStatus != 0 {
		logEvent = logger.Error()
	} else {
		logEvent = logger.Debug()
	}

	logEvent.Err(err).
		Str("method", info.FullMethod).
		Str("user-agent", getRequestMetadata(ctx, "User-Agent")).
		Int("status", respStatus).
		Str("x-forwarded-for", getRequestMetadata(ctx, "X-Forwarded-For")).
		Dur("latency(ms)", latency).
		Int("response size", getResponseSize(resp)).
		Msg("request")

	return resp, err
}

func attachRequestLogger(ctx context.Context, requestID string) (context.Context, zerolog.Logger) {
	logger := *zerolog.Ctx(ctx) // try to get an existing logger from the context

	// if the existing request logger is disabled, there may not be a logger set on the request
	if logger.GetLevel() == zerolog.Disabled {
		// use a child of the global logger
		logger = log.With().Logger()
	}

	// Attach the request ID to the gRPC request context
	logger = logger.With().Str(string(contextRequestID), requestID).Logger()
	ctx = context.WithValue(ctx, contextRequestID, requestID)
	ctx = logger.WithContext(ctx)

	return ctx, logger
}

func getRequestMetadata(ctx context.Context, meta string) string {
	var (
		data metadata.MD
		ok   bool
	)

	if data, ok = metadata.FromIncomingContext(ctx); !ok {
		return ""
	}

	vals := data.Get(meta)
	if len(vals) > 0 {
		return vals[0]
	}

	return ""
}

func getResponseSize(resp interface{}) int {
	if resp != nil {
		if message, ok := resp.(*deviceservice.DeviceUpdateResponse); ok {
			return proto.Size(message)
		}
	}

	return 0
}
