package grpcmiddleware_test

import (
	"context"
	"net"
	"testing"
	"time"

	"carbonblack.com/protocols/psc/ems/server/emdstatus"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/grpcserver"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/grpcserver/grpcmiddleware"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	"gitlab.bit9.local/psc/platform/ems/device-service/proto/grpc/deviceservice"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog/log"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/grpc/test/bufconn"
)

//nolint:gochecknoglobals // singleton test controller
var ctrl *gomock.Controller

func TestMiddleware(t *testing.T) {
	ctrl = gomock.NewController(t)
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "Middleware")
}

var _ = Describe("Middleware", func() {
	var server *grpc.Server
	var client deviceservice.DeviceServiceClient
	var mockMetricsReporter *mocks.MetricReporter
	var mockDB *mocks.DB
	var mockDevEventsService *mocks.DeviceEventsPublisher
	var mockEvictor *mocks.Evictor
	var mockCbd *mocks.Cbd
	const bufSize = 1024 * 1024
	var conn *grpc.ClientConn
	var err error

	var lis *bufconn.Listener

	bufDialer := func(context.Context, string) (net.Conn, error) {
		//nolint:wrapcheck // testing
		return lis.Dial()
	}

	Context("report RED metrics", func() {
		BeforeSuite(func() {
			// set up server with in-memory network comms
			// https://stackoverflow.com/questions/42102496/testing-a-grpc-service
			lis = bufconn.Listen(bufSize)
			server = grpc.NewServer(grpc.UnaryInterceptor(grpcmiddleware.UnaryInterceptor))
			mockDevEventsService = mocks.NewDeviceEventsPublisher(ctrl)
			mockDB = mocks.NewDB(ctrl)
			mockEvictor = mocks.NewEvictor(ctrl)
			mockCbd = mocks.NewCbd(ctrl)

			deviceservice.RegisterDeviceServiceServer(server, grpcserver.NewDeviceServiceServer(mockDB, mockEvictor,
				mockDevEventsService, mockCbd))

			go func() {
				if err := server.Serve(lis); err != nil {
					log.Fatal().Err(err).Msg("Server exited with error")
				}
			}()

			ctx, cancel := context.WithTimeout(context.Background(), time.Second*5)
			defer cancel()
			conn, err = grpc.DialContext(
				ctx,
				"bufnet",
				grpc.WithContextDialer(bufDialer),
				grpc.WithTransportCredentials(insecure.NewCredentials()))

			Expect(err).To(BeNil())
			client = deviceservice.NewDeviceServiceClient(conn)
			mockMetricsReporter = mocks.NewMetricReporter(ctrl)
		})

		AfterSuite(func() {
			err = conn.Close()
			Expect(err).To(BeNil())
			server.GracefulStop()
		})

		It("should report some REDers", func() {
			telemetry.MetricReporter = mockMetricsReporter
			dims := map[string]string{
				"error":  "false",
				"method": "/ems.device_service.DeviceService/Update",
				"status": "0",
			}
			data := map[string]string{
				model.DataFieldRegistrationID:         "1-1",
				model.DataFieldDeviceType:             "1",
				model.DataFieldOSMajorVersion:         "2320",
				model.DataFieldOSMinorVersion:         "5",
				model.DataFieldMaintenanceRelease:     "5",
				model.DataFieldOSBuildNumber:          "5",
				model.DataFieldPlatformType:           "2",
				model.DataFieldOrgKey:                 "R76KWHRN",
				model.DataFieldHostname:               "testHost",
				model.DataFieldProductVersion:         "1",
				model.DataFieldOriginalRequestHeaders: "",
				model.DataFieldOriginalRequestMsgPack: "",
				model.DataFieldMessageType:            "8370",
			}

			mockDB.EXPECT().Update(gomock.Any(), gomock.Any())
			mockMetricsReporter.EXPECT().GetOrRegisterMetric("grpc", gomock.Any(), dims)
			mockMetricsReporter.EXPECT().GetOrRegisterMetric("grpc.timer", gomock.Any(), dims)
			mockCbd.EXPECT().Update(gomock.Any(), gomock.Any()).Return(nil)

			msg := &deviceservice.DeviceUpdateRequest{Msg: &emdstatus.DeviceStatusMsg{Metadata: data}}

			_, err := client.Update(context.Background(), msg)
			Expect(err).To(BeNil())
		})
	})
})
