package logging

import (
	"io"
	stdlog "log"
	"os"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"

	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/rs/zerolog/pkgerrors"
	"github.com/spf13/viper"
)

func Configure() {
	//nolint:reassign // following zerolog pattern to set ErrorStackMarshaler
	// https://betterstack.com/community/guides/logging/zerolog/#logging-errors-with-zerolog
	zerolog.ErrorStackMarshaler = pkgerrors.MarshalStack

	var logOut io.Writer

	switch viper.GetString(config.LogFormatKey) {
	case "pretty":
		logOut = zerolog.NewConsoleWriter()
	case "json":
	default:
		logOut = os.Stdout
	}

	// allow a burst of 20 error logs in 3 seconds, over 20 error logs and 1/100 error messages are logged.
	// so if 320 logs are received in 3 seconds 20 error messages will be logged based on the burst and then
	// 3 logs based on the next sampler resulting in 23 log entries instead of 320
	logger := zerolog.New(logOut).With().
		Timestamp().
		Str("app", viper.GetString(config.AppNameKey)).
		Logger().
		Sample(zerolog.LevelSampler{
			ErrorSampler: &zerolog.BurstSampler{
				Burst:       viper.GetUint32(config.LogSamplerBurstKey),
				Period:      viper.GetDuration(config.LogSamplerPeriodSecKey),
				NextSampler: &zerolog.BasicSampler{N: viper.GetUint32(config.LogSamplerNextSamplerKey)},
			},
		})

	if viper.GetBool(config.LogCallerKey) {
		logger = logger.With().Caller().Logger()
	}

	log.Logger = logger

	if viper.GetBool(config.LogStdlibKey) {
		stdlog.SetFlags(0)
		stdlog.SetOutput(logger)
	} else {
		stdlog.SetOutput(io.Discard)
	}

	rawLevel := viper.GetString(config.LogLevelKey)
	if level, err := zerolog.ParseLevel(rawLevel); err != nil {
		logger.Error().Err(err).Str("value", rawLevel).Msg("invalid log level")
	} else if level != zerolog.GlobalLevel() {
		zerolog.SetGlobalLevel(level)
		log.Info().Str("value", rawLevel).Msg("updated log level")
	}
}
