package logging_test

import (
	"io"
	stdlog "log"
	"os"
	"testing"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/logging"
	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

//nolint:gochecknoglobals // singleton test controller
var ctrl *gomock.Controller

func TestLogging(t *testing.T) {
	ctrl = gomock.NewController(t)
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "Logging")
}

var _ = Describe("Logging", func() {
	BeforeEach(func() {
		viper.Reset()
		// init sampler config
		viper.SetDefault(config.LogSamplerBurstKey, 20)
		viper.SetDefault(config.LogSamplerPeriodSecKey, "3s")
		viper.SetDefault(config.LogSamplerNextSamplerKey, 100)
	})

	Context("Configure function", func() {
		It("should configure a logger", func() {
			Expect(stdlog.Flags()).To(Equal(3))
			Expect(stdlog.Writer()).To(Equal(os.Stderr))

			unconfiguredLogger := log.Logger

			logging.Configure()

			Expect(log.Logger).ToNot(Equal(unconfiguredLogger))
			Expect(stdlog.Flags()).To(Equal(3))
			Expect(stdlog.Writer()).To(Equal(io.Discard))
		})

		It("should set a global log level", func() {
			zerolog.SetGlobalLevel(zerolog.DebugLevel)
			Expect(zerolog.GlobalLevel()).To(Equal(zerolog.DebugLevel))

			viper.Set(config.LogLevelKey, "error")
			logging.Configure()
			Expect(zerolog.GlobalLevel()).To(Equal(zerolog.ErrorLevel))
		})

		It("should be ok with bad level", func() {
			zerolog.SetGlobalLevel(zerolog.DebugLevel)
			Expect(zerolog.GlobalLevel()).To(Equal(zerolog.DebugLevel))

			viper.Set(config.LogLevelKey, "wat")
			logging.Configure()
			Expect(zerolog.GlobalLevel()).To(Equal(zerolog.DebugLevel))
		})

		Context("format settings", func() {
			var mockWriter *mocks.Writer

			BeforeEach(func() {
				mockWriter = mocks.NewWriter(ctrl)

				viper.Set(config.LogLevelKey, "info")
			})

			XIt("should support pretty output", func() {
				var testLog string

				viper.Set(config.LogFormatKey, "pretty")
				logging.Configure()

				logger := log.Output(mockWriter)

				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&testLog),
				).Times(1)

				logger.Info().Str("test1", "one").Int("test2", 2).Msg("testing")

				Expect(testLog).To(ContainSubstring("level=info"))
				Expect(testLog).To(ContainSubstring("test1=one"))
				Expect(testLog).To(ContainSubstring("test2=2"))
				Expect(testLog).To(ContainSubstring("testing"))
				Expect(testLog).ToNot(ContainSubstring("message"))
			})

			It("should default to no log caller", func() {
				var testLog string

				logging.Configure()

				logger := log.Output(mockWriter)

				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&testLog),
				).Times(1)

				logger.Info().Str("test1", "one").Int("test2", 2).Msg("testing")

				Expect(testLog).ToNot(ContainSubstring("\"caller\":\""))
			})

			It("should support log caller", func() {
				var testLog string

				viper.Set(config.LogCallerKey, "true")

				logging.Configure()

				logger := log.Output(mockWriter)

				mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&testLog),
				).Times(1)

				logger.Info().Str("test1", "one").Int("test2", 2).Msg("testing")

				Expect(testLog).To(ContainSubstring("\"caller\":\""))
			})
		})

		Context("with standard logger", func() {
			It("should configure standard logger", func() {
				viper.Set(config.LogStdlibKey, "true")

				Expect(stdlog.Flags()).To(Equal(3))
				Expect(stdlog.Writer()).ToNot(Equal(log.Logger))

				logging.Configure()

				Expect(stdlog.Flags()).To(Equal(0))
				Expect(stdlog.Writer()).To(Equal(log.Logger))
			})
		})
	})
})
