package middleware_test

import (
	"errors"
	"net/http"
	"net/url"
	"testing"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/middleware"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	"github.com/labstack/echo/v4"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

var (
	errTest     = errors.New("test")
	errExpected = errors.New("an expected error")
)

//nolint:gochecknoglobals // singleton test controller
var ctrl *gomock.Controller

func TestMiddleware(t *testing.T) {
	ctrl = gomock.NewController(t)
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "Middleware")
}

var _ = Describe("Middleware", func() {
	const requestIDHeader = "X-Request-ID"
	const wfTraceHeader = "Wf-Ot-Traceid"
	const wfSpanHeader = "Wf-Ot-Spanid"
	const wfSampleHeader = "Wf-Ot-Sample"

	var mockWriter *mocks.Writer
	var mockResponseWriter *mocks.ResponseWriter
	var responseHeaders http.Header
	var request *http.Request
	var testEcho *echo.Echo

	Context("Apply method", func() {
		BeforeEach(func() {
			viper.Reset()
			viper.Set(config.HTTPRequestIDHeaderKey, requestIDHeader)

			//nolint:exhaustruct // only used for testing, just need method, url, and header
			request = &http.Request{
				Method: http.MethodGet,
				//nolint:exhaustruct // only used for testing, just need path
				URL:    &url.URL{Path: "/"},
				Header: make(http.Header),
			}

			mockResponseWriter = mocks.NewResponseWriter(ctrl)
			responseHeaders = http.Header{}
			mockResponseWriter.EXPECT().Header().Return(responseHeaders).AnyTimes()

			testEcho = echo.New()

			telemetry.SetupMetrics(nil)
		})

		Context("with panic", func() {
			BeforeEach(func() {
				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error { panic(errTest) })
				mockResponseWriter.EXPECT().WriteHeader(http.StatusInternalServerError).Times(1)
			})

			It("should recover from panic", func() {
				var response string

				mockResponseWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&response),
				).Times(1)

				Expect(func() { testEcho.ServeHTTP(mockResponseWriter, request) }).ToNot(Panic())
				Expect(response).To(Equal("{\"message\":\"Internal Server Error\"}\n"))
			})
		})

		Context("with plain error", func() {
			BeforeEach(func() {
				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error { return errExpected })
				mockResponseWriter.EXPECT().WriteHeader(http.StatusInternalServerError).Times(1)
			})

			It("should handle error", func() {
				var response string

				mockResponseWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&response),
				).Times(1)

				testEcho.ServeHTTP(mockResponseWriter, request)
				Expect(len(responseHeaders)).To(Equal(2)) // request id and json error content type
				Expect(len(request.Header)).To(BeZero())
				Expect(response).To(Equal("{\"message\":\"Internal Server Error\"}\n"))
			})
		})

		Context("with http error", func() {
			BeforeEach(func() {
				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error {
					return echo.NewHTTPError(http.StatusBadGateway, errExpected)
				})
				mockResponseWriter.EXPECT().WriteHeader(http.StatusBadGateway).Times(1)
			})

			It("should handle error", func() {
				var response string

				mockResponseWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&response),
				).Times(1)

				testEcho.ServeHTTP(mockResponseWriter, request)
				Expect(len(responseHeaders)).To(Equal(2)) // request id and json error content type
				Expect(len(request.Header)).To(BeZero())
				Expect(response).To(Equal("{}\n"))
			})
		})

		Context("with timeout", func() {
			BeforeEach(func() {
				viper.Set(config.HTTPRequestTimeoutKey, "1ms")
				//nolint:godox // temporary TODO
				// TODO: there is a data race condition in the echo timeout middleware
				// so it needs to be tested alone until fixed
				// middleware.Apply(testEcho)
				// testEcho.Use(middleware.RequestLogger())
				testEcho.Use(middleware.Timeout())
				testEcho.Any("/", func(ctx echo.Context) error {
					time.Sleep(1 * time.Second)

					return nil
				})
				mockResponseWriter.EXPECT().WriteHeader(http.StatusServiceUnavailable).Times(1)
			})

			It("should handle error", func() {
				var response string

				mockResponseWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
					helper.GenerateWriterDoStore(&response),
				).Times(1)

				testEcho.ServeHTTP(mockResponseWriter, request)
				Expect(len(responseHeaders)).To(BeZero())
				Expect(len(request.Header)).To(BeZero())
				Expect(response).To(Equal("timeout"))
			})
		})

		Context("with tracing disabled", func() {
			BeforeEach(func() {
				viper.Set(config.TracingEnabledKey, false)
				telemetry.SetupTracing(nil)

				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error { return nil })
			})

			It("should not trace", func() {
				testEcho.ServeHTTP(mockResponseWriter, request)
				Expect(len(responseHeaders)).To(Equal(1))
				Expect(len(request.Header)).To(BeZero())
				Expect(len(request.Header.Get(wfTraceHeader))).To(BeZero())
				Expect(len(request.Header.Get(wfSpanHeader))).To(BeZero())
				Expect(len(request.Header.Get(wfSampleHeader))).To(BeZero())
			})
		})

		Context("with tracing enabled", func() {
			BeforeEach(func() {
				viper.Set(config.TracingEnabledKey, true)
				telemetry.SetupTracing(nil)

				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error { return nil })
				testEcho.GET(config.HealthCheckPath, func(ctx echo.Context) error { return nil })
			})

			Context("without mock logger", func() {
				It("should succeed", func() {
					testEcho.ServeHTTP(mockResponseWriter, request)
					Expect(len(responseHeaders)).To(Equal(1))
					Expect(len(request.Header)).To(Equal(3))
					Expect(len(request.Header.Get(wfTraceHeader))).ToNot(BeZero())
					Expect(len(request.Header.Get(wfSpanHeader))).ToNot(BeZero())
					Expect(len(request.Header.Get(wfSampleHeader))).ToNot(BeZero())
				})
			})

			Context("with mock logger", func() {
				BeforeEach(func() {
					header := http.Header{}
					header.Add("X-Test", "test")
					header.Add("X-Auth-Token", "test")
					header.Add("x-auth-token", "test2")
					request.Header = header

					mockWriter = mocks.NewWriter(ctrl)
					logger := log.With().Logger().Output(mockWriter)
					request = request.WithContext(logger.WithContext(request.Context()))
				})

				It("should use the middleware", func() {
					var requestLoggerLog string

					gomock.InOrder(
						mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
							helper.GenerateWriterDoStore(&requestLoggerLog),
						).Times(1),
					)

					testEcho.ServeHTTP(mockResponseWriter, request)

					Expect(requestLoggerLog).To(ContainSubstring("\"level\":\"info\""))
					Expect(requestLoggerLog).To(ContainSubstring("\"message\":\"request\""))
					Expect(requestLoggerLog).To(ContainSubstring("\"method\":\"GET\""))
					Expect(requestLoggerLog).To(ContainSubstring("\"uri\":\"\""))
					Expect(requestLoggerLog).To(ContainSubstring("\"status\":200"))
					Expect(requestLoggerLog).To(ContainSubstring("\"latency\":"))
					Expect(requestLoggerLog).To(ContainSubstring("\"headers\":{"))
					Expect(requestLoggerLog).To(ContainSubstring("\"X-Test\":[\"test\"]"))
					Expect(requestLoggerLog).To(ContainSubstring("\"X-Auth-Token\":[\"****\",\"*****\"]"))
					Expect(requestLoggerLog).To(ContainSubstring("\"time\":\""))
					Expect(requestLoggerLog).To(ContainSubstring("\"request_id\":\""))
				})
			})

			Context("for health check", func() {
				BeforeEach(func() {
					//nolint:exhaustruct // only used for testing, just need path
					request.URL = &url.URL{Path: config.HealthCheckPath}
				})

				It("should not trace", func() {
					testEcho.ServeHTTP(mockResponseWriter, request)
					Expect(len(responseHeaders)).To(Equal(1))
					Expect(len(request.Header)).To(BeZero())
					Expect(len(request.Header.Get(wfTraceHeader))).To(BeZero())
					Expect(len(request.Header.Get(wfSpanHeader))).To(BeZero())
					Expect(len(request.Header.Get(wfSampleHeader))).To(BeZero())
				})
			})
		})

		Context("with metrics disabled", func() {
			BeforeEach(func() {
				mockSender := mocks.NewMetricSender(ctrl)

				viper.Set(config.MetricsEnabledKey, false)
				telemetry.SetupMetrics(mockSender)

				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error { return nil })
			})

			AfterEach(func() {
				telemetry.MetricReporter.Close()
			})

			It("should not meter", func() {
				testEcho.ServeHTTP(mockResponseWriter, request)
				telemetry.MetricReporter.Report()
			})
		})

		Context("with metrics enabled", func() {
			var (
				mockReporter *mocks.MetricReporter
				badPath      = "/bad"
			)

			BeforeEach(func() {
				viper.Set(config.MetricsEnabledKey, true)

				mockReporter = mocks.NewMetricReporter(ctrl)
				telemetry.MetricReporter = mockReporter

				middleware.Apply(testEcho)
				testEcho.Any("/", func(ctx echo.Context) error { return nil })
				testEcho.Any(badPath, func(ctx echo.Context) error { return errExpected })
				testEcho.GET(config.HealthCheckPath, func(ctx echo.Context) error { return nil })
			})

			It("should succeed", func() {
				dims := map[string]string{
					"error":  "false",
					"method": "GET",
					"path":   "/",
					"status": "200",
				}

				mockReporter.EXPECT().GetOrRegisterMetric("http", gomock.Any(), dims)
				mockReporter.EXPECT().GetOrRegisterMetric("http.timer", gomock.Any(), dims)

				testEcho.ServeHTTP(mockResponseWriter, request)
			})

			It("should report error", func() {
				dims := map[string]string{
					"error":  "true",
					"method": "GET",
					"path":   badPath,
					"status": "500",
				}

				mockReporter.EXPECT().GetOrRegisterMetric("http", gomock.Any(), dims)
				mockReporter.EXPECT().GetOrRegisterMetric("http.timer", gomock.Any(), dims)

				//nolint:exhaustruct // just testing here, only need Path
				request.URL = &url.URL{Path: badPath}

				mockResponseWriter.EXPECT().WriteHeader(http.StatusInternalServerError).Times(1)
				mockResponseWriter.EXPECT().Write(gomock.Any())
				testEcho.ServeHTTP(mockResponseWriter, request)
			})

			Context("for health check", func() {
				BeforeEach(func() {
					//nolint:exhaustruct // only used for testing, just need path
					request.URL = &url.URL{Path: config.HealthCheckPath}
				})

				It("should not meter", func() {
					testEcho.ServeHTTP(mockResponseWriter, request)
				})
			})
		})
	})
})
