package middleware

import (
	"context"
	"errors"
	"net/http"
	"strings"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"

	"github.com/labstack/echo/v4"
	"github.com/labstack/echo/v4/middleware"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
)

var errUnknown = errors.New("unknown error")

// RequestLogger adds a child logger to the request context and logs the request information.
func RequestLogger() echo.MiddlewareFunc {
	return middleware.RequestLoggerWithConfig(
		//nolint: exhaustruct // only need to override RequestLoggerConfig defaults
		middleware.RequestLoggerConfig{
			BeforeNextFunc:  requestLoggerBeforeNext,
			LogError:        true,
			LogHost:         true,
			LogLatency:      true,
			LogMethod:       true,
			LogProtocol:     true,
			LogRemoteIP:     true,
			LogResponseSize: true,
			LogStatus:       true,
			LogURI:          true,
			LogValuesFunc:   requestLoggerLogValues,
			Skipper:         requestLoggerSkipper,
		},
	)
}

func requestLoggerBeforeNext(echoCtx echo.Context) {
	ctx := echoCtx.Request().Context()
	logger := *zerolog.Ctx(ctx) // try to get an existing logger from the context

	// if the existing request logger is disabled, there may not be a logger set on the request
	if logger.GetLevel() == zerolog.Disabled {
		// use a child of the global logger
		logger = log.With().Logger()
	}

	ctx = logger.WithContext(ctx)

	echoCtx.SetRequest(echoCtx.Request().WithContext(ctx))
}

func requestLoggerLogValues(echoCtx echo.Context, values middleware.RequestLoggerValues) error {
	if values.Error != nil {
		var httpErr *echo.HTTPError
		if !errors.As(values.Error, &httpErr) {
			values.Status = http.StatusInternalServerError
		}
	} else if values.Status >= http.StatusInternalServerError {
		values.Error = errUnknown
	}

	ctx := echoCtx.Request().Context()
	logger := zerolog.Ctx(ctx)

	if err := ctx.Err(); errors.Is(err, context.DeadlineExceeded) {
		// A timeout attempts to return a 503.
		values.Status = http.StatusServiceUnavailable
		values.Error = err
	}

	logger.
		Err(values.Error).
		Str("protocol", values.Protocol).
		Str("method", values.Method).
		Str("uri", values.URI).
		Int("status", values.Status).
		Dur("latency", values.Latency).
		Int("response-size", int(values.ResponseSize)).
		Str("host", values.Host).
		Str("remote-ip", values.RemoteIP).
		Dict("headers", headersToLogEvent(echoCtx.Request().Header)).
		Msg("request")

	return nil
}

func requestLoggerSkipper(echoCtx echo.Context) bool {
	return strings.Contains(echoCtx.Request().URL.Path, "healthCheck")
}

// headersToLogEvent creates a structure from HTTP headers suitable for use as a zerolog.Dict field.
//
// A handful of known, PSC-specific headers that might contain sensitive data are redacted.
func headersToLogEvent(headers http.Header) *zerolog.Event {
	evt := zerolog.Dict()

	for header, values := range headers {
		if config.IsRedactedHeader(header) {
			redactedValues := make([]string, len(values))

			for i, value := range values {
				redactedValues[i] = strings.Repeat("*", len(value))
			}

			evt.Strs(header, redactedValues)
		} else {
			evt.Strs(header, values)
		}
	}

	return evt
}
