package middleware

import (
	"context"
	"errors"
	"net/http"
	"strconv"
	"strings"

	"github.com/labstack/echo/v4"
	"github.com/labstack/echo/v4/middleware"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
)

// RequestMetrics records request metrics.
func RequestMetrics() echo.MiddlewareFunc {
	return middleware.RequestLoggerWithConfig(
		//nolint: exhaustruct // only need to override RequestLoggerConfig defaults
		middleware.RequestLoggerConfig{
			LogError:      true,
			LogMethod:     true,
			LogStatus:     true,
			LogURIPath:    true,
			LogValuesFunc: requestMetricsLogValues,
			Skipper:       requestMetricsSkipper,
		},
	)
}

func requestMetricsLogValues(echoCtx echo.Context, values middleware.RequestLoggerValues) error {
	if values.Error != nil {
		var httpErr *echo.HTTPError
		if !errors.As(values.Error, &httpErr) {
			values.Status = http.StatusInternalServerError
		}
	} else if values.Status >= http.StatusInternalServerError {
		values.Error = errUnknown
	}

	ctx := echoCtx.Request().Context()

	if err := ctx.Err(); errors.Is(err, context.DeadlineExceeded) {
		// A timeout attempts to return a 503.
		values.Status = http.StatusServiceUnavailable
		values.Error = err
	}

	dims := map[string]string{
		"method": values.Method,
		"path":   values.URIPath,
		"status": strconv.Itoa(values.Status),
		"error":  strconv.FormatBool(values.Error != nil || values.Status >= http.StatusInternalServerError),
	}

	telemetry.CounterInc("http", dims, 1)
	telemetry.HistogramUpdateTimeSince("http.timer", dims, values.StartTime)

	//nolint:nilerr // umm, error is nil, bad linter!
	return nil
}

func requestMetricsSkipper(echoCtx echo.Context) bool {
	return strings.Contains(echoCtx.Request().URL.Path, "healthCheck")
}
