package middleware

import (
	"net/http"
	"strings"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"

	"github.com/labstack/echo/v4"
	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

func Tracing() echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		if !viper.GetBool(config.TracingEnabledKey) {
			// skip if tracing not enabled
			return next
		}

		return func(echoCtx echo.Context) error {
			if strings.HasSuffix(echoCtx.Path(), config.HealthCheckPath) {
				// ignore health check requests for distributed tracing
				return next(echoCtx)
			}

			request := echoCtx.Request()
			tracer := opentracing.GlobalTracer()
			carrier := opentracing.HTTPHeadersCarrier(request.Header)
			ctx, _ := tracer.Extract(opentracing.HTTPHeaders, carrier)
			op := "HTTP " + request.Method + " " + request.URL.Path

			span := opentracing.StartSpan(op, ext.RPCServerOption(ctx))
			defer span.Finish()

			ext.HTTPMethod.Set(span, request.Method)
			ext.HTTPUrl.Set(span, request.URL.String())
			ext.Component.Set(span, "http-api")

			request = request.WithContext(opentracing.ContextWithSpan(request.Context(), span))
			echoCtx.SetRequest(request)

			if err := tracer.Inject(span.Context(), opentracing.HTTPHeaders, carrier); err != nil {
				log.Err(err).Msg("failed to inject span into carrier")
			}

			err := next(echoCtx)

			ext.HTTPStatusCode.Set(span, uint16(echoCtx.Response().Status))

			if err != nil || echoCtx.Response().Status >= http.StatusInternalServerError {
				span.SetTag("error", true)
			} else {
				span.SetTag("error", false)
			}

			return err
		}
	}
}

func AugmentTraces() echo.MiddlewareFunc {
	return func(next echo.HandlerFunc) echo.HandlerFunc {
		return func(ctx echo.Context) error {
			span := opentracing.SpanFromContext(ctx.Request().Context())
			if span == nil {
				return next(ctx)
			}

			// orgKey := orgKeyFromEchoContext(ctx)
			// if orgKey != nil {
			// 	span.SetTag("org_key", orgKey.String())
			// 	span.SetTag("origin_env", orgKey.Env())
			// }

			return next(ctx)
		}
	}
}
