package model

import (
	"fmt"
	"strconv"
	"strings"

	"github.com/rs/zerolog/log"
	orgKeyHelper "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

const registrationIDsLength = 2

var ErrBadRegistrationID = fmt.Errorf("bad registration id")

func envFromData(data map[string]string) string {
	orgKey, err := orgKeyHelper.Decode(data[DataFieldOrgKey])
	if err != nil {
		log.Err(err).Msg("unable to decode orgkey")

		return ""
	}

	return orgKey.Env()
}

func orgAndDeviceIDFromData(data map[string]string) (int, int, error) {
	ids := strings.Split(data[DataFieldRegistrationID], "-")
	if len(ids) != registrationIDsLength {
		return 0, 0, ErrBadRegistrationID
	}

	orgID, err := strconv.Atoi(ids[0])
	if err != nil {
		return 0, 0, fmt.Errorf("bad org id: %w", err)
	}

	devID, err := strconv.Atoi(ids[1])
	if err != nil {
		return 0, 0, fmt.Errorf("bad device id: %w", err)
	}

	return orgID, devID, nil
}

func deploymentTypeFromInfrastructureProvider(ipe InfrastructureProviderEnum) (string, string) {
	switch ipe {
	case InfrastructureProviderAWS:
		return DeploymentTypeAWS.String(), SubDeploymentTypeAWSVirtualMachineEC2.String()
	case InfrastructureProviderAzure:
		return DeploymentTypeAzure.String(), ""
	case InfrastructureProviderGCP:
		return DeploymentTypeGCP.String(), ""
	case InfrastructureProviderNone: // error? legacy cbd code does not handle this, falls through
	}

	return DeploymentTypeEndpoint.String(), ""
}

func deploymentTypeFromVirtualizationProvider(vpe VirtualizationProviderEnum) (string, string) {
	switch vpe {
	case VirtualizationProviderWorkload:
		return DeploymentTypeWorkload.String(), SubDeploymentTypeVMwareVirtualMachine.String()
	case VirtualizationProviderAWS:
		return DeploymentTypeAWS.String(), SubDeploymentTypeAWSVirtualMachineEC2.String()
	case VirtualizationProviderNone: // error? legacy cbd code does not handle this, falls through
	}

	return DeploymentTypeEndpoint.String(), ""
}

func deploymentTypeFromData(data map[string]string) (string, string) {
	switch {
	case data[DataFieldVDIProvider] != "":
		return DeploymentTypeVDI.String(), NewVDIProvider(data[DataFieldVDIProvider]).String()
	case data[DataFieldInfrastructureProvider] != "":
		return deploymentTypeFromInfrastructureProvider(
			NewInfrastructureProvider(data[DataFieldInfrastructureProvider]),
		)
	case data[DataFieldVirtualizationProvider] != "":
		return deploymentTypeFromVirtualizationProvider(
			NewVirtualizationProvider(data[DataFieldVirtualizationProvider]),
		)
	}

	return DeploymentTypeEndpoint.String(), ""
}
