//nolint:testpackage // testing internal helper funcs
package model

import (
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("data helpers", func() {
	Context("envFromData func", func() {
		It("should return an empty string given data without an orgkey", func() {
			Expect(envFromData(map[string]string{})).To(Equal(""))
		})

		It("should return an empty string given data with a bad orgkey", func() {
			Expect(envFromData(map[string]string{DataFieldOrgKey: "abc123"})).To(Equal(""))
		})

		It("should return an env given data with a good orgkey", func() {
			Expect(envFromData(map[string]string{DataFieldOrgKey: "Y82TXG9PJ8"})).To(Equal("dev01"))
			Expect(envFromData(map[string]string{DataFieldOrgKey: "67ZZDTP7"})).To(Equal("dev01"))
			Expect(envFromData(map[string]string{DataFieldOrgKey: "N4LFZ4DN"})).To(Equal("prod05"))
		})
	})

	Context("orgAndDeviceIDFromData func", func() {
		It("should return not ok with no registration id", func() {
			_, _, err := orgAndDeviceIDFromData(map[string]string{})
			Expect(err).To(HaveOccurred())
		})

		It("should return not ok with empty registration id", func() {
			_, _, err := orgAndDeviceIDFromData(map[string]string{DataFieldRegistrationID: ""})
			Expect(err).To(HaveOccurred())
		})

		It("should return not ok with bad registration id", func() {
			_, _, err := orgAndDeviceIDFromData(map[string]string{DataFieldRegistrationID: "hello"})
			Expect(err).To(HaveOccurred())

			_, _, err = orgAndDeviceIDFromData(map[string]string{DataFieldRegistrationID: "11"})
			Expect(err).To(HaveOccurred())
		})

		It("should return device and org id with valid registration id", func() {
			orgID, deviceID, err := orgAndDeviceIDFromData(map[string]string{DataFieldRegistrationID: "1-2"})

			Expect(err).ToNot(HaveOccurred())
			Expect(orgID).To(Equal(1))
			Expect(deviceID).To(Equal(2))
		})
	})

	Context("deploymentTypeFromInfrastructureProvider func", func() {
		It("should return strings for model.InfrastructureProviderAWS", func() {
			deploymentType, subDeploymentType := deploymentTypeFromInfrastructureProvider(InfrastructureProviderAWS)

			Expect(deploymentType).To(Equal(DeploymentTypeAWS.String()))
			Expect(subDeploymentType).To(Equal(SubDeploymentTypeAWSVirtualMachineEC2.String()))
		})

		It("should return strings for model.InfrastructureProviderAzure", func() {
			deploymentType, subDeploymentType := deploymentTypeFromInfrastructureProvider(InfrastructureProviderAzure)

			Expect(deploymentType).To(Equal(DeploymentTypeAzure.String()))
			Expect(subDeploymentType).To(Equal(""))
		})

		It("should return strings for model.InfrastructureProviderGCP", func() {
			deploymentType, subDeploymentType := deploymentTypeFromInfrastructureProvider(InfrastructureProviderGCP)

			Expect(deploymentType).To(Equal(DeploymentTypeGCP.String()))
			Expect(subDeploymentType).To(Equal(""))
		})

		It("should return strings for model.InfrastructureProviderNone", func() {
			deploymentType, subDeploymentType := deploymentTypeFromInfrastructureProvider(InfrastructureProviderNone)

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))
		})

		It("should return strings for invalid model.InfrastructureProviderEnum", func() {
			deploymentType, subDeploymentType := deploymentTypeFromInfrastructureProvider(InfrastructureProviderEnum(-1))

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))
		})
	})

	Context("deploymentTypeFromVirtualizationProvider func", func() {
		It("should return strings for model.VirtualizationProviderWorkload", func() {
			deploymentType, subDeploymentType := deploymentTypeFromVirtualizationProvider(VirtualizationProviderWorkload)

			Expect(deploymentType).To(Equal(DeploymentTypeWorkload.String()))
			Expect(subDeploymentType).To(Equal(SubDeploymentTypeVMwareVirtualMachine.String()))
		})

		It("should return strings for model.VirtualizationProviderAWS", func() {
			deploymentType, subDeploymentType := deploymentTypeFromVirtualizationProvider(VirtualizationProviderAWS)

			Expect(deploymentType).To(Equal(DeploymentTypeAWS.String()))
			Expect(subDeploymentType).To(Equal(SubDeploymentTypeAWSVirtualMachineEC2.String()))
		})

		It("should return strings for model.VirtualizationProviderNone", func() {
			deploymentType, subDeploymentType := deploymentTypeFromVirtualizationProvider(VirtualizationProviderNone)

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))
		})

		It("should return strings for invalid model.VirtualizationProviderEnum", func() {
			deploymentType, subDeploymentType := deploymentTypeFromVirtualizationProvider(VirtualizationProviderEnum(-1))

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))
		})
	})

	Context("deploymentTypeFromData func", func() {
		var (
			deploymentType    string
			subDeploymentType string

			emptyData = map[string]string{}
			vdiData   = map[string]string{
				DataFieldVDIProvider: "horizon",
			}
			newerSensorData = map[string]string{
				DataFieldInfrastructureProvider: "aws",
			}
			legacySensorData = map[string]string{
				DataFieldVirtualizationProvider: "VMw_ESX",
			}
			badData1 = map[string]string{
				DataFieldVDIProvider: "nope",
			}
			badData2 = map[string]string{
				DataFieldInfrastructureProvider: "nope",
			}
			badData3 = map[string]string{
				DataFieldVirtualizationProvider: "nope",
			}
		)

		It("should return strings", func() {
			deploymentType, subDeploymentType = deploymentTypeFromData(emptyData)

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))

			deploymentType, subDeploymentType = deploymentTypeFromData(vdiData)

			Expect(deploymentType).To(Equal(DeploymentTypeVDI.String()))
			Expect(subDeploymentType).To(Equal(VDIProviderHorizon.String()))

			deploymentType, subDeploymentType = deploymentTypeFromData(newerSensorData)

			Expect(deploymentType).To(Equal(DeploymentTypeAWS.String()))
			Expect(subDeploymentType).To(Equal(SubDeploymentTypeAWSVirtualMachineEC2.String()))

			deploymentType, subDeploymentType = deploymentTypeFromData(legacySensorData)

			Expect(deploymentType).To(Equal(DeploymentTypeWorkload.String()))
			Expect(subDeploymentType).To(Equal(SubDeploymentTypeVMwareVirtualMachine.String()))

			deploymentType, subDeploymentType = deploymentTypeFromData(badData1)

			Expect(deploymentType).To(Equal(DeploymentTypeVDI.String()))
			Expect(subDeploymentType).To(Equal(VDIProviderNone.String()))

			deploymentType, subDeploymentType = deploymentTypeFromData(badData2)

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))

			deploymentType, subDeploymentType = deploymentTypeFromData(badData3)

			Expect(deploymentType).To(Equal(DeploymentTypeEndpoint.String()))
			Expect(subDeploymentType).To(Equal(""))
		})
	})
})
