package model

import (
	"encoding/json"
	"fmt"
)

var (
//	dbFields = []string{
//		"DeviceType",                  // device_type_id
//		"Email",                       // email
//		"Hostname",                    // name
//		"MaintenanceRelease",          // properties
//		"OsBuildNumber",               // properties
//		"OsMajorVersion",              // properties
//		"OsMinorVersion",              // properties
//		"PlatformType",                // properties
//		"ProductVersion",              // properties
//		"UnifiedEndpointManagementID", // properties
//
//		"AgentLocation",              // NEW properties
//		"AvLastScanTime",             // NEW properties
//		"AvReachableServers",         // NEW properties
//		"AvStatus",                   // NEW properties
//		"AvUpdateServers",            // NEW properties
//		"ExternallyAssignedDeviceId", // NEW properties
//		"InfrastructureProvider",     // NEW properties
//		"IsVirtualMachine",           // NEW properties
//		"LastSensorResetTime",        // NEW properties
//		"LocalAddress",               // NEW properties
//		"LoginUserName",              // NEW properties
//		"PolicyName",                 // NEW properties
//		"Quarantined",                // NEW properties
//		"RootkitDetected",            // NEW properties
//		"SensorStatus",               // NEW properties
//		"VirtualizationProvider",     // NEW properties
//	}
)

type (
	//nolint:tagliatelle // CBD json device props don't play by the rules
	dbFieldProperties struct {
		// Existing props
		Hostname                    string `json:"Host Name,omitempty"`
		MaintenanceRelease          string `json:"Maintenance Release,omitempty"`
		OsBuildNumber               string `json:"Build Number,omitempty"`
		OsMajorVersion              string `json:"OS Major Version,omitempty"`
		OsMinorVersion              string `json:"OS Minor Version,omitempty"`
		PlatformType                string `json:"Platform Type,omitempty"`
		ProductVersion              string `json:"Scargo Client Version,omitempty"`
		SensorKitType               string `json:"_SensorKitType,omitempty"`
		UnifiedEndpointManagementID string `json:"uemId,omitempty"`
		VersionInfo                 string `json:"_VersionInfo,omitempty"`

		// New props
		// These are data that have been stored in ES, but we want to store them in PG to take control of them so
		// that we can more quickly deprecate EMS reliance on ES. Setting non-empty values on these will cause every
		// device to update on their next status call which will cause a higher DB load initially.
		AgentLocation              string `json:",omitempty"`
		AvLastScanTime             string `json:",omitempty"`
		AvReachableServers         string `json:",omitempty"`
		AvStatus                   string `json:",omitempty"`
		AvUpdateServers            string `json:",omitempty"`
		AvVersion                  string `json:",omitempty"`
		AvVersionMinor             string `json:"AvMinorVersion,omitempty"`
		DeviceIP                   string `json:"DeviceIp,omitempty"`
		ExternallyAssignedDeviceID string `json:"ExternallyAssignedDeviceId,omitempty"`
		InfrastructureProvider     string `json:",omitempty"`
		IsVirtualMachine           string `json:",omitempty"`
		LastSensorResetTime        string `json:",omitempty"`
		LocalAddress               string `json:",omitempty"`
		LoginUserName              string `json:",omitempty"`
		MACAddress                 string `json:"MacAddress,omitempty"`
		PolicyName                 string `json:",omitempty"`
		Quarantined                string `json:",omitempty"`
		RootkitDetected            string `json:",omitempty"`
		SensorGatewayURL           string `json:"SensorGatewayUrl,omitempty"`
		SensorGatewayUUID          string `json:"SensorGatewayUuid,omitempty"`
		SensorStatus               string `json:",omitempty"`
		TimeReceived               string `json:",omitempty"`
		TimeSent                   string `json:",omitempty"`
		TLSVersion                 string `json:"TlsVersion,omitempty"`
		VDIProvider                string `json:"VdiProvider,omitempty"`
		VirtualizationProvider     string `json:",omitempty"`
	}
)

//nolint:funlen // function is long primarily because of the use of a large struct, and not because of lengthy logic
func propertiesJSON(data map[string]string) (string, error) {
	version, err := version(
		data[DataFieldOSMajorVersion],
		data[DataFieldOSMinorVersion],
		data[DataFieldMaintenanceRelease],
		data[DataFieldOSBuildNumber],
		data[DataFieldPlatformType],
	)
	if err != nil {
		return "", fmt.Errorf("failed to get version from data: %w "+
			"platform=%s major=%s minor=%s maintenance=%s build=%s",
			err,
			data[DataFieldOSMajorVersion],
			data[DataFieldOSMinorVersion],
			data[DataFieldMaintenanceRelease],
			data[DataFieldOSBuildNumber],
			data[DataFieldPlatformType])
	}

	field := &dbFieldProperties{
		Hostname:                    data[DataFieldHostname],
		PlatformType:                data[DataFieldPlatformType],
		OsMajorVersion:              osRepresentation(data[DataFieldOSMajorVersion]),
		OsMinorVersion:              data[DataFieldOSMinorVersion],
		MaintenanceRelease:          data[DataFieldMaintenanceRelease],
		OsBuildNumber:               data[DataFieldOSBuildNumber],
		ProductVersion:              data[DataFieldProductVersion],
		UnifiedEndpointManagementID: data[DataFieldUnifiedEndpointManagementID],
		SensorKitType:               sensorKit(data[DataFieldOSMajorVersion]),
		VersionInfo:                 version,

		// New Props
		// These are data that have been stored in ES, but we want to store them in PG to take control of them so
		// that we can more quickly deprecate EMS reliance on ES. Adding these will cause every device to update on
		// their next status call which will cause a higher DB load initially.
		AgentLocation:              data[DataFieldAgentLocation],
		AvLastScanTime:             data[DataFieldAvLastScanTime],
		AvReachableServers:         data[DataFieldAvReachableServers],
		AvStatus:                   data[DataFieldAvStatus],
		AvUpdateServers:            data[DataFieldAvUpdateServers],
		AvVersion:                  data[DataFieldAvVersion],
		AvVersionMinor:             data[DataFieldAvVersionMinor],
		DeviceIP:                   data[DataFieldDeviceIP],
		ExternallyAssignedDeviceID: data[DataFieldExternallyAssignedDeviceID],
		InfrastructureProvider:     data[DataFieldInfrastructureProvider],
		IsVirtualMachine:           data[DataFieldIsVirtualMachine],
		LastSensorResetTime:        data[DataFieldLastSensorResetTime],
		LocalAddress:               data[DataFieldLocalAddress],
		LoginUserName:              data[DataFieldLoginUserName],
		MACAddress:                 data[DataFieldMACAddress],
		PolicyName:                 data[DataFieldPolicyName],
		Quarantined:                data[DataFieldQuarantined],
		RootkitDetected:            data[DataFieldRootkitDetected],
		SensorGatewayURL:           data[DataFieldSensorGatewayURL],
		SensorGatewayUUID:          data[DataFieldSensorGatewayUUID],
		SensorStatus:               data[DataFieldSensorStatus],
		TimeReceived:               data[DataFieldTimeReceived],
		TimeSent:                   data[DataFieldTimeSent],
		TLSVersion:                 data[DataFieldTLSVersion],
		VDIProvider:                data[DataFieldVDIProvider],
		VirtualizationProvider:     data[DataFieldVirtualizationProvider],
	}

	props, err := json.Marshal(field)
	if err != nil {
		return "", fmt.Errorf("failed to marshal props to JSON: %w", err)
	}

	return string(props), nil
}
