package model

import (
	"errors"
	"fmt"
	"strconv"

	"github.com/hashicorp/go-multierror"
)

var ErrUnSupportedMessageCode = errors.New("unsupported message code")

// Device is a struct containing device data derived / parsed from a status message.
type Device struct {
	DeploymentType         DeploymentTypeEnum
	DryRunMsg              bool
	Email                  string
	Env                    string
	ExternallyAssignedID   string
	Hostname               string
	ID                     int
	InfrastructureProvider InfrastructureProviderEnum
	OrgID                  int
	OrgKey                 string
	PropsJSON              string
	Type                   int
	VDIProvider            VDIProviderEnum
}

// NewDevice takes a status message and parses relevant device data for EMD dependencies.
func NewDevice(metadata map[string]string) (*Device, error) {
	mErr := validateEMDFields(metadata)

	// DeploymentType.
	deploymentType, _ := deploymentTypeFromData(metadata)
	deploymentEnum := NewDeploymentType(deploymentType)

	// DeviceID and OrgID.
	orgID, deviceID, err := orgAndDeviceIDFromData(metadata)
	if err != nil {
		mErr = multierror.Append(mErr, fmt.Errorf("failed to parse registration ID: %w", err))
	}

	// DeviceProps.
	devicePropsJSON, err := propertiesJSON(metadata)
	if err != nil {
		mErr = multierror.Append(mErr, fmt.Errorf("failed to get properties JSON: %w", err))
	}

	// DeviceType
	deviceType, err := strconv.Atoi(metadata[DataFieldDeviceType])
	if err != nil {
		mErr = multierror.Append(mErr, fmt.Errorf("failed to get device type: %w", err))
	}

	// DryRunMsg.
	dryRunMsg := metadata[DataFieldEchoDummyResponse] == "true"

	// Environment.
	env := envFromData(metadata)

	// InfrastructureProvider.
	infraProvider := NewInfrastructureProvider(metadata[DataFieldInfrastructureProvider])

	// VDIProvider.
	vdiProvider := NewVDIProvider(metadata[DataFieldVDIProvider])

	if mErr != nil {
		return nil, fmt.Errorf("failed to parse device data: %w", mErr)
	}

	return &Device{
		DeploymentType:         deploymentEnum,
		DryRunMsg:              dryRunMsg,
		Email:                  metadata[DataFieldEmail],
		Env:                    env,
		ExternallyAssignedID:   metadata[DataFieldExternallyAssignedDeviceID],
		Hostname:               metadata[DataFieldHostname],
		ID:                     deviceID,
		InfrastructureProvider: infraProvider,
		OrgID:                  orgID,
		OrgKey:                 metadata[DataFieldOrgKey],
		PropsJSON:              devicePropsJSON,
		Type:                   deviceType,
		VDIProvider:            vdiProvider,
	}, nil
}

// validateEMDFields validates that a status message has the required fields to derive / parse data for EMD.
func validateEMDFields(metadata map[string]string) error {
	var (
		fields []string
		mErr   error
	)

	if fields, mErr = getRequiredEMDFields(metadata[DataFieldMessageType]); mErr != nil {
		return fmt.Errorf("error getting required EMD fields: %w", mErr)
	}

	for _, field := range fields {
		//nolint:goerr113 // error message should be dynamic to convey which field(s) are missing
		if metadata[field] == "" {
			mErr = multierror.Append(mErr, fmt.Errorf("status message missing field: %s", field))
		}
	}

	if mErr != nil {
		return fmt.Errorf("error validating required EMD fields: %w", mErr)
	}

	return nil
}

func getRequiredEMDFields(msgCode string) ([]string, error) {
	switch msgCode {
	case DeviceStatusRequestCode.String():
		return []string{
			DataFieldDeviceType,
			DataFieldHostname,
			DataFieldMaintenanceRelease,
			DataFieldOrgKey,
			DataFieldOSMajorVersion,
			DataFieldOSMinorVersion,
			DataFieldPlatformType,
			DataFieldProductVersion,
			DataFieldRegistrationID,
		}, nil
	case DeviceMetaDataRequestCode.String():
		return []string{
			DataFieldDeviceType,
			DataFieldHostname,
			DataFieldMaintenanceRelease,
			DataFieldOrgKey,
			DataFieldOSMajorVersion,
			DataFieldOSMinorVersion,
			DataFieldPlatformType,
			DataFieldRegistrationID,
		}, nil
	default:
		return nil, fmt.Errorf("%w: %v", ErrUnSupportedMessageCode, msgCode)
	}
}
