package model

import (
	"errors"
	"fmt"
	"strconv"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"
	"carbonblack.com/protocols/psc/ems/server/deviceevent"

	"google.golang.org/protobuf/proto"
)

var ErrDeviceChangeEventMissing = errors.New("device change event is required on deviceEventsRecord")

// DeviceEventsRecord is a wrapper for the device change event that is placed on the device events kinesis stream.
type DeviceEventsRecord struct {
	Event *deviceevent.EmsChangeEvent
}

func NewDeviceEventsRecord(device *Device, eventType deviceevent.EmsChangeEvent_ChangeEventType) *DeviceEventsRecord {
	props := &deviceevent.EmsChangeEvent_DeviceProperties_{
		DeviceProperties: &deviceevent.EmsChangeEvent_DeviceProperties{
			OsType:               deviceauth.DeviceType(device.Type),
			DeploymentType:       deviceevent.DeploymentType(device.DeploymentType),
			ExternallyAssignedId: device.ExternallyAssignedID,
		},
	}

	event := &deviceevent.EmsChangeEvent{
		EventType:  eventType,
		ObjectId:   int64(device.ID),
		ObjectType: deviceevent.EmsChangeEvent_DEVICE,
		OrgKey:     device.OrgKey,
		Name:       device.Hostname,
		Properties: props,
	}

	return &DeviceEventsRecord{Event: event}
}

// Marshal will marshal the device event record into a slice of bytes.
func (rec *DeviceEventsRecord) Marshal() ([]byte, error) {
	if rec.Event == nil {
		return []byte{}, ErrDeviceChangeEventMissing
	}

	bytes, err := proto.Marshal(rec.Event)
	if err != nil {
		return []byte{}, fmt.Errorf("failed to marshal protobuf to bytes: %w", err)
	}

	return bytes, nil
}

// PartitionKey will return the kinesis partition key for the device event.
func (rec *DeviceEventsRecord) PartitionKey() (string, error) {
	if rec.Event == nil {
		return "", ErrDeviceChangeEventMissing
	}

	return strconv.Itoa(int(rec.Event.ObjectId)), nil
}
