//nolint:testpackage // testing internal helper funcs
package model

import (
	"errors"

	"github.com/hashicorp/go-multierror"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Device", func() {
	var mulitErr *multierror.Error

	Context("NewDevice", func() {
		It("should return struct without error when given valid data", func() {
			data := map[string]string{
				DataFieldRegistrationID:             "1-1",
				DataFieldDeviceType:                 "1",
				DataFieldOSMajorVersion:             "2320",
				DataFieldOSMinorVersion:             "5",
				DataFieldMaintenanceRelease:         "5",
				DataFieldOSBuildNumber:              "5",
				DataFieldPlatformType:               "2",
				DataFieldProductVersion:             "1",
				DataFieldOrgKey:                     "R76KWHRN",
				DataFieldEchoDummyResponse:          "false",
				DataFieldEmail:                      "testEmail",
				DataFieldExternallyAssignedDeviceID: "testExternalID",
				DataFieldHostname:                   "testHost",
				DataFieldVDIProvider:                "CITRIX",
				DataFieldMessageType:                "8370",
			}

			device, err := NewDevice(data)
			Expect(device).ToNot(BeNil())
			Expect(err).To(BeNil())
			Expect(device.DeploymentType).To(Equal(DeploymentTypeVDI))
			Expect(device.DryRunMsg).To(Equal(false))
			Expect(device.Email).To(Equal("testEmail"))
			Expect(device.Env).To(Equal("local"))
			Expect(device.ExternallyAssignedID).To(Equal("testExternalID"))
			Expect(device.Hostname).To(Equal("testHost"))
			Expect(device.ID).To(Equal(1))
			Expect(device.OrgID).To(Equal(1))
			Expect(device.OrgKey).To(Equal("R76KWHRN"))
			Expect(device.PropsJSON).ToNot(Equal(""))
			Expect(device.Type).To(Equal(1))
			Expect(device.InfrastructureProvider).To(Equal(InfrastructureProviderNone))
			Expect(device.VDIProvider).To(Equal(VDIProviderCitrix))
		})

		It("should return error when given invalid registration ID", func() {
			data := map[string]string{
				DataFieldRegistrationID:     "1",
				DataFieldDeviceType:         "abc",
				DataFieldOSMajorVersion:     "2320",
				DataFieldOSMinorVersion:     "5",
				DataFieldMaintenanceRelease: "5",
				DataFieldOSBuildNumber:      "5",
				DataFieldPlatformType:       "2",
				DataFieldOrgKey:             "R76KWHRN",
				DataFieldMessageType:        "8193",
			}
			device, err := NewDevice(data)
			Expect(err).ToNot(BeNil())
			Expect(device).To(BeNil())
			Expect(errors.As(err, &mulitErr)).To(Equal(true))
			Expect(err.Error()).To(ContainSubstring("failed to parse registration ID"))
		})

		It("should return error when given invalid device properties data", func() {
			data := map[string]string{
				DataFieldRegistrationID:     "1-1",
				DataFieldDeviceType:         "1",
				DataFieldOSMajorVersion:     "abc",
				DataFieldOSMinorVersion:     "5",
				DataFieldMaintenanceRelease: "5",
				DataFieldOSBuildNumber:      "5",
				DataFieldPlatformType:       "2",
				DataFieldOrgKey:             "R76KWHRN",
				DataFieldMessageType:        "8193",
			}
			device, err := NewDevice(data)
			Expect(err).ToNot(BeNil())
			Expect(device).To(BeNil())
			Expect(errors.As(err, &mulitErr)).To(Equal(true))
			Expect(err.Error()).To(ContainSubstring("failed to get properties JSON"))
		})

		It("should return error when given invalid device type", func() {
			data := map[string]string{
				DataFieldRegistrationID:     "1-1",
				DataFieldDeviceType:         "abc",
				DataFieldOSMajorVersion:     "2320",
				DataFieldOSMinorVersion:     "5",
				DataFieldMaintenanceRelease: "5",
				DataFieldOSBuildNumber:      "5",
				DataFieldPlatformType:       "2",
				DataFieldOrgKey:             "R76KWHRN",
				DataFieldMessageType:        "8193",
			}
			device, err := NewDevice(data)
			Expect(err).ToNot(BeNil())
			Expect(device).To(BeNil())
			Expect(errors.As(err, &mulitErr)).To(Equal(true))
			Expect(err.Error()).To(ContainSubstring("failed to get device type"))
		})

		It("should return error when status message missing required EMD fields", func() {
			data := map[string]string{
				DataFieldRegistrationID:             "",
				DataFieldDeviceType:                 "",
				DataFieldOSMajorVersion:             "",
				DataFieldOSMinorVersion:             "",
				DataFieldMaintenanceRelease:         "",
				DataFieldOSBuildNumber:              "5",
				DataFieldPlatformType:               "",
				DataFieldProductVersion:             "",
				DataFieldOrgKey:                     "",
				DataFieldEchoDummyResponse:          "false",
				DataFieldEmail:                      "testEmail",
				DataFieldExternallyAssignedDeviceID: "testExternalID",
				DataFieldHostname:                   "",
				DataFieldMessageType:                "8193",
			}

			err := validateEMDFields(data)

			Expect(err).ToNot(BeNil())
			Expect(errors.As(err, &mulitErr)).To(Equal(true))
			Expect(err.Error()).To(ContainSubstring("status message missing field: DeviceType"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: Hostname"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: MaintenanceRelease"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: OrgKey"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: OsMajorVersion"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: OsMinorVersion"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: PlatformType"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: ProductVersion"))
			Expect(err.Error()).To(ContainSubstring("status message missing field: RegistrationID"))
		})

		It("should return error when message is missing message type field", func() {
			data := map[string]string{}

			err := validateEMDFields(data)

			Expect(err).ToNot(BeNil())
			Expect(err.Error()).To(ContainSubstring("unsupported message code"))
		})
	})
})
