package model

import (
	"fmt"
	"strconv"

	"github.com/rs/zerolog/log"
)

const (
	windowsServer2019FirstBuild = 17623
)

//nolint:funlen // long but easy, a single return statement of a map constant
func windowsVersions() map[platformEnum]map[osMajorEnum]map[string]string {
	return map[platformEnum]map[osMajorEnum]map[string]string{
		platformClientX86: {
			osMajorXP: {
				"1": "XP x86",
			},
			osMajorWindows: {
				"0": "Vista x86",
				"1": "Windows 7 x86",
				"2": "Windows 8 x86",
				"3": "Windows 8.1 x86",
			},
			osMajorWindows10: {
				"0": "Windows 10 x86",
			},
		},
		platformClientX64: {
			osMajorWindows: {
				"0": "Vista x64",
				"1": "Windows 7 x64",
				"2": "Windows 8 x64",
				"3": "Windows 8.1 x64",
			},
			osMajorWindows10: {
				"0": "Windows 10 x64",
			},
			osMajorWindows11: {
				"0": "Windows 11 x64",
			},
		},
		platformServerX86: {
			osMajorXP: {
				"2": "Server 2003 x86",
			},
			osMajorWindows: {
				"0": "Server 2008 x86",
			},
		},
		platformServerX64: {
			osMajorWindows: {
				"0": "Server 2008 x64",
				"1": "Server 2008 R2 x64",
				"2": "Server 2012 x64",
				"3": "Server 2012 R2 x64",
			},
			osMajorWindows10: {
				"0": "Windows Server 2016 x64",
			},
			osMajorWindows11: {
				"0": "Windows Server 2022 x64",
			},
		},
		platformClientARM64: {
			osMajorWindows10: {
				"0": "Windows 10 arm64",
			},
			osMajorWindows11: {
				"0": "Windows 11 arm64",
			},
		},
		platformServerARM64: {
			osMajorWindows10: {
				"0": "Windows Server arm64",
			},
		},
	}
}

func macKits() map[sensorKitEnum]bool {
	return map[sensorKitEnum]bool{
		sensorKitMac:    true,
		sensorKitMacOSX: true,
	}
}

func linuxKits() map[sensorKitEnum]bool {
	return map[sensorKitEnum]bool{
		sensorKitAmazonLinux: true,
		sensorKitGoogleLinux: true,
		sensorKitRHEL:        true,
		sensorKitSUSE:        true,
		sensorKitUbuntu:      true,
	}
}

func windowsKits() map[sensorKitEnum]bool {
	return map[sensorKitEnum]bool{
		sensorKitWindows: true,
		sensorKitXP:      true,
	}
}

//nolint:funlen // long but easy, a single return statement of a map constant
func sensorKits() map[osMajorEnum]sensorKitEnum {
	return map[osMajorEnum]sensorKitEnum{
		osMajorXP:        sensorKitXP,
		osMajorWindows:   sensorKitWindows,
		osMajorWindows10: sensorKitWindows,
		osMajorWindows11: sensorKitWindows,

		osMajorMacOSX:  sensorKitMac,
		osMajorMacOS11: sensorKitMac,
		osMajorMacOS12: sensorKitMac,
		osMajorMacOS13: sensorKitMac,
		osMajorMacOS14: sensorKitMac,

		osMajorUnknownRHEL:        sensorKitRHEL,
		osMajorUnknownUbuntu:      sensorKitUbuntu,
		osMajorUnknownSUSE:        sensorKitSUSE,
		osMajorUnknownAmazonLinux: sensorKitAmazonLinux,
		osMajorUnknownGoogleLinux: sensorKitGoogleLinux,

		osMajorCentOS6: sensorKitRHEL,
		osMajorCentOS7: sensorKitRHEL,
		osMajorCentOS8: sensorKitRHEL,
		osMajorCentOS9: sensorKitRHEL,

		osMajorRHEL6: sensorKitRHEL,
		osMajorRHEL7: sensorKitRHEL,
		osMajorRHEL8: sensorKitRHEL,
		osMajorRHEL9: sensorKitRHEL,

		osMajorUbuntu12: sensorKitUbuntu,
		osMajorUbuntu14: sensorKitUbuntu,
		osMajorUbuntu16: sensorKitUbuntu,
		osMajorUbuntu18: sensorKitUbuntu,
		osMajorUbuntu19: sensorKitUbuntu,
		osMajorUbuntu20: sensorKitUbuntu,
		osMajorUbuntu21: sensorKitUbuntu,
		osMajorUbuntu22: sensorKitUbuntu,

		osMajorSLES12: sensorKitSUSE,
		osMajorSLES15: sensorKitSUSE,
		osMajorSLES16: sensorKitSUSE,
		osMajorSLES17: sensorKitSUSE,

		osMajorAmazonLinux2:    sensorKitAmazonLinux,
		osMajorAmazonLinux3:    sensorKitAmazonLinux,
		osMajorAmazonLinux2023: sensorKitAmazonLinux,

		osMajorOpenSUSELeap15: sensorKitSUSE,
		osMajorOpenSUSELeap42: sensorKitSUSE,

		osMajorDebian8:  sensorKitUbuntu,
		osMajorDebian9:  sensorKitUbuntu,
		osMajorDebian10: sensorKitUbuntu,
		osMajorDebian11: sensorKitUbuntu,

		osMajorOracle6: sensorKitRHEL,
		osMajorOracle7: sensorKitRHEL,
		osMajorOracle8: sensorKitRHEL,
		osMajorOracle9: sensorKitRHEL,

		osMajorGoogleContainerOptimized: sensorKitGoogleLinux,

		osMajorVMwareESXi8: sensorKitSUSE,

		osMajorAlma8: sensorKitRHEL,
		osMajorAlma9: sensorKitRHEL,

		osMajorRocky8: sensorKitRHEL,
		osMajorRocky9: sensorKitRHEL,

		osMajorUnknown: sensorKitEnum(0),
	}
}

func unknownLinuxVersions() map[osMajorEnum]bool {
	return map[osMajorEnum]bool{
		osMajorUnknownAmazonLinux: true,
		osMajorUnknownGoogleLinux: true,
		osMajorUnknownRHEL:        true,
		osMajorUnknownSUSE:        true,
		osMajorUnknownUbuntu:      true,
	}
}

func slesLinuxVersions() map[osMajorEnum]bool {
	return map[osMajorEnum]bool{
		osMajorSLES12: true,
		osMajorSLES15: true,
		osMajorSLES16: true,
		osMajorSLES17: true,
	}
}

func googleLinuxVersions() map[osMajorEnum]bool {
	return map[osMajorEnum]bool{
		osMajorUnknownGoogleLinux:       true,
		osMajorGoogleContainerOptimized: true,
	}
}

//nolint:funlen // long but easy, a single return statement of a map constant
func representations() map[osMajorEnum]string {
	return map[osMajorEnum]string{
		osMajorXP:        "Windows_XP",
		osMajorWindows:   "Windows",
		osMajorWindows10: "Windows 10",
		osMajorWindows11: "Windows 11",

		osMajorMacOSX:  "MAC OS X 10",
		osMajorMacOS11: "MAC OS 11",
		osMajorMacOS12: "MAC OS 12",
		osMajorMacOS13: "MAC OS 13",
		osMajorMacOS14: "MAC OS 14",

		osMajorUnknownRHEL:        "Linux (Unsupported)",
		osMajorUnknownUbuntu:      "Linux (Unsupported)",
		osMajorUnknownSUSE:        "Linux (Unsupported)",
		osMajorUnknownAmazonLinux: "Linux (Unsupported)",
		osMajorUnknownGoogleLinux: "Linux (Unsupported)",

		osMajorCentOS6: "CentOS 6",
		osMajorCentOS7: "CentOS 7",
		osMajorCentOS8: "CentOS 8",
		osMajorCentOS9: "CentOS 9",

		osMajorRHEL6: "RHEL 6",
		osMajorRHEL7: "RHEL 7",
		osMajorRHEL8: "RHEL 8",
		osMajorRHEL9: "RHEL 9",

		osMajorUbuntu12: "Ubuntu 12",
		osMajorUbuntu14: "Ubuntu 14",
		osMajorUbuntu16: "Ubuntu 16",
		osMajorUbuntu18: "Ubuntu 18",
		osMajorUbuntu19: "Ubuntu 19",
		osMajorUbuntu20: "Ubuntu 20",
		osMajorUbuntu21: "Ubuntu 21",
		osMajorUbuntu22: "Ubuntu 22",

		osMajorSLES12: "SLES 12",
		osMajorSLES15: "SLES 15",
		osMajorSLES16: "SLES 16",
		osMajorSLES17: "SLES 17",

		osMajorAmazonLinux2:    "Amazon Linux 2",
		osMajorAmazonLinux3:    "Amazon Linux 3",
		osMajorAmazonLinux2023: "Amazon Linux 2023",

		osMajorOpenSUSELeap15: "OpenSUSE Leap 15",
		osMajorOpenSUSELeap42: "OpenSUSE Leap 42",

		osMajorDebian8:  "Debian 8",
		osMajorDebian9:  "Debian 9",
		osMajorDebian10: "Debian 10",
		osMajorDebian11: "Debian 11",

		osMajorOracle6: "Oracle 6",
		osMajorOracle7: "Oracle 7",
		osMajorOracle8: "Oracle 8",
		osMajorOracle9: "Oracle 9",

		osMajorGoogleContainerOptimized: "Google Container-Optimized",

		osMajorVMwareESXi8: "VMware ESXi 8",

		osMajorAlma8: "Alma Linux 8",
		osMajorAlma9: "Alma Linux 9",

		osMajorRocky8: "Rocky Linux 8",
		osMajorRocky9: "Rocky Linux 9",

		osMajorUnknown: "Unknown",
	}
}

func isMacVersion(major osMajorEnum) bool {
	return macKits()[sensorKits()[major]]
}

func isLinuxVersion(major osMajorEnum) bool {
	return linuxKits()[sensorKits()[major]]
}

func isUnknownLinuxVersion(major osMajorEnum) bool {
	return unknownLinuxVersions()[major]
}

func isUbuntuVersion(major osMajorEnum) bool {
	return sensorKits()[major] == sensorKitUbuntu
}

func isSLESVersion(major osMajorEnum) bool {
	return slesLinuxVersions()[major]
}

func isGoogle(major osMajorEnum) bool {
	return googleLinuxVersions()[major]
}

func isRHELVersion(major osMajorEnum) bool {
	return sensorKits()[major] == sensorKitRHEL
}

func isWindowsVersion(major osMajorEnum) bool {
	return windowsKits()[sensorKits()[major]]
}

func osRepresentation(majorString string) string {
	majorCode, err := strconv.Atoi(majorString)
	if err != nil {
		return ""
	}

	major := osMajorEnum(majorCode)

	var osRep string

	reps := representations()
	if osRep = reps[major]; osRep == "" {
		return reps[osMajorUnknown]
	}

	return osRep
}

func sensorKit(majorString string) string {
	majorCode, err := strconv.Atoi(majorString)
	if err != nil {
		return ""
	}

	major := osMajorEnum(majorCode)

	return sensorKits()[major].String()
}

func version(
	majorString string,
	minor string,
	maintenance string,
	build string,
	platformString string,
) (string, error) {
	majorCode, err := strconv.Atoi(majorString)
	if err != nil {
		return "", fmt.Errorf("failed to convert major to int: %w, major=%s", err, majorString)
	}

	platformCode, err := strconv.Atoi(platformString)
	if err != nil {
		return "", fmt.Errorf("failed to convert platform to int: %w, platform=%s", err, platformString)
	}

	major := osMajorEnum(majorCode)
	platform := platformEnum(platformCode)

	switch {
	case isWindowsVersion(major):
		return windowsVersion(platform, major, minor, maintenance, build)
	case isLinuxVersion(major):
		return linuxVersion(platform, major, minor, maintenance)
	case isMacVersion(major):
		return fmt.Sprintf("%v.%v.%v", representations()[major], minor, maintenance), nil
	}

	return representations()[osMajorUnknown], nil
}

func windowsVersion(
	platform platformEnum,
	major osMajorEnum,
	minor string,
	maintenance string,
	build string,
) (string, error) {
	log.Debug().Msgf(
		"platform=%s major=%v minor=%s maintenance=%s build=%s",
		platform,
		major,
		minor,
		maintenance,
		build)

	buildNum, err := strconv.Atoi(build)
	if err != nil {
		// temporary for debugging in scl2 testing
		log.Warn().Err(err).Msg("setting build to 0 as it is either invalid or unset")

		buildNum = 0
	}

	version := windowsVersions()[platform][major][minor]

	// DSER-10831: check if this key would cause a collision (due to two servers with the same key) and use build
	// number to distinguish if so
	if version == "Windows Server 2016 x64" && buildNum >= windowsServer2019FirstBuild {
		version = "Windows Server 2019 x64"
	}

	if maintenance != "" {
		version = fmt.Sprintf("%v SP: %v", version, maintenance)
	}

	return version, nil
}

func linuxVersion(platform platformEnum, major osMajorEnum, minor string, maintenance string) (string, error) {
	rep := representations()[major]
	min := minor
	sep := "."
	msep := "."

	switch {
	case isUnknownLinuxVersion(major):
		return representations()[major], nil
	case isUbuntuVersion(major):
		minNum, err := strconv.Atoi(min)
		if err != nil {
			return "", fmt.Errorf("failed to convert minor to int: %w", err)
		}

		min = fmt.Sprintf("%02d", minNum)
	case isSLESVersion(major):
		sep = " SP"
	case isGoogle(major):
		sep = " "
	case isRHELVersion(major):
		msep = "-"
	}

	maint := maintenance
	if maint != "" {
		maint = fmt.Sprintf("%v%v", msep, maintenance)
	}

	return fmt.Sprintf("%v%v%v%v %v", rep, sep, min, maint, platform), nil
}
