package service

import (
	"context"
	"errors"
	"fmt"

	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/client"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
)

var ErrUnknownRequestCode = errors.New("cbd client only services status or metadata requests")

type Cbd interface {
	Update(context.Context, map[string]string) error
}

type CbdService struct {
	enabled   bool
	cbdClient client.CbdCaller
}

func NewCbdService(cbdClient client.CbdCaller, enabled bool) *CbdService {
	log.Warn().Msgf("Cbd Service enabled: %t", enabled)

	return &CbdService{
		enabled,
		cbdClient,
	}
}

func (cbds *CbdService) Update(ctx context.Context, metadata map[string]string) error {
	headers := metadata[model.DataFieldOriginalRequestHeaders]
	message := metadata[model.DataFieldOriginalRequestMsgPack]
	host := metadata[model.DataFieldOriginalRequestHost]

	log.Debug().Msgf("-- CBD Update\n%v\n%v", headers, message)

	dryRunMsg := metadata[model.DataFieldEchoDummyResponse] == "true"

	if cbds.enabled && !dryRunMsg {
		log.Debug().Msg("Cbd Service is enabled, updating through Cbd ...")

		msgCode, err := model.MessageCode(metadata[model.DataFieldMessageType])
		if err != nil {
			return fmt.Errorf("cbd service only supports status or metadata requests | code: %v, err: %w",
				metadata[model.DataFieldMessageType], err)
		}

		switch msgCode {
		case model.DeviceStatusRequestCode:
			if err := cbds.cbdClient.Call(
				ctx, host, viper.GetString(config.CbdClientStatusURI), headers, message); err != nil {
				return fmt.Errorf("status request to cbd client failed %w", err)
			}
		case model.DeviceMetaDataRequestCode:
			if err := cbds.cbdClient.Call(
				ctx, host, viper.GetString(config.CbdClientMetadataURI), headers, message); err != nil {
				return fmt.Errorf("metadta request to cbd client failed %w", err)
			}
		default:
			return fmt.Errorf("%w : unknown code: %s", ErrUnknownRequestCode, msgCode)
		}
	} else {
		log.Debug().Msg("Cbd Service is disabled... not updating through Cbd")
	}

	return nil
}
