package service

import (
	"context"
	"errors"
	"fmt"

	"carbonblack.com/protocols/psc/ems/server/deviceevent"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/data"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"

	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

var ErrInvalidParamsProvided = errors.New("failed to publish device event record")

type DeviceEventsPublisher interface {
	PublishDeviceEventsRecord(context.Context, *model.Device, deviceevent.EmsChangeEvent_ChangeEventType) error
}

type DeviceEventsService struct {
	producer data.Producer
	dryRun   bool
	enabled  bool
}

func NewDeviceEventsService(producer data.Producer, enabled, dryRun bool) *DeviceEventsService {
	svc := &DeviceEventsService{
		producer: producer,
		enabled:  enabled,
		dryRun:   dryRun,
	}

	log.Info().Msgf("device events service enabled in %s: %v", viper.GetString(config.AppEnvKey), svc.enabled)

	return svc
}

// PublishDeviceEventsRecord generates and publishes a device events record to the device events kinesis stream.
func (s *DeviceEventsService) PublishDeviceEventsRecord(
	ctx context.Context,
	device *model.Device,
	eventType deviceevent.EmsChangeEvent_ChangeEventType,
) error {
	if !s.enabled {
		return nil
	}

	rec := model.NewDeviceEventsRecord(device, eventType)

	log.Debug().Msgf("generated device events record %v", rec)

	if s.dryRun {
		log.Debug().Msg("device event service dry run")

		return nil
	}

	dims := map[string]string{
		"type":  rec.Event.GetEventType().String(),
		"error": "false",
	}
	defer telemetry.CounterInc("device-event-stream.put", dims, 1)

	// We should figure out how we want to handle errors while placing record on stream.
	// CBD catches errors and logs them but notes that errors should  be handled and references work on DSER-30635.
	// The ticket is marked as complete but I don't see any work done??
	if err := s.producer.PutRecord(ctx, rec); err != nil {
		dims["error"] = "true"

		return fmt.Errorf("failed to add %s device event to stream for device %d in org %s: %w",
			rec.Event.GetEventType().String(), rec.Event.GetObjectId(), rec.Event.GetOrgKey(), err)
	}

	log.Debug().Msgf("successfully published device events record: %v", rec)

	return nil
}
