package service_test

import (
	"context"
	"errors"

	"carbonblack.com/protocols/psc/ems/server/deviceevent"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/model"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/service"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	helper "gitlab.bit9.local/psc/platform/ems/device-service/test"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
)

var errPutRecord = errors.New("bad record")

var _ = Describe("Device Events Service", func() {
	var (
		svc          *service.DeviceEventsService
		mockProducer *mocks.Producer
		mockWriter   *mocks.Writer
		mockReporter *mocks.MetricReporter
		ctx          context.Context
		oldLogger    zerolog.Logger
	)

	device := &model.Device{}

	Context("service enabled", func() {
		BeforeEach(func() {
			ctx = context.Background()
			mockProducer = mocks.NewProducer(ctrl)
			mockWriter = mocks.NewWriter(ctrl)
			mockReporter = mocks.NewMetricReporter(ctrl)
			telemetry.MetricReporter = mockReporter
			oldLogger = log.Logger
			log.Logger = zerolog.New(mockWriter).Level(zerolog.DebugLevel)
		})

		AfterEach(func() {
			log.Logger = oldLogger
		})

		It("should report metric and log success message when publishing record succeeds", func() {
			var msg string

			mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
				helper.GenerateWriterDoStore(&msg),
			).Times(3)
			mockProducer.EXPECT().PutRecord(ctx, gomock.Any()).Times(1).Return(nil)
			mockReporter.EXPECT().GetOrRegisterMetric("device-event-stream.put", gomock.Any(), gomock.Any()).Times(1)

			svc = service.NewDeviceEventsService(mockProducer, true, false)
			err := svc.PublishDeviceEventsRecord(ctx, device, deviceevent.EmsChangeEvent_UPDATE)

			Expect(err).To(BeNil())
			Expect(msg).To(ContainSubstring("\"message\":\"successfully published device events record: "))
		})

		It("should report metric and return error when publishing record fails", func() {
			mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Times(2)
			mockProducer.EXPECT().PutRecord(ctx, gomock.Any()).Times(1).Return(errPutRecord)
			mockReporter.EXPECT().GetOrRegisterMetric("device-event-stream.put", gomock.Any(), gomock.Any()).Times(1)
			svc = service.NewDeviceEventsService(mockProducer, true, false)

			err := svc.PublishDeviceEventsRecord(ctx, device, deviceevent.EmsChangeEvent_UPDATE)

			Expect(err).ToNot(BeNil())
		})

		It("should not publish record to kinesis during dry run", func() {
			var msg string

			mockWriter.EXPECT().Write(gomock.AssignableToTypeOf([]byte{})).Do(
				helper.GenerateWriterDoStore(&msg),
			).Times(3)

			mockProducer.EXPECT().PutRecord(ctx, gomock.Any()).Times(0)
			svc = service.NewDeviceEventsService(mockProducer, true, true)
			err := svc.PublishDeviceEventsRecord(ctx, device, deviceevent.EmsChangeEvent_UPDATE)
			Expect(err).To(BeNil())
			Expect(msg).To(ContainSubstring("\"message\":\"device event service dry run"))
		})
	})

	Context("service disabled", func() {
		BeforeEach(func() {
			ctx = context.Background()
			mockProducer = mocks.NewProducer(ctrl)
		})

		It("should not publish record to kinesis", func() {
			mockProducer.EXPECT().PutRecord(ctx, gomock.Any()).Times(0)
			svc = service.NewDeviceEventsService(mockProducer, false, true)
			err := svc.PublishDeviceEventsRecord(ctx, device, deviceevent.EmsChangeEvent_UPDATE)
			Expect(err).To(BeNil())
		})
	})
})
