package service

import (
	"context"
	"errors"
	"fmt"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/client"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"

	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

const maxWaitTimeToCloseQueue = 5

var ErrEvictingDevice = errors.New("failed to evict device from cache")

type msgContext struct {
	deviceID int
	orgKey   string
}

type EvictServiceDelayConfig struct {
	evictDelayTime    time.Duration
	evictDelayEnabled bool // enable eviction delays to reduce number of delete calls to ems-device-auth
	maxBatchSize      int
}

type EvictService struct {
	cfg   EvictServiceDelayConfig
	queue chan *msgContext
	stop  chan bool
	done  chan bool
	dac   client.Evictor
}

func NewEvictServiceConfig() EvictServiceDelayConfig {
	return EvictServiceDelayConfig{
		evictDelayTime:    viper.GetDuration(config.EvictDelayTimeKey),
		evictDelayEnabled: viper.GetBool(config.EvictDelayEnabledKey),
		maxBatchSize:      viper.GetInt(config.EvictDelayMaxBatchSizeKey),
	}
}

func NewEvictService(cfg EvictServiceDelayConfig, dac client.Evictor) *EvictService {
	var evictService *EvictService
	if cfg.evictDelayEnabled {
		evictService = &EvictService{
			cfg:   cfg,
			queue: make(chan *msgContext, cfg.maxBatchSize),
			stop:  make(chan bool),
			done:  make(chan bool),
			dac:   dac,
		}
		// launch processor to collect devices to evict, delays eviction calls and attempts to reduce
		// number of API calls to ems-device-auth
		go evictService.receiveDevicesToEvict()
	} else {
		//nolint:exhaustruct
		evictService = &EvictService{
			cfg: cfg,
			dac: dac,
		}
	}

	return evictService
}

func (s *EvictService) EvictDevice(ctx context.Context, orgKey string, deviceID int) error {
	if s.cfg.evictDelayEnabled {
		msg := &msgContext{
			deviceID: deviceID,
			orgKey:   orgKey,
		}
		s.queue <- msg

		return nil
	}

	start := time.Now()
	defer telemetry.HistogramUpdateTimeSince("evict.processTime", map[string]string{"batch": "false"}, start)

	if err := s.dac.EvictDevice(ctx, orgKey, deviceID); err != nil {
		telemetry.CounterInc("evict.error", map[string]string{"batch": "false"}, 1)

		return fmt.Errorf("%w", err)
	}

	return nil
}

// FlushAndClose all opened channels managed by the service, close all downstream client connections.
func (s *EvictService) FlushAndClose() error {
	if s.cfg.evictDelayEnabled {
		s.stop <- true
		// block until go routine has finished processing all messages in queue
		select {
		case <-s.done:
			log.Info().Msg("queue has been closed, go routine exited, service properly closed")
		case <-time.After(time.Duration(maxWaitTimeToCloseQueue) * time.Second):
			log.Warn().Msgf("didn't close queue after waiting %vs", maxWaitTimeToCloseQueue)
		}
	}

	// close client connections
	if err := s.dac.Close(); err != nil {
		return fmt.Errorf("failed to close device auth client: %w", err)
	}

	return nil
}

func (s *EvictService) Close() error {
	return s.FlushAndClose()
}

func (s *EvictService) Name() string {
	return "evict_service"
}

// processDevicesToEvict calls DeviceAuthClient to evict devices found in provided batch.
// Currently, it eats all errors in adding items, future work can be to
// handle it more gracefully (error item channels for eventual retries?).
func (s *EvictService) processDevicesToEvict(devicesToEvict map[string]*msgContext) {
	batchSize := len(devicesToEvict)
	if batchSize == 0 {
		return
	}

	telemetry.GaugeUpdate("evict.batch.size", nil, int64(batchSize))

	start := time.Now()
	defer telemetry.HistogramUpdateTimeSince("evict.processTime", map[string]string{"batch": "true"}, start)

	for key, item := range devicesToEvict {
		if err := s.dac.EvictDevice(context.Background(), item.orgKey, item.deviceID); err != nil {
			telemetry.CounterInc("evict.error", map[string]string{"batch": "true"}, 1)
			log.Err(err).Msg(ErrEvictingDevice.Error())
		}
		// clear item from devicesToEvict (successes+failures), not initially retrying failed evictions as
		// TTL will clear them eventually
		delete(devicesToEvict, key)
	}
}

// receiveDevicesToEvict continuously runs adding items to a set of devices to evict,
// flushes batch once batch is full or if timer is triggered.
// Shuts down processing once queue is closed.
func (s *EvictService) receiveDevicesToEvict() {
	devicesToEvict := make(map[string]*msgContext, s.cfg.maxBatchSize)
	ticker := time.NewTicker(s.cfg.evictDelayTime)

	for {
		select {
		case msg := <-s.queue:
			key := fmt.Sprintf("%s-%d", msg.orgKey, msg.deviceID)

			if devicesToEvict[key] = msg; len(devicesToEvict) >= s.cfg.maxBatchSize {
				s.processDevicesToEvict(devicesToEvict)
				telemetry.CounterInc("evict.batch.triggered", map[string]string{"batch": "full"}, 1)
			}
		case <-ticker.C:
			s.processDevicesToEvict(devicesToEvict)
			telemetry.CounterInc("evict.batch.triggered", map[string]string{"batch": "timer"}, 1)
		case <-s.stop:
			// read all msgs from the queue if the service is stopping
			if len(s.queue) > 0 {
				for msg := range s.queue {
					devicesToEvict[fmt.Sprintf("%s-%d", msg.orgKey, msg.deviceID)] = msg
				}
			}

			if len(devicesToEvict) > 0 {
				log.Info().Msgf("stopping batch processing, processing final batch of size %d", len(devicesToEvict))
				s.processDevicesToEvict(devicesToEvict)
			}

			s.done <- true

			ticker.Stop()

			return
		}
	}
}
