package service_test

import (
	"context"
	"time"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/service"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"

	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
)

var _ = Describe("Evict Service", func() {
	var mockEvictor *mocks.Evictor

	Context("Without delay", func() {
		var svc *service.EvictService

		BeforeEach(func() {
			mockEvictor = mocks.NewEvictor(ctrl)
			viper.Reset()
			telemetry.SetupMetrics(nil)
			svc = service.NewEvictService(service.NewEvictServiceConfig(), mockEvictor)
		})

		It("should return new service", func() {
			Expect(svc).ToNot(BeNil())
		})

		It("should not error and call evictor", func() {
			mockEvictor.EXPECT().EvictDevice(context.Background(), "testOrgKey", 1).Times(1)
			err := svc.EvictDevice(context.Background(), "testOrgKey", 1)
			Expect(err).To(BeNil())
		})

		It("should error when evictor call fails", func() {
			mockEvictor.EXPECT().EvictDevice(context.Background(), "testOrgKey", 1).Return(service.ErrEvictingDevice)
			err := svc.EvictDevice(context.Background(), "testOrgKey", 1)
			Expect(err.Error()).To(Equal(service.ErrEvictingDevice.Error()))
		})
	})

	Context("With delay", func() {
		var mockReporter *mocks.MetricReporter
		BeforeEach(func() {
			mockReporter = mocks.NewMetricReporter(ctrl)
			telemetry.MetricReporter = mockReporter
			mockEvictor = mocks.NewEvictor(ctrl)
		})

		It("should call evict device when map is full ", func() {
			viper.Reset()
			viper.Set(config.EvictDelayEnabledKey, true)
			viper.Set(config.EvictDelayMaxBatchSizeKey, 2)
			viper.Set(config.EvictDelayTimeKey, "100s")
			// don't test order since map can be randomly ordered
			mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1)
			mockEvictor.EXPECT().EvictDevice(context.Background(), "testOrgKey", 2).Times(1)
			mockEvictor.EXPECT().EvictDevice(context.Background(), "a", 1).Times(1)
			mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1)
			mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1)
			mockEvictor.EXPECT().Close().Times(1)

			svc := service.NewEvictService(service.NewEvictServiceConfig(), mockEvictor)
			err := svc.EvictDevice(context.Background(), "testOrgKey", 2)
			Expect(err).To(BeNil())
			err = svc.EvictDevice(context.Background(), "a", 1)
			Expect(err).To(BeNil())
			// wait allowing processor to trigger if batch is full
			time.Sleep(time.Millisecond * 10)
			Expect(svc.FlushAndClose()).To(BeNil())
		})

		It("should call evict device when map isn't full but timer is triggered", func() {
			viper.Reset()
			viper.Set(config.EvictDelayEnabledKey, true)
			viper.Set(config.EvictDelayMaxBatchSizeKey, 2)
			viper.Set(config.EvictDelayTimeKey, "1s")
			gomock.InOrder(
				mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1),
				mockEvictor.EXPECT().EvictDevice(context.Background(), "testOrgKey", 2).Times(1),
				mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1),
				mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1),
				mockEvictor.EXPECT().Close().Times(1),
			)

			svc := service.NewEvictService(service.NewEvictServiceConfig(), mockEvictor)
			err := svc.EvictDevice(context.Background(), "testOrgKey", 2)
			Expect(err).To(BeNil())
			// wait allowing processor to trigger once timer is hit
			time.Sleep(time.Millisecond * 1010)
			Expect(svc.FlushAndClose()).To(BeNil())
		})

		It("should call evict device when map isn't full and timer isn't triggered but stopped received", func() {
			viper.Reset()
			viper.Set(config.EvictDelayEnabledKey, true)
			viper.Set(config.EvictDelayTimeKey, "100s")
			viper.Set(config.EvictDelayMaxBatchSizeKey, 100)
			gomock.InOrder(
				mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1),
				mockEvictor.EXPECT().EvictDevice(context.Background(), "testOrgKey", 2).Times(1),
				mockReporter.EXPECT().GetOrRegisterMetric(gomock.Any(), gomock.Any(), gomock.Any()).Times(1),
				mockEvictor.EXPECT().Close().Times(1),
			)

			svc := service.NewEvictService(service.NewEvictServiceConfig(), mockEvictor)
			err := svc.EvictDevice(context.Background(), "testOrgKey", 2)
			Expect(err).To(BeNil())
			time.Sleep(time.Millisecond * 1000)
			Expect(svc.FlushAndClose()).To(BeNil())
		})
	})
})
