package service

import (
	"errors"
	"fmt"
	"sync"

	"github.com/hashicorp/go-multierror"
)

var ErrAppNotReady = errors.New("can't connect to downstream dependencies: %v")

// ReadinessService wraps multiple readiness probes to provide a single IsReady call.
type ReadinessService struct {
	probes []ReadinessChecker
}

type ReadinessChecker interface {
	IsReady() error
}

func NewReadinessService(probes ...ReadinessChecker) *ReadinessService {
	return &ReadinessService{probes: probes}
}

func (s *ReadinessService) IsReady() error {
	var mErr *multierror.Error

	probing := &sync.WaitGroup{}
	probing.Add(len(s.probes))

	// attempt connections to all downstream dependencies
	for _, probe := range s.probes {
		go func(probe ReadinessChecker) {
			if err := probe.IsReady(); err != nil {
				mErr = multierror.Append(mErr, err)
			}

			probing.Done()
		}(probe)
	}

	probing.Wait()

	if mErr != nil {
		//nolint:errorlint // already wrapped an error
		return fmt.Errorf("%w: %v", ErrAppNotReady, mErr)
	}

	return nil
}
