package telemetry_test

import (
	"time"

	"github.com/golang/mock/gomock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry"
	"gitlab.bit9.local/psc/platform/ems/device-service/test/mocks"
)

var _ = Describe("SetupMetrics", func() {
	const metricName = "test"

	Context("SetUpMetrics", func() {
		It("should create a reporter", func() {
			telemetry.SetupMetrics(&telemetry.NoopSender{})
			Expect(telemetry.MetricReporter).ToNot(BeNil())
			telemetry.MetricReporter.Start()
			defer telemetry.MetricReporter.Close()

			telemetry.MetricReporter.GetOrRegisterMetric(metricName, 1, map[string]string{})
			defer telemetry.MetricReporter.UnregisterMetric(metricName, map[string]string{})

			telemetry.MetricReporter.Report()
			Expect(telemetry.MetricReporter.ErrorsCount()).To(BeZero())
		})
	})

	Context("IncCounter", func() {
		It("should get or register the metric and then increment", func() {
			mockReporter := mocks.NewMetricReporter(ctrl)
			mockCounter := mocks.NewCounter(ctrl)
			dims := map[string]string{metricName: "one"}
			incrBy := int64(5)

			telemetry.MetricReporter = mockReporter

			gomock.InOrder(
				mockReporter.EXPECT().GetOrRegisterMetric(metricName, gomock.Any(), dims).Return(mockCounter).Times(1),
				mockCounter.EXPECT().Inc(incrBy).Times(1),
			)

			telemetry.CounterInc(metricName, dims, incrBy)
		})
	})

	Context("HistogramUpdateTimeSince", func() {
		It("should get or register the metric and then update since", func() {
			mockReporter := mocks.NewMetricReporter(ctrl)
			mockHistogram := mocks.NewHistogram(ctrl)
			metricName := "test"
			dims := map[string]string{"test": "one"}
			start := time.Now()

			telemetry.MetricReporter = mockReporter

			gomock.InOrder(
				mockReporter.EXPECT().GetOrRegisterMetric(metricName, gomock.Any(), dims).Return(mockHistogram).Times(1),
				mockHistogram.EXPECT().Update(time.Since(start).Milliseconds()).Times(1),
			)

			telemetry.HistogramUpdateTimeSince(metricName, dims, start)
		})
	})

	Context("UpdateGauge", func() {
		It("should get or register the metric and then increment", func() {
			mockReporter := mocks.NewMetricReporter(ctrl)
			mockGauge := mocks.NewGauge(ctrl)
			metricName := "test"
			dims := map[string]string{"test": "one"}
			val := int64(5)

			telemetry.MetricReporter = mockReporter

			gomock.InOrder(
				mockReporter.EXPECT().GetOrRegisterMetric(metricName, gomock.Any(), dims).Return(mockGauge).Times(1),
				mockGauge.EXPECT().Update(val).Times(1),
			)

			telemetry.GaugeUpdate(metricName, dims, val)
		})
	})
})
