package wavefront_test

import (
	"errors"
	"testing"

	"gitlab.bit9.local/psc/platform/ems/device-service/internal/config"
	"gitlab.bit9.local/psc/platform/ems/device-service/internal/telemetry/wavefront"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/spf13/viper"
)

func TestWavefront(t *testing.T) {
	t.Parallel()
	RegisterFailHandler(Fail)
	RunSpecs(t, "Wavefront")
}

var _ = Describe("InitSender", func() {
	BeforeEach(func() {
		viper.Reset()

		err := config.Init()
		Expect(err).To(HaveOccurred())
		Expect(errors.Unwrap(err)).Should(BeAssignableToTypeOf(viper.ConfigFileNotFoundError{}))
	})

	Context("with no config", func() {
		It("should return a sender", func() {
			sender := wavefront.InitSender()
			Expect(sender).NotTo(BeNil())
		})
	})

	Context("with disabled config", func() {
		BeforeEach(func() {
			viper.Set(config.WFModeKey, "disabled")
		})

		It("should not return a sender", func() {
			sender := wavefront.InitSender()
			Expect(sender).To(BeNil())
		})
	})

	Context("with unknown config", func() {
		BeforeEach(func() {
			viper.Set(config.WFModeKey, "wat?")
		})

		It("should not return a sender", func() {
			sender := wavefront.InitSender()
			Expect(sender).To(BeNil())
		})
	})

	Context("with empty config", func() {
		BeforeEach(func() {
			viper.Set(config.WFModeKey, "")
		})

		It("should not return a sender", func() {
			sender := wavefront.InitSender()
			Expect(sender).To(BeNil())
		})
	})

	Context("with config", func() {
		Context("that is good", func() {
			Context("for direct sending", func() {
				BeforeEach(func() {
					viper.Set(config.WFModeKey, "direct")
					viper.Set(config.WFTokenKey, "a-token")
				})

				It("should return a direct sender", func() {
					sender := wavefront.InitSender()
					Expect(sender).NotTo(BeNil())
				})
			})

			Context("for proxy sending", func() {
				BeforeEach(func() {
					viper.Set(config.WFModeKey, "proxy")
				})

				It("should return a proxy sender", func() {
					sender := wavefront.InitSender()
					Expect(sender).NotTo(BeNil())
				})
			})
		})

		Context("that is bad", func() {
			Context("for direct sending", func() {
				BeforeEach(func() {
					viper.Set(config.WFModeKey, "direct")
				})

				Context("with no token", func() {
					It("should not return a direct sender", func() {
						sender := wavefront.InitSender()
						Expect(sender).To(BeNil())
					})
				})

				Context("with a bad token", func() {
					It("should not return a direct sender", func() {
						viper.Set(config.WFTokenKey, "bad-token\\")
						sender := wavefront.InitSender()
						Expect(sender).To(BeNil())
					})
				})
			})

			Context("for proxy sending", func() {
				BeforeEach(func() {
					viper.Set(config.WFModeKey, "proxy")
				})

				Context("with a bad host", func() {
					It("should not return a direct sender", func() {
						viper.Set(config.WFHostKey, "\\bad-host")
						sender := wavefront.InitSender()
						Expect(sender).To(BeNil())
					})
				})
			})
		})
	})
})
